/**
 * Dasu Project
 *
 * File:         DasuConfiguration.java
 * RCS:          $Id$
 * Description:  DasuConfiguration class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 24, 2010 at 2:23:49 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Properties;
import java.util.UUID;
import java.util.Map.Entry;
import java.util.logging.Logger;

import com.azureus.plugins.aznetmon.main.RSTPacketStats;

import edu.northwestern.dasu.measurement.NetstatBanwidthPerformer;
import edu.northwestern.dasu.stats.Statistics;

/**
 * @author Mario Sanchez / David Choffnes &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The DasuConfiguration class ...
 */

public class DasuConfiguration {
	
	private final static Logger LOGGER = Logger.getLogger(DasuConfiguration.class.getName());
	
	private static DasuConfiguration self = null;
	
	public static final String DASU_SERVER_LIST_LOC = "http://www.aqua-lab.org/dasu/ws/dasu-list.dat";
	public static final String DASU_DEFAULT_SERVER = "http://haleakala.cs.northwestern.edu";

//	public static final String DASU_SERVER_LIST_LOC = "http://mas939-imac.cs.northwestern.edu/dasu/ws/dasu-list.dat";
//	public static final String DASU_DEFAULT_SERVER = "http://mas939-imac.cs.northwestern.edu";

	public static final String DASU_DEFAULT_REPORTER_SCRIPT = "/dasu/ws/reportStats.php";
	public static final String NDT_SCRIPT = "/dasu/ws/ndt.php";
	public static final String CONFIG_LIST = "/dasu/ws/config-list.dat";
	public static final String INTERF_TYPES_SCRIPT = "/dasu/ws/InterferenceTypes.php";
	public static final String AME_SCRIPT = "/dasu/ws/AMExperiment.php";
	public static final String DASU_CONFIG = "/dasu/ws/dasuConfig.properties";
	public static final String DASU_MESSAGES = "/dasu/ws/message-list.dat";	
	
	private boolean recordStats = true;
	private boolean sendToDatabase = true;
	private boolean enablePopups = true;
	private boolean enableAzureusSpeedTest = true;
	private int peerUpdateIntervalSec = 30; //interval (seconds) for subsequent updates for the peerManager et al.
	private boolean doTraceroute = true;
	private boolean doAnaLookup = true;
	private int tracerouteTimeout = 60; //seconds
	private int dnsTimeout = 1; //seconds
	private int dbUpdateFreqSec = 30; //seconds
	private int updateCheckFreqSec = 43200;
	private int maxPeers = 100;
	private String uuid = null;
	private int UUIDid = -1;
	private int schedulerInterval = 302; //5*60 + 2 seconds
	private int AMCSleepInterval = 30; //seconds
	private int managerSleepInterval = 30; //seconds
	private int AMECheckupInterval = 5*60; //seconds
	private int messageManagerInterval = 60*60; //60*60 seconds = 1 hour
	private int dbStaticFlushInterval = 15*60; //seconds (15 min)
	private int DBManagerSleepInterval = 2*60; //seconds (2 min)

	private int maxLogEntries = 200;
	private int maxHistoricSignals = 20;
	private int maxNumEntriesPerHistoricSignal = 370;
	private float inferenceVersion = 2.0F;
	
	private final int maxHistoricSignalsHARDLIMIT = 40;
	private final int maxNumEntriesPerHistoricSignalHARDLIMIT = 740;
	private final int maxPingsHARDLIMIT = 40;

	private int dbReconnectInterval = 15*60; // 10 minutes to retry
	private int reportingTimeout = 60;
	private int maxPings = 10;
	private boolean streamDebug = false;
	private boolean runExperiments = true;
	private boolean visualize=false;
	private boolean ignoreStateChangeRST=false;
	private int stateChangeRSTInterval = 2; // seconds

	
	/** the sample interval for news, in seconds */
	private int sampleInterval = 30;

	private boolean override;
	private int pctDiffTrigger = 10;
	
	private boolean suppressWarnings = false;
	
	/** if true, override the user's settings with those of the server */
	private boolean globalOverride = false;
	
	private int RSTimeWindowSize = 1; /*minutes*/
	private int timeSpan = 5; /*minutes*/
	
	public int getDbUpdateFreqSec() {
		return dbUpdateFreqSec;
	}


	public void setDbUpdateFreqSec(int dbUpdateFreqSec) {
		this.dbUpdateFreqSec = dbUpdateFreqSec;
	}
	
	public int getTracerouteTimeout() {
		return tracerouteTimeout;
	}

	public void setTracerouteTimeout(int tracerouteTimeout) {
		this.tracerouteTimeout = tracerouteTimeout;
	}
	
	public int getDnsTimeout() {
		return dnsTimeout;
	}

	public void setDnsTimeout(int dnsTimeout) {
		this.dnsTimeout = dnsTimeout;
	}

	public int getMaxPeers() {
		return maxPeers;
	}

	public void setMaxPeers(int maxPeers) {
		this.maxPeers = maxPeers;
	}

	public int getUpdateCheckFreqSec() {
		return updateCheckFreqSec;
	}

	public void setUpdateCheckFreqSec(int updateCheckFreqSec) {
		this.updateCheckFreqSec = updateCheckFreqSec;
	}

	public synchronized static DasuConfiguration getInstance(){
		
		if (self==null){
			self = new DasuConfiguration();
		}
		return self;
	}

	public boolean isDoTraceroute() {
		return doTraceroute;
	}

	public boolean isDoAnaLookup() {
		return doAnaLookup;
	}
	
	public void setDoTraceroute(boolean doTraceroute) {
		this.doTraceroute = doTraceroute;
	}

	public void setDoAnaLookup(boolean doAnaLookup) {
		this.doAnaLookup = doAnaLookup;
	}
	
	public int getPeerUpdateIntervalSec() {
		return peerUpdateIntervalSec;
		//return 1;
	}


	public void setPeerUpdateIntervalSec(int peerUpdateIntervalSec) {
		this.peerUpdateIntervalSec = peerUpdateIntervalSec;
	}


	public boolean isRecordStats() {
		return recordStats;
	}


	public void setRecordStats(boolean recordStats) {
		this.recordStats = recordStats;
	}
	
	public boolean isEnablePopups() {
		return enablePopups;
	}

	public void setEnablePopups(boolean enablePopups) {
		this.enablePopups = enablePopups;
	}

	public boolean isEnableAzureusSpeedTest() {
		return enableAzureusSpeedTest;
	}


	public void setEnableAzureusSpeedTest(boolean enableAzureusSpeedTest) {
		this.enableAzureusSpeedTest = enableAzureusSpeedTest;
	}

	public boolean isSendToDatabase() {
		return sendToDatabase;
	}


	public void setSendToDatabase(boolean sendToDatabase) {
		this.sendToDatabase = sendToDatabase;
	}

	public void writeProperties(Properties props, ArrayList<String> locations) {
		props.put("dasu.enableStats", recordStats ? "true" : "false");
		props.put("dasu.enablePopups", enablePopups ? "true" : "false");
		props.put("dasu.enableAzureusSpeedTest", enableAzureusSpeedTest ? "true" : "false");
		props.put("dasu.sendToDatabase", sendToDatabase ? "true" : "false");		
		props.put("dasu.override", override ? "true" : "false");
		props.put("dasu.stats.dbUpdate", dbUpdateFreqSec+"" );
		props.put("dasu.stats.tracerouteTimeout", tracerouteTimeout+"" );
		props.put("dasu.stats.dnsTimeout", dnsTimeout+"" );
		props.put("dasu.update.updateCheck", updateCheckFreqSec+"");
		props.put("dasu.doTraceroute", doTraceroute ? "true" : "false");
		props.put("dasu.doAnaLookup", doAnaLookup ? "true" : "false");
		props.put("dasu.stats.peerUpdate", peerUpdateIntervalSec+"");
		props.put("dasu.reportingTimeout", reportingTimeout+"");
		props.put("dasu.maxPings", maxPings+"");
		props.put("dasu.uuid", uuid);
		props.put("dasu.UUIDid", UUIDid+"");
		props.put("dasu.streamDebug", streamDebug ? "true" : "false");
		props.put("dasu.runExperiments", runExperiments ? "true" : "false");
		props.put("dasu.visualize", visualize ? "true" : "false");
		props.put("dasu.ignoreStateChangeRST", ignoreStateChangeRST ? "true" : "false");
		props.put("dasu.sampleInterval", sampleInterval+"");
		props.put("dasu.suppressWarnings", suppressWarnings ? "true" : "false");
		props.put("dasu.globalOverride", globalOverride ? "true" : "false");
		props.put("dasu.schedulerInterval", schedulerInterval+"");
		props.put("dasu.messageManagerInterval", messageManagerInterval+"");
		props.put("dasu.AMECheckupInterval", AMECheckupInterval+"");
		props.put("dasu.stateChangeRSTInterval", stateChangeRSTInterval+"");
		props.put("dasu.AMCSleepInterval", AMCSleepInterval+"");
		props.put("dasu.managerSleepInterval", managerSleepInterval+"");
		props.put("dasu.db.staticFlushInterval", dbStaticFlushInterval+"");
		props.put("dasu.db.managerSleepInterval", DBManagerSleepInterval+"");
		props.put("dasu.maxLogEntries", maxLogEntries+"");
		props.put("dasu.maxHistoricSignals", maxHistoricSignals+"");
		props.put("dasu.maxNumEntriesPerHistoricSignal", maxNumEntriesPerHistoricSignal+"");
		props.put("dasu.inference.version", inferenceVersion+"");
		props.put("dasu.RSTimeWindowSize", RSTimeWindowSize+"");
		props.put("dasu.RSTimeSpan", timeSpan+"");


		for (String filename : locations){
			try {
				FileOutputStream fos = new FileOutputStream(filename);
				props.store(fos, "");
				break;
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
	}

	public void loadProperties(Properties props) {

		LOGGER.info("Loading properties from server");
		
		for (Entry<Object, Object> ent : props.entrySet()){
			try {
			String key = (String)ent.getKey();
			String value = (String)ent.getValue();
			if (key.equals("dasu.enableStats")){
				recordStats= value.equals("true") ? true : false;
			} else if (key.equals("dasu.stats.dbUpdate")){
				setDbUpdateFreqSec(Integer.valueOf(value));
			} else if (key.equals("dasu.stats.tracerouteTimeout")){
				setTracerouteTimeout(Integer.valueOf(value));
			} else if (key.equals("dasu.stats.dnsTimeout")){
				setDnsTimeout(Integer.valueOf(value));
			} else if (key.equals("dasu.update.updateCheck")){
				setUpdateCheckFreqSec(Integer.valueOf(value));
			}else if (key.equals("dasu.doTraceroute")){
				doTraceroute= value.equals("true") ? true : false;
			}else if (key.equals("dasu.runExperiments")){
				runExperiments= value.equals("true") ? true : false;
			}else if (key.equals("dasu.doAnaLookup")){
				doAnaLookup= value.equals("true") ? true : false;
			}else if (key.equals("dasu.enablePopups")){
				enablePopups= value.equals("true") ? true : false;
			}else if (key.equals("dasu.enableAzureusSpeedTest")){
				enableAzureusSpeedTest= value.equals("true") ? true : false;
			}else if (key.equals("dasu.sendToDatabase")){
				sendToDatabase= value.equals("true") ? true : false;			
			}else if (key.equals("dasu.stats.peerUpdate")){
				setPeerUpdateIntervalSec(Integer.valueOf(value));
			}else if (key.equals("dasu.reportingTimeout")){
				setReportingTimeout(Integer.valueOf(value));
			} else if (key.equals("dasu.maxPings")){
				setMaxPings(Integer.valueOf(value));
			} else if (key.equals("dasu.uuid")){
				uuid = value;
			} else if (key.equals("dasu.UUIDid")){
				setUUIDid(Integer.valueOf(value));
			} else if (key.equals("dasu.streamDebug")){
				streamDebug = value.equals("true") ? true : false;;
			} else if (key.equals("dasu.visualize")){
				visualize = value.equals("true") ? true : false;
			} else if (key.equals("dasu.ignoreStateChangeRST")){
				ignoreStateChangeRST = value.equals("true") ? true : false;
			} else if (key.equals("dasu.sampleInterval")){
				sampleInterval = Integer.valueOf(value);
				RSTPacketStats.INTERVAL = sampleInterval*1000;
				//Main.getInstance().updateRSTTimer();
			} else if (key.equals("dasu.override")){
				override= value.equals("true") ? true : false;
			} else if (key.equals("dasu.pctDiffTrigger")){
				setPctDiffTrigger(Integer.valueOf(value));
			} else if (key.equals("dasu.suppressWarnings")){
				suppressWarnings = value.equals("true") ? true : false;
			} else if (key.equals("dasu.globalOverride")){
				globalOverride = value.equals("true") ? true : false;
			}else if (key.equals("dasu.inference.version")){
				setInferenceVersion(Float.valueOf(value));
			}else if (key.equals("dasu.schedulerInterval")){
				setSchedulerInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.messageManagerInterval")){
				setMessageManagerInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.AMECheckupInterval")){
				setAMECheckupInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.stateChangeRSTInterval")){
				setStateChangeRSTInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.AMCSleepInterval")){
				setAMCSleepInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.db.managerSleepInterval")){
				setDBManagerSleepInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.db.staticFlushInterval")){
				setDbStaticFlushInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.managerSleepInterval")){
				setManagerSleepInterval(Integer.valueOf(value));
			}else if (key.equals("dasu.maxLogEntries")){
				setMaxLogEntries(Integer.valueOf(value));
			}else if (key.equals("dasu.maxHistoricSignals")){
				setMaxHistoricSignals(Integer.valueOf(value));
			}else if (key.equals("dasu.maxNumEntriesPerHistoricSignal")){
				setMaxNumEntriesPerHistoricSignal(Integer.valueOf(value));
			}else if (key.equals("dasu.RSTimeWindowSize")){
				setRSTimeWindowSize(Integer.valueOf(value));
			}else if (key.equals("dasu.RSTimeSpan")){
				setTimeSpan(Integer.valueOf(value));
			} else {
				LOGGER.warning("Unknown key: "+key);
			}
			} catch (Exception e){
				System.err.println("Error processing property...");
				e.printStackTrace();
			} catch (Error e){
				System.err.println("Error processing property...");
				e.printStackTrace();
			}
			
		}
		
		boolean write = false;
		// if no uuid, generate it now
		if (uuid==null){
			uuid = UUID.randomUUID().toString();
			props.put("dasu.uuid", uuid);
			write = true;
		}
		
		if (UUIDid == -1) {
			
			UUIDid = Statistics.getInstance().makeUUIDid(uuid);
			if (UUIDid != -1) {
				props.put("dasu.UUIDid", UUIDid);
				write = true;
			}
		} 
		
		if (write)
			writeProperties(props, Main.getPropertiesLocations());
		
		LOGGER.info("UUID " + uuid + " " + UUIDid);
		
	}


	public int getDbReconnectInterval() {
		return dbReconnectInterval;
	}


	public void setDbReconnectInterval(int dbReconnectInterval) {
		this.dbReconnectInterval = dbReconnectInterval;
	}

	public void setReportingTimeout(int timeout) {
		this.reportingTimeout = timeout;
	}

	public int getReportingTimeout() {
		return reportingTimeout;
	}


	public void setMaxPings(int value) {
		
		if (value <= maxPingsHARDLIMIT)
			this.maxPings = value;
		
	}
	
	public int getMaxPings(){
		return maxPings;
	}

	public void setInferenceVersion(float value) {
		this.inferenceVersion = value;
	}

	public float getInferenceVersion(){
		return inferenceVersion;
	}

	public String getUUID() {
		return uuid;
	}
	
	public int getUUIDid() {
		return UUIDid;
	}
	
	public void setUUID(String uuid){
		this.uuid = uuid;
	}

	public void setUUIDid(int uuidid){
		this.UUIDid = uuidid;
	}

	public boolean isStreamDebug() {
		// TODO Auto-generated method stub
		return this.streamDebug;
	}
	
	public void setStreamDebug(boolean streamDebug){
		this.streamDebug = streamDebug;
	}


	public boolean isRunExperiments() {
		// TODO Auto-generated method stub
		return runExperiments;
	}
	
	public void setRunExperiments(boolean runExperiments){
		this.runExperiments = runExperiments;
	}


	public boolean isVisualize() {
		// TODO Auto-generated method stub
		return visualize;
	}

	public boolean isIgnoreStateChangeRST() {
		return ignoreStateChangeRST;
	}
	public int getSampleInterval() {
		return sampleInterval;
	}


	public void setSampleInterval(int sampleInterval) {
		this.sampleInterval = sampleInterval;
	}


	public void setOverride(boolean override) {
		this.override = override;
	}
	
	public boolean isOverride(){ 
		return override;
	}

	public int getSchedulerInterval() {
		return schedulerInterval ;
	}

	public void setSchedulerInterval(int interval){
		this.schedulerInterval = interval;
	}
	
	public int getAMECheckupInterval() {
		return AMECheckupInterval ;
	}
	
	public void setAMECheckupInterval(int interval){
		this.AMECheckupInterval = interval;
	}
	
	public int getMessageManagerInterval() {
		return messageManagerInterval ;
	}
	
	public void setMessageManagerInterval(int interval){
		this.messageManagerInterval = interval;
	}

	public int getStateChangeRSTInterval() {
		return stateChangeRSTInterval ;
	}

	public void setStateChangeRSTInterval(int interval){
		this.stateChangeRSTInterval = interval;
	}


	public int getAMCSleepInterval() {
		return AMCSleepInterval;
	}

	public void setAMCSleepInterval(int interval){
		this.AMCSleepInterval = interval;
	}
	
	public int getDBManagerSleepInterval() {
		return DBManagerSleepInterval;
	}

	public void setDBManagerSleepInterval(int interval){
		this.DBManagerSleepInterval = interval;
	}
	
	public int getDbStaticFlushInterval() {
		return DBManagerSleepInterval;
	}

	public void setDbStaticFlushInterval(int interval){
		this.dbStaticFlushInterval = interval;
	}
	
	public int getManagerSleepInterval() {
		return managerSleepInterval;
	}

	public void setManagerSleepInterval(int interval){
		this.managerSleepInterval = interval;
	}

	public int getMaxLogEntries() {
		return maxLogEntries ;
	}
	
	public void setMaxLogEntries(int entries){
		this.maxLogEntries = entries;
	}
	
	public int getMaxHistoricSignals() {
		return maxHistoricSignals ;
	}
	
	public void setMaxHistoricSignals(int entries){
		
		if (entries <= maxHistoricSignalsHARDLIMIT)
			this.maxHistoricSignals = entries;
	}

	public int getMaxNumEntriesPerHistoricSignal() {
		return maxNumEntriesPerHistoricSignal ;
	}
	
	public void setMaxNumEntriesPerHistoricSignal(int entries){

		if (entries <= maxNumEntriesPerHistoricSignalHARDLIMIT)
			this.maxNumEntriesPerHistoricSignal = entries;
	}

	public Integer getPctDiffTrigger() {
		return pctDiffTrigger;
	}

	public void setPctDiffTrigger(Integer pctDiffTrigger) {
		this.pctDiffTrigger = pctDiffTrigger;
	}

	public boolean isSuppressWarnings() {
		return suppressWarnings;
	}
	
	public void setSuppressWarnings(boolean suppressWarnings) {
		this.suppressWarnings = suppressWarnings;
	}

	public boolean isGlobalOverride() {
		return globalOverride;
	}

	public void setGlobalOverride(boolean globalOverride) {
		this.globalOverride = globalOverride;
	}

	public boolean getVisualize() {
		return visualize;
	}
	
	public boolean getIgnoreStateChangeRST() {
		return ignoreStateChangeRST;
	}
	
	public void setVisualize(boolean visualize) {
		this.visualize = visualize;
	}

	public void setIgnoreStateChangeRST(boolean ignoreStateChangeRST) {
		this.ignoreStateChangeRST = ignoreStateChangeRST;
	}

	public int getRSTimeWindowSize() {
		return RSTimeWindowSize;
	}

	public int getTimeSpan() {
		return timeSpan;
	}

	public void setRSTimeWindowSize(int timeWindowSize) {
		RSTimeWindowSize = timeWindowSize;
	}

	public void setTimeSpan(int timeSpan) {
		this.timeSpan = timeSpan;
	}

}
