/**
 * Dasu Project
 *
 * File:         DasuData.java
 * RCS:          $Id$
 * Description:  DasuData class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Feb 21, 2011 at 6:38:55 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;

import edu.northwestern.dasu.util.json.JSONValue;
import edu.northwestern.dasu.util.json.parser.JSONParser;
import edu.northwestern.dasu.util.json.parser.ParseException;

/**
 * Dasu Project
 *
 * File:         DasuData.java
 * RCS:          $Id$
 * Description:  DasuData class (see below)
 * Author:       John Otto
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Feb 21, 2011 at 6:38:55 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

/**
 * @author John Otto &lt;jotto@eecs.northwestern.edu&gt;
 *
 * The DasuData class ...
 * 
 * Provide a key-value store that is synchronized with the client.
 * Keys are always strings, but objects are anything that can be
 * converted to JSON format (i.e. List, Map, primitives)...
 * NOTE: native Java arrays are not supported in the converter.
 * 
 * Objects that implement the DasuDataListener interface can
 * register for callbacks when a key is updated.
 */

public class DasuData {

	private static DasuData self = null;
	
	private DasuView ui = null;
	public static Hashtable<String, String> values = new Hashtable<String, String>();
	public HashMap<String, HashSet<DasuDataListener>> listeners = new HashMap<String, HashSet<DasuDataListener>>();
	public static HashSet<String> persistentKeys = new HashSet<String>();
	
	/*
	 * Private on purpose.
	 * Don't make new instances, just use getInstance().
	 */
	private DasuData() {		
	} 
	
	public static DasuData getInstance() {
		if(self == null) {
			self = new DasuData();
		}
				
		return self;
	}
	
	public void setUi(DasuView dv) {
		ui = dv;
	}
	
	public void updateFromUi(String key, Object val) {
		synchronized(values) {
			updateDb(key, val);
		}
	}
	
	public String dump() {
		synchronized(values) {
			return JSONValue.toJSONString(values);
		}
	}
	
	public Object get(String key) {
		synchronized(values) {
			if(!values.containsKey(key)) return null;
			try {
				return new JSONParser().parse(values.get(key));
			} catch (ParseException e) {
				e.printStackTrace();
			}
			return null;
		}
	}
	
	public void markPersistent(String key) {
		
		//System.out.println("\n\n\n\nDasuData: marking! " + key);
		persistentKeys.add(key);
	}
	
	public void unmarkPersistent(String key) {
		
		if (persistentKeys.contains(key))
			persistentKeys.remove(key);
	}
	
	public void update(String key, Object val) {
		synchronized(values) {
			updateDb(key, val);
			
			// Sync with the remote DB
			
			if(ui != null && ui.isLoaded()) {
				ui.asyncExecJs("storage_updated('" + key + "', " + values.get(key) + ");");
			}
		}
	}

	private void updateDb(String key, Object val) {
		values.put(key, JSONValue.toJSONString(val));
		
		// Run callbacks on updates
		if(listeners.containsKey(key)) {
			for(DasuDataListener ddl: listeners.get(key)) {
				ddl.updated(key, val);
			}
		}
	}
	
	public void flushToUi() {
		if(ui == null || !ui.isLoaded()) {
			return;
		}
		
		synchronized(values) {
			for(String key: values.keySet()) {
				ui.asyncExecJs("storage_updated('" + key + "', " + values.get(key) + ");");
			}
		}
	}
	
	public static interface DasuDataListener {
		public void updated(String key, Object val);
	}
	
	public void registerListener(String key, DasuDataListener ddl) {
		if(!listeners.containsKey(key)) {
			listeners.put(key, new HashSet<DasuDataListener>());
		}
		listeners.get(key).add(ddl);
	}
	
}