/**
 * Dasu Project
 *
 * File:         DasuManager.java
 * RCS:          $Id$
 * Description:  DasuManager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 24, 2010 at 2:27:54 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderConfiguration;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.ResourceFactory;
import org.gudy.azureus2.core3.config.COConfigurationManager;
import org.gudy.azureus2.core3.config.ParameterListener;
import org.gudy.azureus2.core3.logging.Logger;

import edu.northwestern.dasu.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.dasu.download.DownloadMonitor;
import edu.northwestern.dasu.drools.ActiveMeasurementEngine;
import edu.northwestern.dasu.drools.FactDidEvent;
import edu.northwestern.dasu.drools.FactMeasurementLpsd;
import edu.northwestern.dasu.drools.FactMeasurementSystem;
import edu.northwestern.dasu.drools.FactMeasurementUPnP;
import edu.northwestern.dasu.drools.RuleEngine;
import edu.northwestern.dasu.measurement.*;
import edu.northwestern.dasu.stats.*;
import edu.northwestern.dasu.timer.DasuCronScheduler;
import edu.northwestern.dasu.timer.DasuMessageManager;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.ui.BarPlot;
import edu.northwestern.dasu.ui.LinePlot;
import edu.northwestern.dasu.util.DigitalSignatureManager;
import edu.northwestern.dasu.util.PluginInterface;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 * 
 *         The DasuManager class ...
 */
public class DasuManager implements Runnable {

	private final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger(DasuManager.class.getName());
	
	private static boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;
	public static boolean isInitialized = false;
	private int sleepInterval = DasuConfiguration.getInstance().getManagerSleepInterval();
	public static boolean USE_SIMULATION = false;
	private static DasuManager self;

	/** Monitor for added/removed downloads */
	public DownloadMonitor dlm;
	public NetstatBanwidthPerformer nbp;
	public UPnPBanwidthPerformer ubp;
	public ResetMonitor rm;
	
	/* DROOLS ENGINES */
	public static RuleEngine engine = null;
	public static ActiveMeasurementEngine AMEngine = null;
	private static ITimer AMEngineTimer = null;
	//private static ITimer LogMessageTimer = null;
	private static Hashtable<String, ParameterListener> COConfigurationListeners = new Hashtable<String, ParameterListener>();
	
	/* FACTS */
	public static FactMeasurementSystem systemFact = null;
	public static FactMeasurementLpsd localPeerFact = null;
	public static FactMeasurementUPnP localNetworkFact = null;
	
	public static PersistentData persistentDataManager = new PersistentData();
	protected static PluginInterface pi = Main.getPluginInterface();
	
	private static String persistentDataFileName = Main.getDir() + File.separatorChar + "dasu.data." + Main.getDomainName();

	private static final boolean DASU_SKIP_SERVER_SWITCH = false;
	private static TreeMap<Double, String> DasuUrls = new TreeMap<Double, String>();
	public static String currentDasuUrl = DasuConfiguration.DASU_DEFAULT_SERVER;

	/* Pertaining to updating the UI */
	public static Updater updater = null;
	
	public DasuManager() {
		self = this;
		isActive = true;
		
		LOGGER.info("loading persistent data");
		Object pd = Main.loadObject(persistentDataFileName);
		if (pd != null) persistentDataManager = (PersistentData) pd;
		
	}

	public static DasuManager getInstance() {
		return self;
	}
	
	private static void startActiveEngine(String filename, String sigFilename, Integer taskId) {

		if (AMEngine == null) {
			
			LOGGER.info("File Found... Starting ActiveMeasurementEngine...");
			LOGGER.info(filename + "\n" + sigFilename + "\ntaskId: " + taskId + "\n");
			
			try {
				AMEngine = new ActiveMeasurementEngine(filename, sigFilename, taskId);
				AMEngine.initialize();
				AMEngine.setDaemon(true);
				AMEngine.start();
			}
			catch (Exception e) {
				
				LOGGER.severe(e.getMessage());
				e.printStackTrace();
				
			}
		}

	}

	private static void launchActiveMeasrementTimer() {

		int AMECheckupInterval = DasuConfiguration.getInstance()
				.getAMECheckupInterval();
		
		if (Main.DEBUG)
			AMECheckupInterval = 15;

		AMEngineTimer = Main.createTimer("AMEnginePeriodicTimer");
		AMEngineTimer.addPeriodicEvent(AMECheckupInterval * 1000,
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {

						if (Main.DEBUG && AMEngine != null)
							AMEngine.displayFacts();

						if ((AMEngine != null) && (!AMEngine.isActive())) {
							AMEngine = null;
						}

						if (AMEngine == null) {

							ArrayList<String> task = getAMExperiment();
							
							if (task != null) {
								DasuManager.startActiveEngine(task.get(0),
										task.get(1), Integer.parseInt(task.get(2)));
							}
						}
					}

					public ArrayList<String> getAMExperiment() {

						if(DasuManager.getInstance() == null) {
							return null;
						}
						
						String myIp = Main.getPublicIpAddress();
						String url = DasuManager.getInstance().getNewRandomDasuUrl() + DasuConfiguration.AME_SCRIPT + "?ip=" + myIp +
							"&UUIDid=" + DasuConfiguration.getInstance().getUUIDid();
						
						if ((NetworkAdminASNLookupImpl.getInstance() != null) 
								&& (NetworkAdminASNLookupImpl.getInstance().getASNInfo() != null)) {

							url = url +  "&asn=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS() +
								"&cc=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getCountry();
						}
							
						
						String interfFile = "AMExperiment.drl";
						String interfSignature = "AMExperiment.signature";
						String dir = Main.getDir() + java.io.File.separator;

						/* load file from server */
						LOGGER.info("Connecting to " + url);
						
						Properties props = Main.downloadFromURL(url,
								5 * 1000);

						if (props == null) {
							LOGGER.info("AMEngine: No new experiment to load...");
							return null;
						}
						
						String configUrl = null;
						Integer taskId = null;
						Integer signatureId = null;
						String signatureUrl = null;
						boolean isNewFile = false;
						boolean isSigFile = false;

						for (Entry<Object, Object> ent : props.entrySet()) {
							String key = (String) ent.getKey();
							String value = (String) ent.getValue();
							if (key.equals("dasu.active.experiment")) {
								configUrl = value.split(";")[0];
								taskId = Integer
										.valueOf(value.split(";")[1]);
							} else if (key.equals("dasu.active.signature")) {
								signatureUrl = value.split(";")[0];
								signatureId = Integer.valueOf(value
										.split(";")[1]);
							}
						}

						// taskId greater than 0 are not valid active task
						// ids
						if (taskId != null
								&& signatureId != null
								&& taskId > 0
								&& taskId.equals(signatureId)
								&& !getPersistentDataManager()
										.getTaskLogHistory().containsKey(
												taskId)) {
							LOGGER.info("New AM experiment found... downloading");

							isNewFile = Main.downloadFileFromURL(
									configUrl, interfFile, dir);
							isSigFile = Main.downloadFileFromURL(signatureUrl,
									interfSignature, dir);

							if ((isNewFile) && (isSigFile)) {
								
								ArrayList<String> ret = new ArrayList<String>();
								ret.add(dir + interfFile);
								ret.add(dir + interfSignature);
								ret.add(taskId.toString());
								return ret;
								
							}
							else {
								LOGGER.info("Active Measurement file and/or signature file missing!");
							}
						} else {
							LOGGER.info("AMEngine: experiment & signature taskId dont match:... "
												+ " taskId: "
												+ taskId
												+ " signatureId: "
												+ signatureId
												+ " or "
												+ " task already executed... skipping... ");
						}
					
						

					return null;
				}

			});

	}
	
	public static PersistentData getPersistentDataManager() {
		return persistentDataManager;
	}

	private void launchConfigurationChangeListener() {

		final HashMap<String, String> paramsEvent = new HashMap<String, String>();
		paramsEvent.put("Max Download Speed KBs", "MaxDownloadChanged");
		paramsEvent.put("Max Upload Speed KBs", "MaxUploadChanged");
		paramsEvent.put("Max.Peer.Connections.Total", "MaxConnectionsChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent", "MaxConnectionsPerTorrentChanged");
		paramsEvent.put("Max Seeds Per Torrent", "MaxSeedsPerTorrentChanged");
		paramsEvent.put("Use Lazy Bitfield", "LazyBitfieldChanged");
		paramsEvent.put("network.transport.encrypted.require", "EncryptionRequiredChanged");
		paramsEvent.put("network.transport.encrypted.min_level", "EncrytionLevelChanged");
		paramsEvent.put("network.transport.encrypted.fallback.outgoing", "EncryptionOutgoingChanged");
		paramsEvent.put("network.transport.encrypted.fallback.incoming", "EncryptionIncominChanged");
		paramsEvent.put("network.transport.encrypted.use.crypto.port", "EncryptionCryptoPortChanged");
		paramsEvent.put("TCP.Listen.Port", "TCPPortChanged");
		paramsEvent.put("UDP.Listen.Port", "UDPPortChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent.When.Seeding.Enable", "MaxConnectionsPerTorrentWhenSeedingChanged");
		paramsEvent.put("enable.seedingonly.maxuploads", "MaxUploadWhenSeedingChanged");
		
		for (String param : paramsEvent.keySet()) {
			
			ParameterListener pl = new ParameterListener() {
				public void parameterChanged(String parameterName) {
					LOGGER.fine("CHANGED! " + parameterName);
					AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDidEvent(paramsEvent.get(parameterName), -1L));
				}
			};
			
			COConfigurationManager.addParameterListener(param, pl);
			COConfigurationListeners.put(param, pl);
			
		}		

	}
	
	private void printDebug() {
	
		if ((localPeerFact != null) && (systemFact != null)) {

			LOGGER.fine("\n" + "local lpsd: avgUploadRate: "
					+ (localPeerFact.getBytesSent()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (localPeerFact.getBytesReceived()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (localPeerFact.getShortDeltaTime()));
			
			LOGGER.fine("local netstat: avgUploadRate: "
					+ (systemFact.getBytesSent()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (systemFact.getBytesReceived()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (systemFact.getShortDeltaTime()));
			
			LOGGER.fine("local upnpstat: avgUploadRate: "
					+ (localNetworkFact.getTotalBytesSent()/(localNetworkFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (localNetworkFact.getTotalBytesReceived()/(localNetworkFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (localNetworkFact.getShortDeltaTime()));


			LOGGER.fine("local lpsd: uploadRate: "
					+ localPeerFact.getUploadRate()
					+ " avgUpload: "
					+ localPeerFact.getAverageUpload()
					+ " stdevUpload: "
					+ localPeerFact.getStdevUpload()
					+ " downloadRate: "
					+ localPeerFact.getDownloadRate()
					+ " avgDownload: "
					+ localPeerFact.getAverageDownload()
					+ " stdevDownload: "
					+ localPeerFact.getStdevDownload()
					+ " deltaTime: "
					+ localPeerFact.getLongDeltaTime());
			
			LOGGER.fine("local netstat: bytesSent: "
					+ systemFact.getBytesSent()
					+ " avgBytesSent: "
					+ systemFact.getAverageBytesSent()
					+ " stdevBytesSent: "
					+ systemFact.getStdevBytesSent()
					+ " bytesReceived: "
					+ systemFact.getBytesReceived()
					+ " avgBytesReceived: "
					+ systemFact.getAverageBytesReceived()
					+ " stdevBytesReceived: "
					+ systemFact.getStdevBytesReceived()
					+ " deltaTime: "
					+ systemFact.getLongDeltaTime());
			
			LOGGER.fine("local lpsd: numTorrents: "
					+ localPeerFact.getNumTorrents()
					+ " numConnctedSeeds: "
					+ localPeerFact.getNumConnectedSeeds()
					+ " avgNumConnectedSeeds: "
					+ localPeerFact.getAverageNumConnectedSeeds()
					+ " numConnctedLeechers: "
					+ localPeerFact.getNumConnectedLeechers()
					+ " avgNumConnectedLeechers: "
					+ localPeerFact.getAverageNumConnectedLeechers()
					+ " maxPeerConnections: "
					+ localPeerFact.getMaxPeerConnectionsGlobal()
					+ " maxPeerConnectionsPerTorrent: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrent()
					+ " maxPeerConnectionsPerTorrentSeeding: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrentSeeding()
					+ " maxUploadRateSeeding: "
					+ localPeerFact.getMaxUploadRateSeeding()
					+ " maxSeedsPerTorrent: "
					+ localPeerFact.getMaxSeedsPerTorrent()
					+ " lazyBitfield: "
					+ localPeerFact.isLazyBitfield()
					+ " requireEncryption: "
					+ localPeerFact.isRequireEncryption()
					+ " encryptionMinLevel: "
					+ localPeerFact.getEncryptionMinLevel()
					+ " allowNonEncryptedOut: "
					+ localPeerFact.isAllowNonEncryptedOut()
					+ " allowNonEncryptedIn: "
					+ localPeerFact.isAllowNonEncryptedIn()
					+ " useCryptoPortExtension: "
					+ localPeerFact.isUseCryptoPortExtension()
					+ " tcpPortIn: " 
					+ localPeerFact.getTcpPortIn()
					+ " udpPortIn: "
					+ localPeerFact.getUdpPortIn());
		}
		
		dlm.printDebug();		

		if (engine != null)
			engine.displayFacts();
		
	}
	
	public void testing() {
		
//		List<String> urls = Arrays.asList(new String[]{"http://google.com"});		
//		Map<String, String> headers = new HashMap();
//		headers.put("User-agent", "Mozzilla");		
//		Map<String, Object> params = new HashMap();		
//		params.put("headers", new HashMap());
//		params.put("headers", headers);		
//		ProbeTask task = new ProbeTask(-1, -1, 0L, 0L, urls, ProbeType.HTTP, TaskType.BULK, SourceType.RE, params);		
//		AMCoordinatorV2.addProbeTask(task);
		
//		AMCoordinatorV2.getInstance().addPriorityProbe("4.4.4.2", ProbeType.DNS);
		
//		Statistics.getInstance().addDnsTimeout(AMCoordinatorV2.getMyDnsServers().get(0).get(0), 
//				Main.getPublicIpAddress(), System.currentTimeMillis(), System.currentTimeMillis(), true);
		
		//AMCoordinatorV2.addPriorityProbe(null, ProbeType.NDT2);
		LOGGER.info("Azureus name: " + Main.getPluginInterface().getAzureusName());
		LOGGER.info(Main.getPluginInterface().getAzureusVersion());
		LOGGER.info(Main.getPluginInterface().getApplicationName());
		LOGGER.info(System.getProperty("os.name"));
		LOGGER.info(System.getProperty("os.arch"));
		LOGGER.info(System.getProperty("os.version"));
		LOGGER.info(System.getProperty("java.version"));
		LOGGER.info(System.getProperty("java.vendor"));
		LOGGER.info(Main.getInterfaceName());
		LOGGER.info(Main.getInterfaceMAC());
		
		LOGGER.info("Disabling AME experiments due to testing mode...");
		DasuConfiguration.getInstance().setRunExperiments(false);
		
	//    	if (!AMCoordinatorV2.historicSampleExists("time_of_day_bw_limit_day_minute_5")) {
	//		AMCoordinatorV2.historicSampleCreate("time_of_day_bw_limit_day_minute_5", TimeType.DAY, TimeType.MINUTE, 5);
	//
	//    ArrayList maxDownloadRateLpsdPastTwoHours = (ArrayList) AMCoordinatorV2.historicSampleGetValuesSubsetPast(
	//            "time_of_day_bw_limit_day_minute_5", "maxDownloadRateLpsd", Util.currentDasuTime(), 60);
	//        System.out.println("\n\n\n\n\n\nENGINE: " + maxDownloadRateLpsdPastTwoHours);
		
	//	SignalSample ss = null;
	//	try {
	//		ss = new SignalSample("uploadRate", TimeType.DAY, TimeType.HOUR, 1, "BytesPerSecond");
	//	} catch (Exception e1) {
	//		e1.printStackTrace();
	//	}
	//	System.out.println(new Date(Util.currentDasuTime()));
	//	ss.set(Util.currentDasuTime(), 4532L);
	//	
	//	System.out.println(ss.getName());
	//	System.out.println(ss.getValuesHash());
	//	System.out.println(ss.getTimeStampsHash());
		
	//	Collection<Float> nt = AMCoordinatorV2.historicSampleGetValues("netstat_max_downloadRate_hour_minute_1");
	//	Collection<Float> bt  = AMCoordinatorV2.historicSampleGetValues("bt_max_downloadRate_hour_minute_1");
	//	System.out.println(SignalManipulation.evalAverage(new Float[]{1f, 2f, 3f}, 1));
		
	//	System.out.println(nt);
	//	System.out.println(bt);
	//	

		Float[] nt = new Float[]{77.0f, 65.0f, 55.0f, 58.0f, 52.0f, 52.0f, 76.0f, 61.0f, 58.0f, 59.0f, 
				67.0f, 62.0f, 67.0f, 76.0f, 77.0f, 65.0f, 55.0f, 58.0f, 51.0f, 52.0f, 73.0f, 66.0f, 78.0f, 57.0f, 
				67.0f, 72.0f, 52.0f, 76.0f, 58.0f, 59.0f, 67.0f, 62.0f, 67.0f, 76.0f, 77.0f, 65.0f, 55.0f, 
				58.0f, 51.0f, 52.0f, 73.0f, 66.0f, 78.0f, 57.0f, 67.0f, 72.0f, 52.0f, 76.0f, 61.0f, 58.0f, 59.0f, 
				67.0f, 62.0f, 67.0f, 76f, 73.0f, 66.0f, 78.0f, 57.0f};
		
		Float[] bt = new Float[]{
				40.0f, 40.0f, 43.0f, 41.0f, 44.0f, 50.0f, 55.0f, 42.0f, 12.0f, 4.0f, 10.0f, 12.0f, 11.0f, 14.0f, 
				44.0f, 50.0f, 53.0f, 48.0f, 40.0f, 53.0f, 41.0f, 44.0f, 42.0f, 50.0f, 55.0f, 42.0f, 59.0f,
				44.0f, 50.0f, 66.0f, 51.0f, 40.0f, 43.0f, 41.0f, 44.0f, 42.0f, 50.0f, 55.0f, 42.0f, 59.0f, 44.0f, 50.0f, 
				66.0f, 59.0f, 66.0f, 44.0f, 50.0f, 53.0f, 48.0f, 40.0f, 66.0f, 44.0f, 50.0f, 53.0f, 48.0f, 53.0f,
				42.0f, 50.0f, 55.0f};
	
		Float[] signal = new Float[]{0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, .0f, 24.0f, 34f, 13f, 44f, 53f}; 
	
		Long[] time = new Long[bt.length];
		for (int i=0; i < bt.length; i++ )
			time[i] = Util.currentDasuTime()+i*1000*60;
	
		LOGGER.info("signal: " + Arrays.asList(signal));
		LOGGER.info("getMean: " + SignalManipulation.getAverage(signal));
		LOGGER.info("getStdev: " + SignalManipulation.getStdDev(signal));
		LOGGER.info("getPercentile: " + SignalManipulation.getPercentile(signal, 95));
		LOGGER.info("getPercentileAboveValue: " + SignalManipulation.getPercentileAboveValue(signal, 95, SignalManipulation.getAverage(signal)));
		LOGGER.info("getPercentileIndexPosition: " + SignalManipulation.getPercentileIndexPosition(signal, 95));
		
		MovingWindow mw =  new MovingWindow(bt, time, 5);
		LOGGER.info(mw.toString());	
		
		LOGGER.info("correlation 5: " + SignalManipulation.getCorrelationCollection(nt, bt, 5));
		LOGGER.info("correlation 5 cumulative: " + SignalManipulation.getCorrelationCollection(
				SignalManipulation.getCumulative(nt), SignalManipulation.getCumulative(bt), 5));
		
		LOGGER.info("time: (" + time.length + ") " + Arrays.asList(time));
		LOGGER.info("cumulative bt: (" + SignalManipulation.getCumulative(bt).size() + ") " + SignalManipulation.getCumulative(bt));
	
		LOGGER.info("eval Slope 1: " + SignalManipulation.getSlopeCollection( 
				SignalManipulation.getCumulative(bt), Arrays.asList(time), 2));
		
		LinePlot plot1 = new LinePlot("bt.lpsd.plot.test","test1",  mw.getTime(), mw.getMean(), "(time)", "(kbps)", LinePlot.FORMAT_NUMBER);
		DasuData.getInstance().update(plot1.getId(), plot1.generatePlot());		
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot1.generatePlot());
		
		LinePlot plot2 = new LinePlot("bt.lpsd.plot.test2", "test2", null, mw.getSignal(), "(time)", "(kbps)", LinePlot.FORMAT_NUMBER);
		DasuData.getInstance().update(plot2.getId(), plot2.generatePlot());
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot2.generatePlot());

		LinePlot plot3 = new LinePlot("bt.lpsd.plot.test3","test3", null, mw.getMean(), "(time)", "(kbps)", LinePlot.FORMAT_NUMBER);
		plot3.addSeries("title", mw.getStdDev());
		DasuData.getInstance().update(plot3.getId(), plot3.generatePlot());
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot3.generatePlot());
		
		BarPlot plot4 = new BarPlot("bt.lpsd.plot.test4", "test", new String[]{"Sun", "Mon", "Tue", "Wed", "Thr", "Fri", "Sat"}, new Float[]{6f, 5f, 4f, 3f, 2f, 1f, 0f}, "(test)", "(test)");
		plot4.addOption("horizontal", true);
		DasuData.getInstance().update(plot4.getId(), plot4.generatePlot());		
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot4.generatePlot());
		
		BarPlot plot5 = new BarPlot("bt.lpsd.plot.test5", "test", new String[]{"Sun", "Mon", "Tue", "Wed", "Thr", "Fri", "Sat"}, new Float[]{6f, 5f, 4f, 3f, 2f, 1f, 0f}, "(test)", "(test)");
		plot5.addSeries("test2", new Float[]{6f, 5f, 6f, 5f, 6f, 5f, 6f});
		DasuData.getInstance().update(plot5.getId(), plot5.generatePlot());		
		LOGGER.info("\n\n\n\n\n\nMARIO: " + plot5.generatePlot());
		
	}
	
	/**
	 * This should only run if DEBUG is not set.
	 * Follows the same scheme that the RuleEngine uses to update the inference rules.
	 * 
	 * @author JSO
	 *
	 * The Updater class ...
	 */
	public class Updater extends Thread {
		
		private final String zipPrefix = "DasuUiRules_";
		private final String zipSuffix = ".zip";
		private final String zipSigSuffix = ".signature";
		
		private final String zipVersionFile = "VERSION";
		private final String zipRulesPrefix = "rules/";
		private final String zipRulesSuffix = ".drl";
		private final String zipRulesBaseName = "base";
		private final String zipUiPrefix = "ui/";
		
		// XXX: This must be set to the version that is packed in the Jar!
		private final Float zipDefaultVersion = 1.0F;
		
		private Float toUse = null;
		private boolean ready = false;
		
		private String latestZipUrl = null;
		private Float latestZipVersion = null;
		private String latestSignatureUrl = null;
		
		private final String updateServerZipKey = "dasu.inference.config";
		private final String updateServerSigKey = "dasu.inference.signature";
		
		public boolean canUpdate() {
			boolean checkSucceeded = checkLatestVersion();
			return checkSucceeded && (toUse == null || latestZipVersion > toUse);
		}
		
		public boolean isReady() {
			return ready;
		}
		
		public boolean haveValidFile() {
			return toUse != null;
		}
		
		public Float getVersion() {
			return toUse;
		}

		/*
		 * return success of check
		 */
		private boolean checkLatestVersion() {
			String url = DasuManager.getInstance().getNewRandomDasuUrl() + DasuConfiguration.CONFIG_LIST;
			Properties props = Main.downloadFromURL(url,  2*1000);
			
			if(props == null) return false; // Return failure
			
			for (Entry<Object, Object> ent : props.entrySet()){				
				String key = (String)ent.getKey();
				String value = (String)ent.getValue();
				if (key.equals(updateServerZipKey)){
					latestZipUrl = value.split(";")[0];
					latestZipVersion = Float.valueOf(value.split(";")[1]);
				} else if (key.equals(updateServerSigKey)) {
					latestSignatureUrl = value.split(";")[0];
				}
			}
			boolean result = latestZipUrl != null && latestZipVersion != null && latestSignatureUrl != null;
			LOGGER.info("checkLatestVersion returning " + result + " with found version " + latestZipVersion);
			return result;
		}
		
		/*
		 * Return a list of version numbers for which we have a zip and sig file, in descending order.
		 */
		private ArrayList<Float> getLocalVersions() {
			HashMap<Float, File> sigs = new HashMap<Float, File>();
			HashMap<Float, File> zips = new HashMap<Float, File>();
			
			File dir = new File(Main.getDir());
			for(File f: dir.listFiles()) {
				String name = f.getName();
				if(!name.startsWith(zipPrefix)) continue;
				if(!name.endsWith(zipSuffix) && !name.endsWith(zipSuffix + zipSigSuffix)) continue;
				
				Float version = null;
				int toCrop = 0;
				boolean sig = name.endsWith(zipSigSuffix);
				
				if(sig) {
					// sig file
					toCrop = zipSuffix.length() + zipSigSuffix.length();
				} else {
					// zip file
					toCrop = zipSuffix.length();
				}
				
				try {
					version = Float.parseFloat(name.subSequence(zipPrefix.length(), name.length() - toCrop).toString());
				} catch(NumberFormatException e) {}
				
				if(version != null) {
					if(sig) {
						sigs.put(version, f);
					} else {
						zips.put(version, f);
					}
				}
				
			}
			Set<Float> lhs = sigs.keySet();
			lhs.retainAll(zips.keySet());
			ArrayList<Float> versions = new ArrayList<Float>(lhs);
			Collections.sort(versions);
			Collections.reverse(versions);
			LOGGER.info("getLocalVersions found these versions: " + versions);
			return versions;
		}
		
		private String getZipName(Float version) {
			return zipPrefix + version.toString() + zipSuffix;
		}
		
		private String getSigName(Float version) {
			return zipPrefix + version.toString() + zipSuffix + zipSigSuffix;
		}
		
		private String getZipPath(Float version) {
			return Main.getDir() + File.separator + getZipName(version);
		}
		
		private String getSigPath(Float version) {
			return Main.getDir() + File.separator + getSigName(version);
		}
		
		private boolean isVerified(Float version) {
			boolean result = DigitalSignatureManager.verify(getZipPath(version), getSigPath(version));
			LOGGER.info("checking if signature is valid for version " + version + " -- " + result + "!");
			return result;
		}
		
		@SuppressWarnings("unchecked")
		private boolean versionMatches(Float version) {
			ZipFile zf;
			boolean result = false;
			try {
				zf = new ZipFile(getZipPath(version));
				Enumeration e = zf.entries();
				ZipEntry ze = null;
				while(e.hasMoreElements()) {
					ZipEntry obj = (ZipEntry) e.nextElement();
					if(obj.getName().equals(zipVersionFile)) {
						ze = obj;
						break;
					}
				}
				if(ze == null) {
					LOGGER.info("no " + zipVersionFile + " in " + getZipPath(version));
					zf.close();
				} else {
					BufferedReader br = new BufferedReader(new InputStreamReader(zf.getInputStream(ze)));
					String line = br.readLine().trim();
					br.close();
					zf.close();
					Float thisVersion = Float.parseFloat(line);
					result = thisVersion.equals(version);
				}
			} catch (IOException e) {
				//e.printStackTrace();
			} catch (NumberFormatException e) {
				//e.printStackTrace();
			}
			LOGGER.info("checking if version matches for version " + version + " -- " + result + "!");
			return result;
		}
		
		@SuppressWarnings("unchecked")
		private boolean rulesBuild(Float version) {
			boolean result = false;
			try {
				ZipFile zf = new ZipFile(getZipPath(version));
				
				Properties props = new Properties();
		        props.setProperty("drools.dialect.java.compiler", "JANINO");
		        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, (ClassLoader[]) null);
				KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
				
				// load the base rule file
				boolean found = false;
				Enumeration e = zf.entries();
				while(e.hasMoreElements()) {
					ZipEntry obj = (ZipEntry) e.nextElement();
					if(obj.getName().equals(zipRulesPrefix + zipRulesBaseName + zipRulesSuffix)) {
						found = true;
						
						// try to build this file
						try {
							InputStream is = zf.getInputStream(obj);
							kbuilder.add(ResourceFactory.newInputStreamResource(is), ResourceType.DRL);
							is.close();
							if (kbuilder.getErrors().size() > 0) {
								throw new IllegalArgumentException("Could not parse knowledge.");
							}
							result = true;
						} catch(Exception e2) {
							LOGGER.info("exception while building " + zipRulesBaseName + " in " + getZipPath(version));
							e2.printStackTrace();
						}
						break;
					}
				}
				
				if(!found) {
					LOGGER.info("did not find base rule file " + zipRulesPrefix + zipRulesBaseName + zipRulesSuffix + " in " + getZipPath(version));
				}
				if (found && result) {
					// Make sure all other files build, too.
					boolean okay = true;
					
					e = zf.entries();
					while(e.hasMoreElements()) {
						ZipEntry obj = (ZipEntry) e.nextElement();
						if(obj.getName().startsWith(zipRulesPrefix) && obj.getName().endsWith(zipRulesSuffix) && !obj.getName().equals(zipRulesPrefix + zipRulesBaseName + zipRulesSuffix)) {
							
							// try to build this file
							try {
								InputStream is = zf.getInputStream(obj);
								kbuilder.add(ResourceFactory.newInputStreamResource(is), ResourceType.DRL);
								is.close();
								if (kbuilder.getErrors().size() > 0) {
									throw new IllegalArgumentException("Could not parse knowledge.");
								}
							} catch(Exception e2) {
								okay = false;
								LOGGER.info("exception while building " + obj.getName() + " in " + getZipPath(version));
								break;
							}
						}
					}
					
					if(!okay) result = false;
				}
				zf.close();
			} catch (IOException e) {
				//e.printStackTrace();
			} catch (NumberFormatException e) {
				//e.printStackTrace();
			}
			LOGGER.info("checking if rules build for version " + version + " -- " + result + "!");
			return result;
		}
		
		private void delete(Float version) {
			File zf = new File(getZipPath(version));
			if(zf.exists() && zf.canWrite()) zf.delete(); 
			File sf = new File(getSigPath(version));
			if(sf.exists() && sf.canWrite()) sf.delete();
		}
		
		private boolean extractFromJar(String fileToExtract) throws IOException {
			ClassLoader classLoader = Main.getPluginInterface().getPluginClassLoader();

			if (!(classLoader instanceof URLClassLoader)) {
				LOGGER.info("classloader not instanceof URL classloader");
				return false;
			}
			
			URLClassLoader current = (URLClassLoader) classLoader;

			URL url = current.findResource(fileToExtract);
			if(url == null) {
				LOGGER.info("can't find file " + fileToExtract);
				return false;
			}
			
			DataInputStream in = new DataInputStream(url.openStream());	        
		    byte[] bytes = new byte[in.available()];
		  
		    int offset = 0;
		    int numRead = 0;
		    while (offset < bytes.length
		             && (numRead=in.read(bytes, offset, bytes.length-offset)) >= 0) {
		          offset += numRead;
		    }
	        
	        String filename = Main.getDir() + java.io.File.separator + fileToExtract;

		    File file=new File(filename);
		    DataOutputStream outs = new DataOutputStream(new FileOutputStream(file,false));
		    outs.write(bytes);
		    outs.close();
		    
			return true;
		}
		
		private Float getBestWorkingLocalVersion() {
			Float bestWorkingLocalVersion = null;
			for(Float version: getLocalVersions()) {
				if(isVerified(version) && versionMatches(version) && rulesBuild(version) && bestWorkingLocalVersion == null) {
					bestWorkingLocalVersion = version;
					LOGGER.info("testing version " + version + " with path " + getZipPath(version) + " -- OK!");
				} else {
					// Delete versions that fail or when we already have a newer working version.
					LOGGER.info("testing version " + version + " with path " + getZipPath(version) + " -- failed or old. Deleting!");
					delete(version);
				}
			}
			return bestWorkingLocalVersion;
		}
		
		@SuppressWarnings("unchecked")
		public HashMap<String, ByteArrayOutputStream> getUiData() {
			HashMap<String, ByteArrayOutputStream> data = new HashMap<String, ByteArrayOutputStream>();
			
			if(toUse != null) {
				try {
					ZipFile zf = new ZipFile(getZipPath(toUse));
					Enumeration e = zf.entries();
					while(e.hasMoreElements()) {
						ZipEntry ze = (ZipEntry) e.nextElement();
						if(!ze.getName().startsWith(zipUiPrefix)) continue;
						
						// get the input stream, load the byte array
						InputStream is = zf.getInputStream(ze);
						ByteArrayOutputStream baos = new ByteArrayOutputStream();
						int nRead;
						byte[] arr = new byte[4096];
						while((nRead = is.read(arr, 0, arr.length)) != -1) {
							baos.write(arr, 0, nRead);
						}
						is.close();
						data.put(ze.getName().substring(zipUiPrefix.length()), baos);
					}
					zf.close();
				} catch (IOException e) {
					e.printStackTrace();
				} catch (NumberFormatException e) {
					e.printStackTrace();
				}
			}
			return data;
		}
		
		@SuppressWarnings("unchecked")
		public Vector<ByteArrayOutputStream> getRulesData() {
			Vector<ByteArrayOutputStream> fileData = new Vector<ByteArrayOutputStream>();
			if(toUse != null) {
				try {
					ZipFile zf = new ZipFile(getZipPath(toUse));
					Enumeration e = zf.entries();
					while(e.hasMoreElements()) {
						ZipEntry obj = (ZipEntry) e.nextElement();
						if(obj.getName().equals(zipRulesPrefix + zipRulesBaseName + zipRulesSuffix)) {
							// base rule file -- load first.
							InputStream is = zf.getInputStream(obj);
							ByteArrayOutputStream baos = new ByteArrayOutputStream();
							int nRead;
							byte[] data = new byte[16384];
							while((nRead = is.read(data, 0, data.length)) != -1) {
								baos.write(data, 0, nRead);
							}
							
							is.close();
							fileData.add(baos);
							break;
						}
					}
					
					e = zf.entries();
					while(e.hasMoreElements()) {
						ZipEntry obj = (ZipEntry) e.nextElement();
						if(obj.getName().startsWith(zipRulesPrefix) && obj.getName().endsWith(zipRulesSuffix) && !obj.getName().equals(zipRulesPrefix + zipRulesBaseName + zipRulesSuffix)) {
							// additional rule files -- load in any order.
							InputStream is = zf.getInputStream(obj);
							ByteArrayOutputStream baos = new ByteArrayOutputStream();
							int nRead;
							byte[] data = new byte[16384];
							while((nRead = is.read(data, 0, data.length)) != -1) {
								baos.write(data, 0, nRead);
							}
							
							is.close();
							fileData.add(baos);
						}
					}					
					
					zf.close();
				} catch (IOException e) {
					e.printStackTrace();
				} catch (NumberFormatException e) {
					e.printStackTrace();
				}
			}
			return fileData;
		}
		
		public void run() {
			LOGGER.info("Updater started");
			
			// Get the latest local version that works.
			Float bestWorkingLocalVersion = getBestWorkingLocalVersion();
			LOGGER.info("at start, bestWorkingVersion = " + bestWorkingLocalVersion);
			
			// What is the latest version, as reported by the server?
			boolean checkSucceeded = checkLatestVersion();
			
			if(checkSucceeded && (bestWorkingLocalVersion == null || latestZipVersion > bestWorkingLocalVersion)) {
				LOGGER.info("trying to update");
				
				// Is the newest version (remote) newer than the best available local version?
				// Or, do we not have a local working version?
				// Then...download the newest files
				boolean zipSuccess = Main.downloadFileFromURL(latestZipUrl, getZipPath(latestZipVersion));
				boolean sigSuccess = Main.downloadFileFromURL(latestSignatureUrl, getSigPath(latestZipVersion));
				
				if(!zipSuccess || !sigSuccess) {
					LOGGER.info("update download failed");
					// update failed. delete one or more of the files.
					delete(latestZipVersion);
				} else {
					// Re-update to see what's working
					bestWorkingLocalVersion = getBestWorkingLocalVersion();
					LOGGER.info("update download succeeded. new bestWorkingVersion = " + bestWorkingLocalVersion);
				}
			}
			
			if(bestWorkingLocalVersion == null) {
				LOGGER.info("after update, bestWorkingVersion is null. trying to extract from jar.");
				// after update, still don't have a working version.
				// extract from the jar
				try {
					boolean zipSuccess = extractFromJar(getZipName(zipDefaultVersion));
					boolean sigSuccess = extractFromJar(getSigName(zipDefaultVersion));
					
					if(!zipSuccess || !sigSuccess) {
						LOGGER.severe("extract from jar failed.");
						LOGGER.severe("using zipName " + getZipName(zipDefaultVersion));
						LOGGER.severe("using sigName " + getSigName(zipDefaultVersion));
						delete(zipDefaultVersion);
					} else {
						LOGGER.info("extract from jar succeeded.");
					}
					
					// Get the best working version, again.
					bestWorkingLocalVersion = getBestWorkingLocalVersion();
					LOGGER.info("after extract from jar, bestWorkingLocalVersion = " + bestWorkingLocalVersion);
				} catch(IOException e) {
					LOGGER.severe("extract from jar failed");
					e.printStackTrace();
					delete(zipDefaultVersion);
				}
			}
			
			if(bestWorkingLocalVersion == null) {
				// Even version from the jar must be busted. Whoa.
				LOGGER.severe("all attempts to load a UI/rules version failed");
			} else {
				toUse = bestWorkingLocalVersion;
				LOGGER.info("using UI version " + toUse);
			}
			
			// Set the version being used...
			DasuConfiguration.getInstance().setInferenceVersion(toUse);
			Main.updateProperties();
			
			ready = true;
			LOGGER.info("UI update finished; ready");
		}
	}
	
	/**
	 * Perform periodic data-collections
	 */
	
	public void run() {

		// Initialize
		
		if (!Logger.isEnabled())
			COConfigurationManager.setParameter("Logger.Enabled", true);

		Iterator<String> iter = persistentDataManager.UIValues.keySet().iterator();
		synchronized(DasuData.values) {
			while (iter.hasNext()) {
				String key = iter.next();
				DasuData.values.put(key, persistentDataManager.UIValues.get(key));
				iter.remove();
			}
		}
		
		// Has to be before AMC does init
		nbp = new NetstatBanwidthPerformer(Main.getInterfaceName(), "Netstat BW");
		nbp.runOnce(1000);
		
		ubp = new UPnPBanwidthPerformer(Main.getInterfaceName(), "UPnP BW");
		ubp.runOnce(1000);
	
		getDasuServerList();
		
		// Start updating the UI and rules
		updater = new Updater();
		updater.run(); // just inline this :) don't run as a thread.
		
		if(!updater.isReady() || !updater.haveValidFile()) {
			LOGGER.severe("Updater failed to find a working set of rules/UI.");
		}
		
		pi.getUtilities().createThread("DasuAMCoordinator", AMCoordinatorV2.getInstance());
		
		LOGGER.info("initializing facts");
		systemFact = new FactMeasurementSystem(new NetstatSystemStats());
		localPeerFact = new FactMeasurementLpsd(new DynamicLocalPeerStat());
		localNetworkFact = new FactMeasurementUPnP(new UPnPSystemStats());
		
		LOGGER.info("starting rule engine");
		try {
			engine = new RuleEngine();
			engine.setDaemon(true);
			engine.initialize();

			if (!engine.isAlive() && engine.isInitialized()) engine.start();
		}
		catch(Exception e) {
			e.printStackTrace();
			// TODO: If we get here, we couldn't load any rules. Make this stop the startup process...
		}
		LOGGER.info("started rule engine");
		
		pi.getUtilities().createThread("DasuCronScheduler", DasuCronScheduler.getInstance());
		pi.getUtilities().createThread("DasuMessageManager", DasuMessageManager.getInstance());
		
		if (DasuConfiguration.getInstance().isRunExperiments())
			launchActiveMeasrementTimer();
		
		//launchLogMessageTimer();
		isInitialized = true;
		
		Main.logPlugin("Inference engine enabled...");

		//testing();
		
		if (!USE_SIMULATION) {
			dlm = new DownloadMonitor(engine);
			
			rm = new ResetMonitor(dlm);
			rm.runOnce("Netstat Reset Monitor", 1000);
			
			launchConfigurationChangeListener();
			
		} else {
			localPeerFact = null;
			systemFact = null;
			localNetworkFact = null;
		}
		
		while (isActive) {

			if (USE_SIMULATION) {
				performSimulation();
			} else {
				
				/*
				 * update measurement facts associated with the collected
				 * variables
				 */
				
				rm.update();
				nbp.update();
				ubp.update();

				if (engine != null) {

					//Rpss facts are inserted by the PeerStats Manager 
					//asynchronously as peers disconnect, only for peers
					//that actually transmitted/received something from us
					
					dlm.updateFacts();
					
					systemFact.addEntry(new NetstatSystemStats());
					localPeerFact.addEntry(new DynamicLocalPeerStat());
					localNetworkFact.addEntry(new UPnPSystemStats());

					engine.addFact(systemFact);
					engine.addFact(localPeerFact);
					engine.addFact(localNetworkFact);

				}

				if (Main.DEBUG)
					printDebug();

			}// else

			LOGGER.info("DasuManager: sleeping for " + sleepInterval + " seconds...");

			synchronized (halt) {

				try {
					if (!terminate) {
						if (USE_SIMULATION)
							Thread.sleep(2);
						else
							halt.wait(sleepInterval * 1000);
					}
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			}

			if (terminate) {
				self = null;
				return;
			}

		}// while

	}
	
	public boolean isRunning() {
		return isActive && !terminate;
	}

	private boolean performSimulation() {

		if (!InterferenceSimulator.isActive()) {
			if (!InterferenceSimulator.loadTimeline())
				return false;
			InterferenceSimulator.printEvents();
		}

		
		//why do you need this?
		// JSO: don't think we need to update the reset monitor specifically for simulation mode...
		/*if (edu.northwestern.dasu.Main.perf != null) {
			edu.northwestern.dasu.Main.perf.perform(new UTTimerEvent() {
				public void cancel() {
				}
			});
		}*/

		/*
		 * collect all sample variables first to minimize time discrepancy
		 * between samples
		 */

		HashMap<Long, DownloadStatsDynamic> didTempMap = InterferenceSimulator
				.getModifiedDownloadStatsDynamicMap();

		HashMap<Long, RSTStat> didRstMap = InterferenceSimulator
				.getModifiedRSTStatMap();

		NetstatSystemStats systemRst = InterferenceSimulator.getModifiedNetstatSystemStats();
		
		DynamicLocalPeerStat lpsd = InterferenceSimulator.getModifiedDynamicLocalPeerStat();
		
		for (Entry<Long, RSTStat> e : didRstMap.entrySet()) {

			LOGGER.info("performSimulation(): did: "
					+ e.getValue().downloadId + " numRst: "
					+ e.getValue().numRst + " time: "
					+ e.getValue().timeRecorded);
		}

		/*
		 * update/create measurement facts associated with the collected
		 * variables
		 */

		if (systemFact == null) {
			systemFact = new FactMeasurementSystem(systemRst);
			if (engine != null)
				engine.addFact(systemFact);
		} else
			systemFact.addEntry(systemRst);

		if (localPeerFact == null) {
			localPeerFact = new FactMeasurementLpsd(lpsd);
			if (engine != null)
				engine.addFact(localPeerFact);
		} else
			localPeerFact.addEntry(lpsd);

		// TODO: JSO: commented out simulation functionality that has been moved to DownloadMonitor
		/*
		
		for (long did : didTempMap.keySet()) {

			if (!didFactMap.containsKey(did)) {

				didFactMap
						.put(did, new FactMeasurementDsd(didTempMap.get(did)));
				if (engine != null)
					engine.addFact(didFactMap.get(did));
			} else
				didFactMap.get(did).addEntry(didTempMap.get(did));
		}

		for (long did : didRstMap.keySet()) {

			if (!rstFactMap.containsKey(did))
				rstFactMap.put(did, new FactMeasurementRst(didRstMap.get(did)));
			else
				rstFactMap.get(did).addEntry(didRstMap.get(did));

			if (engine != null) {
				engine.addFact(rstFactMap.get(did));
				//System.out.println("\n\n\n\nENGINE: " + engine + "\n\n\n\n");
			}

			// engine.updateFact(rstFactMap.get(did));

			// System.out.println(rstFactMap);

		}

		Iterator<Entry<Long, FactMeasurementRst>> iter = rstFactMap.entrySet()
				.iterator();
		while (iter.hasNext()) {
			Entry<Long, FactMeasurementRst> e = iter.next();
			e.getValue().updateEntryList();
			// if (e.getValue().updateEntryList() == 0) {
			// engine.retractFact(e.getValue());
			// iter.remove();
			// }
		}
		
		*/

		if (Main.DEBUG)
			printDebug();

		InterferenceSimulator.increaseTick();

		return true;

	}

	public void stop() {

		isActive = false;
		synchronized (halt) {
			terminate = true;
		}

		for ( String key : DasuData.persistentKeys ) {
			if ( DasuData.values.containsKey(key))
				persistentDataManager.UIValues.put(key, DasuData.values.get(key));
		}		
		Main.saveObject(persistentDataManager, persistentDataFileName);
		
		if (AMEngineTimer != null) AMEngineTimer.destroy();
		DasuCronScheduler.getInstance().stop();
		DasuMessageManager.getInstance().stop();

		if (AMCoordinatorV2.isActive()) {
			LOGGER.info("Stopping AMCoordinator");
			AMCoordinatorV2.stop();
		}

		LOGGER.info("Stopping rule engines");

		if (engine != null && engine.isAlive()) engine.stopEngine();
		if (AMEngine != null && AMEngine.isAlive()) AMEngine.stopEngine();
		
		if(dlm != null) dlm.remove();
		
		if(rm != null) rm.destroy();
		
		for (String param : COConfigurationListeners.keySet()) {
			COConfigurationManager.removeParameterListener(param, COConfigurationListeners.get(param));
		}

		Main.logPlugin("Inference engine disabled...");		
		try {
			finalize();
		} catch (Throwable e) {
			e.printStackTrace();
		}

		System.out.println("DasuManager: Buh-bye");
		LOGGER.info("Done...");

	}

	/**
	 * 
	 */
	public void getDasuServerList() {
		
		DasuUrls.put(1.0, DasuConfiguration.DASU_DEFAULT_SERVER);
		Properties servers = Main.downloadFromURL(DasuConfiguration.DASU_SERVER_LIST_LOC, 30 * 1000);
		for (Entry<Object, Object> ent : servers.entrySet()) {
			try {
				String pair[] = ((String) ent.getValue()).split(";");
				DasuUrls.put(Double.parseDouble(pair[1]), pair[0]);
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
		// remove default if there's more than one server and its weight wasn't
		// changed
		if (DASU_SKIP_SERVER_SWITCH) 
			return;
		
		if (DasuUrls.size() > 1 && DasuUrls.get(1.0).equals(DasuConfiguration.DASU_DEFAULT_SERVER))
			DasuUrls.remove(1.0);

		double val = Main.getRandom().nextDouble();
		for (Entry<Double, String> ent : DasuUrls.entrySet()) {
			if (val <= ent.getKey()) {
				currentDasuUrl = ent.getValue();
				break;
			}
		}
	}
	
	public String getNewRandomDasuUrl() {

		String thisUrl = currentDasuUrl;
		String oldUrl = currentDasuUrl;
		
		if (DasuUrls.size() <= 1) {
			currentDasuUrl = thisUrl;
			return thisUrl;
		}
		else {
			do {
				double val = Main.getRandom().nextDouble();
				for (Entry<Double, String> ent : DasuUrls.entrySet()) {
					if (val <= ent.getKey())
						thisUrl = ent.getValue();
				}
			} while (!thisUrl.equals(oldUrl));
		}
		currentDasuUrl = thisUrl;
		return thisUrl;
	}

}
