/**
 * Dasu Project
 *
 * File:         AMCoordinatorV2.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Dec 15, 2009 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.database;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DBManager class periodically reports data to the Main Database.
 **/

public class DBManager implements Runnable {	
	
	private static DBManager self = null;
	private static DatabaseReporter reporter;
	private static boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;

	private TreeMap<Long, Hashtable<String, Hashtable<String, Integer>>> dbIdMaps =
		DasuManager.getPersistentDataManager().getDbIdMaps();
	Long currSessionTime = 0l;
	long lastStaticFlushTime = -1;

	private static boolean DEBUG = false;
	
	
	
	public DBManager(DatabaseReporter reporter) {
		self = this;
		isActive = true;
		DBManager.reporter = reporter;
	}

	public static DBManager getInstance() {
		return self;
	}

	public static DatabaseReporter getReporter() {
		return reporter;
	}
	
	
	private static Object invokeMethod(String methodName, List<Object> params) {
		
		if (DEBUG) {
			System.out.println("phpRporter id_size: " + PhpReporter.peerIds.size() +  " map_size: " + PhpReporter.peerMap.size());			
			System.out.println("Statistics id_size: " + Statistics.peerIds.size() + " map_size: " + Statistics.peerMap.size());
			System.out.println("invokeMethod: " + methodName + " " + params);
		}
		
		try {

            Class<?> paramTypes[] = new Class[params.size()];
            for (int i=0; i < params.size(); i++) { 
            	if ( params.get(i) == null) {
            		paramTypes[i] = String.class;
            		continue;
            	}
            	paramTypes[i] = params.get(i).getClass();
            	if ( paramTypes[i].getName().contains("Integer"))
            		paramTypes[i] = Integer.TYPE;
            	if ( paramTypes[i].getName().contains("Long"))
            		paramTypes[i] = Long.TYPE;
            	if ( paramTypes[i].getName().contains("Float"))
            		paramTypes[i] = Float.TYPE;
            	if ( paramTypes[i].getName().contains("Boolean"))
            		paramTypes[i] = Boolean.TYPE;
            	if ( paramTypes[i].getName().contains("Double"))
            		paramTypes[i] = Double.TYPE;
            }
            
            Method method = PhpReporter.getInstance().getClass().getMethod(methodName, paramTypes);
            return method.invoke(PhpReporter.getInstance(), params.toArray());
        }
        catch (Throwable e) {
      	  e.printStackTrace();
        }
      
        return null;
	}

	private Long getSessionTime(long currTime) {

		Long ret = 0l;
		ArrayList<Long> toRemove = new ArrayList<Long>();
		
		Iterator<Long> iter = dbIdMaps.keySet().iterator();
		while (iter.hasNext()) {
			Long sessionTime = iter.next();
			if ( sessionTime <= currTime ) {
				
				//finished with old session, remove from map
				if ( !ret.equals(0l))
					toRemove.add(ret);

				ret = sessionTime;
				
			}
		}
		
		for ( Long r : toRemove)
			dbIdMaps.remove(r);
		
		return ret;

	}
	
	private Long switchMap(long currTime) {
		
		if (DEBUG)
			System.out.println("\nswitch map! before" + dbIdMaps.keySet());

		synchronized(Statistics.sessionRestartMonitor) {

			Long sessionStartTime = getSessionTime(currTime);
			
			if (sessionStartTime != currSessionTime) {
				
				PhpReporter.peerIds.clear();
				PhpReporter.routerIds.clear();
				PhpReporter.urlIds.clear();
				PhpReporter.peerClientIds.clear();
				PhpReporter.dnsIds.clear();
				
			}
	
			PhpReporter.peerMap = dbIdMaps.get(sessionStartTime).get("peerMap");
			PhpReporter.routerMap = dbIdMaps.get(sessionStartTime).get("routerMap");
			PhpReporter.urlMap = dbIdMaps.get(sessionStartTime).get("urlMap");
			PhpReporter.peerClientMap = dbIdMaps.get(sessionStartTime).get("peerClientMap");
			PhpReporter.dnsMap = dbIdMaps.get(sessionStartTime).get("dnsMap");

			if (DEBUG)
				System.out.println("switch map! after" + dbIdMaps.keySet() + "\n");

			return sessionStartTime;
			
		}
	}
	
	public void initialize() {	
		
		currSessionTime = Main.sessionStartTime;	
		switchMap(currSessionTime);
	}
	
	public void run() {
				
		//TODO: ensure the db is up before removing entry from list
		//TODO: ensure the entry was successfully inserted before moving on to next one
		while (isActive) {

			int byteCount = 0;
			if ( reporter instanceof CallbackReporter) {
				
				List<Long> entries = new ArrayList<Long>();
				synchronized( ((CallbackReporter) reporter).dbQueue.keySet() ) {
					entries.addAll(((CallbackReporter) reporter).dbQueue.keySet());
				}
				Collections.sort(entries);

				/** the largest post that will be sent, 20K for now */
				while (byteCount < 20*1024 && entries.size() > 0) {

					if (!isActive)
						break;
					
					Long time = entries.remove(0);
					
					if (( time < currSessionTime) || ( currSessionTime != Main.sessionStartTime ))
						currSessionTime = switchMap(time);

					//update all ids
					batchProcessIds();
					List<Object> toInvoke = null;
					synchronized( ((CallbackReporter) reporter).dbQueue ) {
						toInvoke = ((CallbackReporter) reporter).dbQueue.remove(time);
						if (DEBUG)
							System.out.println( "map size: " + ((CallbackReporter) reporter).dbQueue.size() + " " + toInvoke);
					}
					
					if ( toInvoke != null ){

						byteCount += sizeof(toInvoke);

						if (DEBUG)
							System.out.println("byteCount: " + byteCount);
						
						String methodName = (String) toInvoke.get(0);
						List<Object> params = toInvoke.subList(1, toInvoke.size());
						Object success = invokeMethod(methodName, params);
						
						//System.out.println(success.getClass() + " " + success + "\n\n\n\n\n");
						if ( success instanceof Boolean) {
							
							if ( !((Boolean)success) ) {
							
								if (DEBUG)
									System.out.println( "Submission failed?");
								
								//send a request to the db to ensure the db is down and not a fluke of this
								//report, this will avoid losing data reporting over a one-entry error
								if ( Statistics.getInstance().makeUUIDid(DasuConfiguration.getInstance().getUUID()) == -1) {
								
									synchronized( ((CallbackReporter) reporter).dbQueue ) {
										toInvoke = ((CallbackReporter) reporter).dbQueue.put(time, toInvoke);
										
										if (DEBUG)
											System.out.println( "Submission really failed!! wait for some time and try again later!! \n\n\n\n\n");
										
										//reduce buffer size to 1byte to avoid losing data
										PhpReporter.POST_THRESHOLD = 1;
										break;
									}
								}
							}
							else {
								
								//data submitted successfully, reset buffer size
								if (PhpReporter.POST_THRESHOLD < 5*1024) {
									PhpReporter.POST_THRESHOLD = 5*1024;
									
									if (DEBUG) System.out.println( "Submission success?!! Increase buffer size  to 5K\n\n\n\n\n");
								}
								
							}
						}
							
					}					
				}				
				
			}
			
			//perform static flush
			int staticFlushInterval = DasuConfiguration.getInstance().getDbStaticFlushInterval();
			
			if (PhpReporter.POST_THRESHOLD > 1 && Util.currentGMTTime()-lastStaticFlushTime >= staticFlushInterval * 1000){
				
				if (DEBUG) System.out.println("Static flush! Next one in " + staticFlushInterval + " seconds...");										
				
				Main.getReporter().flush();
				lastStaticFlushTime = Util.currentGMTTime();
				
			}
			
			synchronized (halt) {

				try {
					if (!terminate) {

						int sleepInterval = DasuConfiguration.getInstance().getDBManagerSleepInterval();
						if (DEBUG) {
							System.out.println("DBManager(): sleeping for " + sleepInterval + " seconds...");
							halt.wait(sleepInterval * 1000);
						}
						else {
							halt.wait(sleepInterval * 1000);
						}
						
					}
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
	
				
			}
		}
		

		if (DEBUG)
			System.out.println("DBManager(): Hasta Luego...");
	}
	
	
	public int batchProcessTable(String tableName, String columnName,
			Hashtable<String, Integer> localMap, Hashtable<Integer, Integer> dbMap) {

		int numRemaining = 0;
		synchronized (localMap) {
			LinkedList<String> keys = new LinkedList<String>();
			for (Entry<String, Integer> ent : localMap.entrySet()) {
				if (!dbMap.containsKey(ent.getValue())) {
					keys.add(ent.getKey());
				}
			}

			if (keys.size() == 0)
				return 0;
			
			if (DEBUG)
				System.out.println(keys);
			
			Map<String, Integer> dbKeys = PhpReporter.getInstance().makeIdBatch(tableName, columnName, keys);

			if (dbKeys == null)
				dbKeys = new HashMap<String, Integer>();

			int numFound = 0;
			for (Entry<String, Integer> ent : dbKeys.entrySet()) {
				dbMap.put(localMap.get(ent.getKey()), ent.getValue());
				numFound++;
			}
			
			//in case the db returned 0 mappings, set mappings to -1 to avoid infinite loop
			if (numFound == 0) {
				for (String ent : keys) {
					dbMap.put(localMap.get(ent), -1);
					numFound++;
				}				
			}
			numRemaining = keys.size() - numFound;
		}
		return numRemaining;
	}
	
	public void batchProcessIds() {
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return;

		int totalLeft;
		int numLoops = 0; // to ensure termination
		do {
			totalLeft = 0;
			totalLeft += batchProcessTable("peerips", "ip", PhpReporter.peerMap, PhpReporter.peerIds);
			totalLeft += batchProcessTable("peerclients", "name", PhpReporter.peerClientMap, PhpReporter.peerClientIds);
			totalLeft += batchProcessTable("routerips", "ip", PhpReporter.routerMap, PhpReporter.routerIds);
			totalLeft += batchProcessTable("urlids", "url", PhpReporter.urlMap, PhpReporter.urlIds);
			totalLeft += batchProcessTable("dnsips", "ip", PhpReporter.dnsMap, PhpReporter.dnsIds);
			numLoops++;
		} while (totalLeft > 0 && Main.getReporter().canConnect()
				&& numLoops < 1000);

	}

	public void stop() {

		//do I want to awake the manager one last time to send flushed data		
		//before closing up?
		isActive = false;
		synchronized (halt) {
			terminate = true;
			halt.notifyAll();
		}
		
	}
	
    private static final int SZ_REF = 4;
    
    public static int sizeof(boolean b) {
            return 1;
    }

    public static int sizeof(byte b) {
            return 1;
    }

    public static int sizeof(char c) {
            return 2;
    }

    public static int sizeof(short s) {
            return 2;
    }

    public static int sizeof(int i) {
            return 4;
    }

    public static int sizeof(long l) {
            return 8;
    }

    public static int sizeof(float f) {
            return 4;
    }

    public static int sizeof(double d) {
            return 8;
    }

    private static int size_inst(Class<?> c) {
            Field flds[] = c.getDeclaredFields();
            int sz = 0;

            for (int i = 0; i < flds.length; i++) {
                    Field f = flds[i];
                    if (!c.isInterface() &&
                        (f.getModifiers() & Modifier.STATIC) != 0)
                            continue;
                    sz += size_prim(f.getType());
            }

            if (c.getSuperclass() != null)
                    sz += size_inst(c.getSuperclass());

            Class<?> cv[] = c.getInterfaces();
            for (int i = 0; i < cv.length; i++)
                    sz += size_inst(cv[i]);

            return sz;
    }

    private static int size_prim(Class<?> t) { 
            if (t == Boolean.TYPE)
                    return 1;
            else if (t == Byte.TYPE)
                    return 1;
            else if (t == Character.TYPE)
                    return 2;
            else if (t == Short.TYPE)
                    return 2;
            else if (t == Integer.TYPE)
                    return 4;
            else if (t == Long.TYPE)
                    return 8;
            else if (t == Float.TYPE)
                    return 4;
            else if (t == Double.TYPE)
                    return 8;
            else if (t == Void.TYPE)
                    return 0;
            else
                    return SZ_REF;
    }

    private static int size_arr(Object obj, Class<?> c) {
            Class<?> ct = c.getComponentType();
            int len = Array.getLength(obj);

            if (ct.isPrimitive()) {
                    return len * size_prim(ct);
            }
            else {
                    int sz = 0;
                    for (int i = 0; i < len; i++) {
                            sz += SZ_REF;
                            Object obj2 = Array.get(obj, i);
                            if (obj2 == null)
                                    continue;
                            Class<?> c2 = obj2.getClass();
                            if (!c2.isArray())
                                    continue;
                            sz += size_arr(obj2, c2);
                    }
                    return sz;
            }
    }

    public static int sizeof(Object obj) {
            if (obj == null)
                    return 0;

            Class<?> c = obj.getClass();

            if (c.isArray())
            	return size_arr(obj, c);
            else
            	return size_inst(c);
    }

	
}
