/**
 * Dasu Project
 *
 * File:         RuleBasedEngine.java
 * RCS:          $Id$
 * Description:  RuleBasedEngine class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:23:51 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.drools
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */


package edu.northwestern.dasu.drools;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.Vector;
import java.util.logging.Logger;

import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderConfiguration;
import org.drools.builder.KnowledgeBuilderError;
import org.drools.builder.KnowledgeBuilderErrors;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.event.rule.ObjectInsertedEvent;
import org.drools.event.rule.ObjectRetractedEvent;
import org.drools.event.rule.ObjectUpdatedEvent;
import org.drools.event.rule.WorkingMemoryEventListener;
import org.drools.io.ResourceFactory;
import org.drools.logger.KnowledgeRuntimeLogger;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.rule.ConsequenceException;
import org.drools.runtime.rule.FactHandle;

import edu.northwestern.dasu.Main;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The RuleBasedEngine class ...
 */
public abstract class RuleBasedEngine extends Thread {

	private final static Logger LOGGER = Logger.getLogger(RuleBasedEngine.class.getName());
	
	private boolean active = false;
	private boolean initialized = false;
	protected String filename = null;
	protected String sigFilename = null;
	protected int taskId = -1;

	protected KnowledgeBase kbase = null;
	protected StatefulKnowledgeSession ksession = null; 
	protected KnowledgeRuntimeLogger logger = null;

	private Hashtable<Long, FactHandle> facts = new Hashtable<Long, FactHandle>();
	private Set<FactHandle> modifiedFacts = 
		Collections.synchronizedSet(new HashSet<FactHandle>());
	
	protected RuleBasedEngine(String name) throws Exception {
		super(name);
	}
	
	protected RuleBasedEngine(String name, String filename, String sigFilename, int taskId) throws Exception {
		super(name);
		this.filename = filename;
		this.sigFilename = sigFilename;
		this.taskId = taskId;
	}
	

    // Add the object and its contents to working memory
	public void addFact(final FactTemplate factObject) {
		
		if (isActive()) {
		
			if (!facts.containsKey(factObject.getId())) {
				FactHandle fh = ksession.insert(factObject);
				facts.put(factObject.getId(), fh);
			
				factObject.addPropertyChangeListener(
						new PropertyChangeListener() {
							public void propertyChange(PropertyChangeEvent e) {
						        updateFact(factObject);
							}
						});
			}
			//else 
			//	updateFact(factObject);
		}
	}
		
	/*
	 * XXX: modifying fact can quietly fail if it is being modified...
	 */
	private void updateFact(FactTemplate factObject) {
					
		synchronized(factObject) {
					
			if (isActive()) {
				FactHandle fh = ksession.getFactHandle(factObject);
				
				if (fh != null) {
					
					/* object is being modified in working memory,
					 * avoid concurrent modification by doing nothing 
					 */
	
					if (modifiedFacts.contains(fh)) {
						LOGGER.fine("!!!!!DEBUG: updateFact(): Fact being modified, skip: " + factObject);
						return;
					}
					
					//LOGGER.fine("!!!!!DEBUG: updateFact(): " + factObject);
					ksession.update(fh, factObject);
					//ksession.update(facts.get(factObject.getId()), factObject);
					
				}
				else
					facts.remove(fh);
			}

		}
				
	}
	
    // Remove the object and its contents from working memory
	public void retractFact(FactTemplate factObject) {
		if (isActive()) {
			FactHandle fh = ksession.getFactHandle(factObject);
			if (fh != null)
				ksession.retract(fh);
//				ksession.retract(facts.get(factObject.getId()));
			else
				facts.remove(factObject.getId());
		}
			
	}


	/* FireAllRules, GetGlobal, SetGlobal, InsertObject, InsertElements
	Query, StartProcess, BatchExecution */
	
	public void executeCommand(String command) {
	
		System.out.println(command);
		//ksession.execute(command);
		
	}
	
	public void initialize() throws Exception {
		
		if (loadRules()) {
			loadFacts();
			setInitialized(true);
		}
		else {
			throw new Exception("No rules to load... aborting engine startup...");
		}
			
		
	}

	protected boolean loadRules() {
		return false;
	}

	@Override
	public void run() {
		
		if (isInitialized()) {
			setActive(true);
			
				LOGGER.info(this.getName() + ": Starting execution Drools engine\n\n");
		}

		try {
			while (!Main.isShuttingDown() && isActive()) {
				
				ksession.fireAllRules();
				if ( ksession.getGlobal("halt") != null && ((FactHalt)ksession.getGlobal("halt")).isHalt())
					break;				
				Thread.sleep(100);
			} // end while active
		}catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		ksession.halt();
		LOGGER.info(this.getName() + ": Finishing execution of Drools engine\n\n");
		setActive(false);
		
		destroySession();

	}
	
	protected void destroySession() {

		if (ksession != null)
			ksession.dispose();
	}
	
	protected KnowledgeBase loadRules(Vector<ByteArrayOutputStream> data) {
		Properties props = new Properties();
        props.setProperty("drools.dialect.java.compiler", "JANINO");
        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, (ClassLoader[]) null);
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
		
		for(ByteArrayOutputStream baos: data) {
			kbuilder.add(ResourceFactory.newByteArrayResource(baos.toByteArray()), ResourceType.DRL);		
			KnowledgeBuilderErrors errors = kbuilder.getErrors();
			if (errors.size() > 0) {
				for (KnowledgeBuilderError error: errors) {
					LOGGER.severe(error.toString());
				}
				throw new IllegalArgumentException("Could not parse knowledge.");
			}
		}
		
		//KnowledgeBaseConfiguration baseconfig = KnowledgeBaseFactory.newKnowledgeBaseConfiguration(null, Main.getPluginInterface().getPluginClassLoader());
		//kbase = KnowledgeBaseFactory.newKnowledgeBase(baseconfig);
		kbase = KnowledgeBaseFactory.newKnowledgeBase();
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		
		return kbase;

	}
	
	protected KnowledgeBase loadRulesFromDir(String dir, String firstName) {
		Properties props = new Properties();
        props.setProperty("drools.dialect.java.compiler", "JANINO");
        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, (ClassLoader[]) null);
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
		
		File dirFile = new File(dir);
		
		kbuilder.add(ResourceFactory.newFileResource(new File(dirFile, firstName)), ResourceType.DRL);		
		KnowledgeBuilderErrors errors = kbuilder.getErrors();
		if (errors.size() > 0) {
			for (KnowledgeBuilderError error: errors) {
				LOGGER.severe(error.toString());
			}
			throw new IllegalArgumentException("Could not parse knowledge.");
		}
		
		for(File f: dirFile.listFiles()) {
			if(!f.getName().equals(firstName) && f.getName().endsWith(".drl")) {
				kbuilder.add(ResourceFactory.newFileResource(f), ResourceType.DRL);		
				errors = kbuilder.getErrors();
				if (errors.size() > 0) {
					for (KnowledgeBuilderError error: errors) {
						LOGGER.severe(error.toString());
					}
					throw new IllegalArgumentException("Could not parse knowledge.");
				}
			}
		}
		
		//KnowledgeBaseConfiguration baseconfig = KnowledgeBaseFactory.newKnowledgeBaseConfiguration(null, Main.getPluginInterface().getPluginClassLoader());
		//kbase = KnowledgeBaseFactory.newKnowledgeBase(baseconfig);
		kbase = KnowledgeBaseFactory.newKnowledgeBase();
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		
		return kbase;

	}
	
	protected KnowledgeBase loadRules(String fileName) {
		Properties props = new Properties();
        props.setProperty("drools.dialect.java.compiler", "JANINO");
        KnowledgeBuilderConfiguration config = KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, (ClassLoader[]) null);
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder(config);
		
		kbuilder.add(ResourceFactory.newFileResource(fileName), ResourceType.DRL);		
		KnowledgeBuilderErrors errors = kbuilder.getErrors();
		if (errors.size() > 0) {
			for (KnowledgeBuilderError error: errors) {
				LOGGER.severe(error.toString());
			}
			throw new IllegalArgumentException("Could not parse knowledge.");
		}
		//KnowledgeBaseConfiguration baseconfig = KnowledgeBaseFactory.newKnowledgeBaseConfiguration(null, Main.getPluginInterface().getPluginClassLoader());
		//kbase = KnowledgeBaseFactory.newKnowledgeBase(baseconfig);
		kbase = KnowledgeBaseFactory.newKnowledgeBase();
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		
		return kbase;

	}
	
	protected void loadFacts() {}

	public void displayFacts() {
		
		if (isActive()) {
			Iterator<FactHandle> f = ksession.getFactHandles().iterator();
			while (f.hasNext()) {
				FactHandle fh = f.next();
				Object obj = ksession.getObject(fh);
				if(obj != null)
					System.out.println(getName() + " => " + obj.toString());
			}
		}
		
	}
	
	
	public void stopEngine() {
		LOGGER.info("stopEngine()");
		
		if (isActive()) {
			ksession.halt();
			setActive(false);
		}
	}
	
	public boolean isActive() {
		return active;
	}
	
	public void setActive(boolean val) {
		this.active = val;
	}
	
	public void setInitialized(boolean state) {
		initialized = state;
	}
	
	public boolean isInitialized() {
		return initialized;
	}

	public void setModifiedFacts(Set<FactHandle> modifiedFacts) {
		this.modifiedFacts = modifiedFacts;
	}

	public Set<FactHandle> getModifiedFacts() {
		return modifiedFacts;
	}
	
}
