/**
 * Dasu Project
 *
 * File:         AMCoordinatorV2.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Dec 15, 2009 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.logging.Logger;

import org.gudy.azureus2.core3.util.AEThread;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadException;
import org.gudy.azureus2.plugins.download.DownloadManager;
import org.gudy.azureus2.plugins.torrent.Torrent;
import org.gudy.azureus2.plugins.torrent.TorrentDownloader;
import org.gudy.azureus2.plugins.torrent.TorrentException;
import org.gudy.azureus2.plugins.torrent.TorrentManager;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactDnsProbeResult;
import edu.northwestern.dasu.drools.FactFireAction;
import edu.northwestern.dasu.drools.FactForcedSeedingThroughPortTestResult;
import edu.northwestern.dasu.drools.FactHttpProbeResult;
import edu.northwestern.dasu.drools.FactNdtProbeResult;
import edu.northwestern.dasu.drools.FactPingProbeResult;
import edu.northwestern.dasu.drools.FactScrapeProbeResult;
import edu.northwestern.dasu.drools.FactSpeedTestResult;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.drools.FactTestResult;
import edu.northwestern.dasu.drools.FactTraceResult;
import edu.northwestern.dasu.drools.FactUpnpTestResult;
import edu.northwestern.dasu.drools.RuleBasedEngine;
import edu.northwestern.dasu.measurement.AzureusSpeedTest;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.timer.CronjobEntry;
import edu.northwestern.dasu.timer.DasuCronScheduler;
import edu.northwestern.dasu.timer.DasuMessageManager;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.util.HashMapCache;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The AMCoordinatorV2 class coordinates measurements/tests
 *         between all measurement/test modules.
 **/

public class AMCoordinatorV2 implements Runnable {	
	
	private static class ProbeSubTask {
		String ipAddress;
		SourceType sourceType;
		Integer taskId;
		Map params;
		String target = null;
		
		private ProbeSubTask(String ipAddress, SourceType sourceType, Integer taskId) {
			this.ipAddress = ipAddress;
			this.sourceType = sourceType;
			this.taskId = taskId;
		}
		
		private ProbeSubTask(String ipAddress, SourceType sourceType, Integer taskId, String target) {
			this.ipAddress = ipAddress;
			this.sourceType = sourceType;
			this.taskId = taskId;
			this.target = target;
		}
		
		private ProbeSubTask(String ipAddress, SourceType sourceType, Integer taskId, Map params) {
			this.ipAddress = ipAddress;
			this.sourceType = sourceType;
			this.taskId = taskId;
			this.params = params;
		}
		
		private ProbeSubTask(String ipAddress, SourceType sourceType, Integer taskId, Map params, String target) {
			this.ipAddress = ipAddress;
			this.sourceType = sourceType;
			this.taskId = taskId;
			this.params = params;
			this.target = target;
		}

	}
	
	private final static Logger LOGGER = Logger.getLogger(AMCoordinatorV2.class.getName());
	
	private static boolean active = false;
	private static boolean pause = false;
	private static AMCoordinatorV2 self = null;
	
	/* sleepInterval in seconds */
	private static int sleepInterval = -1;
	
	/* Map of ProbeType to DasuProbeModule instances */
	private static Hashtable<ProbeType, DasuProbeModule> probeModules = null;
 
	/* Queue of Tasks to be executed by the AMC */
	private static Hashtable<ProbeType, PriorityBlockingQueue<Pair<Long, ProbeTask>>> 
		probeTaskQueue = null;
	
	
	/************* TESTS' RELATED STRUCTURES **************/
	
	/* Queue of Tests to be executed by the AMC */
	private static BlockingQueue<TestTask> testQueue = null;
	
	/* Holds the results from dispatched tests, written to by the test thread itself*/
	public static Hashtable<Long, FactTestResult> testResultTable = null;
	
	/* Holds the tests that are already running */
	private static Hashtable<TestType, SourceType> dispatchedTests = null;
	
	/*************** AMC CONTROL STRUCTURES & VARIABLES ************/

	private static int shortTermInterval = 1;  //multiples of one minute
	private static ITimer hourlyHistoryUpdaterTimer;
	
	private static Hashtable<ProbeType, Integer> maxProbePerMin = 
		new Hashtable<ProbeType, Integer>();
	private static Hashtable<TestType, Integer> maxTestPerMin = 
		new Hashtable<TestType, Integer>();
	private static Hashtable<ProbeType, Integer> maxProbePerHour = 
		new Hashtable<ProbeType, Integer>();
	private static Hashtable<TestType, Integer> maxTestPerHour = 
		new Hashtable<TestType, Integer>();
	
	/*holds probe/test history for the past minute*/
	private static LinkedHashMap<Pair<ProbeType, String>, Long> shortTermProbeHistory 
		= null;
	private static LinkedHashMap<Pair<TestType, String>, Long> shortTermTestHistory 
		= null;
	private static Hashtable<ProbeType, Integer> shortTermProbeTally 
		= null;
	private static Hashtable<TestType, Integer> shortTermTestTally 
		= null;

	private static LinkedHashMap<Pair<ProbeType, String>, Long> temporaryShortTermProbeHistory 
		= new LinkedHashMap<Pair<ProbeType, String>, Long>();
	private static LinkedHashMap<TestType, Pair<String, Long>> temporaryShortTermTestHistory 
		= new LinkedHashMap<TestType, Pair<String, Long>>();

	
	/*holds the hourly tally (count) of probes*/
	private static Hashtable<ProbeType, Integer> hourlyProbeTally =
		null;
	private static Hashtable<TestType, Integer> hourlyTestTally =
		null;

	/*
	 * To keep track of Inference-Engine launched priority tasks that will
	 * awake the AMC
	 */
	private static BlockingQueue<ProbeType> REProbeTaskQueue = null;
	
	/* 
	 * Queue used by every DasuProbeModule (and Jess Engine) used to alert (awake) the AMC
	 * of probe results it needs to collect (or of probe tasks it needs to perform)
	 */
	protected static PriorityBlockingQueue<Pair<Long, EventType>> events = null;

	/* 
	 * Used by the AMC to keep track of the probes already dispatched and to know
	 * which ProbeTask launched what ProbeType when the results are back 
	 */
	private static Hashtable<Pair<ProbeType, String>, SourceType> dispatchedEvents = null;

	/* probeTask history */
	private static HashMapCache<Integer, Long> taskLogHistory = null;
	
	private static LinkedList<Download> activeTorrents =
		new LinkedList<Download>();
	
	private static Hashtable<String, HistoricSignalHolder> historicSamples
		= null;
	
	private static Hashtable<String, SessionSignalHolder> sessionSamples =
		new Hashtable<String, SessionSignalHolder>();
	
	private AMCoordinatorV2() {
		sleepInterval = DasuConfiguration.getInstance().getAMCSleepInterval();
		shortTermProbeHistory 
			= DasuManager.getPersistentDataManager().getShortTermProbeHistory();
		shortTermTestHistory 
			= DasuManager.getPersistentDataManager().getShortTermTestHistory();
		shortTermProbeTally 
			= DasuManager.getPersistentDataManager().getShortTermProbeTally();
		shortTermTestTally 
			= DasuManager.getPersistentDataManager().getShortTermTestTally();
	
		hourlyProbeTally =
			DasuManager.getPersistentDataManager().getHourlyProbeTally();
		hourlyTestTally =
			DasuManager.getPersistentDataManager().getHourlyTestTally();
	
		taskLogHistory =
			DasuManager.getPersistentDataManager().getTaskLogHistory();
	
		historicSamples
			= DasuManager.getPersistentDataManager().getHistoricSamplesV2();
	
		setActive(true);
		
		//System.out.println("\n\n\n\n\n\n\n\n\n\n\n" + DasuManager.getPersistentDataManager().getUIValues() + "\n\n\n\n\n\n\n\n\n\n");
	}

	public static AMCoordinatorV2 getInstance() {
		if (self == null)
			self = new AMCoordinatorV2();
		return self;
	}

	private static void initializeProbeLimits() {
		//XXX: These should be in a configuration file (JSO)
		//XXX: Not really, configuration files make the values easy to change,
		// these thresholds should be hard-coded in the source (MAS)
				
		maxProbePerHour.put(ProbeType.PING, 600);
		maxProbePerHour.put(ProbeType.TRACEROUTE, 600);
		maxProbePerHour.put(ProbeType.NDT, 1);
		maxProbePerHour.put(ProbeType.HTTP, 600);
		maxProbePerHour.put(ProbeType.NDT2, 5);
		maxProbePerHour.put(ProbeType.SCRAPE, 60);
		maxProbePerHour.put(ProbeType.DNS, 2000);
		
		maxTestPerHour.put(TestType.AZUREUS_DL, 2);
		maxTestPerHour.put(TestType.AZUREUS_UL, 2);
		maxTestPerHour.put(TestType.AZUREUS_DL_ENCRYPTION, 2);
		maxTestPerHour.put(TestType.AZUREUS_UL_ENCRYPTION, 2);
		maxTestPerHour.put(TestType.AZUREUS_UPNP, 100);
		
		maxProbePerMin.put(ProbeType.PING, 20);
		maxProbePerMin.put(ProbeType.TRACEROUTE, 20);
		maxProbePerMin.put(ProbeType.HTTP, 20);
		maxProbePerMin.put(ProbeType.NDT, 1);
		maxProbePerMin.put(ProbeType.NDT2, 1);
		maxProbePerMin.put(ProbeType.SCRAPE, 1);
		maxProbePerMin.put(ProbeType.DNS, 20);
		
		maxTestPerMin.put(TestType.AZUREUS_UPNP, 10);
		maxTestPerMin.put(TestType.AZUREUS_DL, 2);
		maxTestPerMin.put(TestType.AZUREUS_UL, 2);
		maxTestPerMin.put(TestType.AZUREUS_DL_ENCRYPTION, 2);
		maxTestPerMin.put(TestType.AZUREUS_UL_ENCRYPTION, 2);
		
		for ( ProbeType p : ProbeType.values()) {
			
			if ( !shortTermProbeTally.contains(p) ) {
				shortTermProbeTally.put(p, 0);
				hourlyProbeTally.put(p, 0);
			}
		}

		for ( TestType t : TestType.values()) {
			
			if ( !shortTermTestTally.contains(t) ) {
				shortTermTestTally.put(t, 0);
				hourlyTestTally.put(t, 0);
			}
		}
		
		launchHourlyProbeHistoryUpdater();
		
	}
	
	private static void initialize() {

		probeModules = new Hashtable<ProbeType, DasuProbeModule>();
		probeTaskQueue = new Hashtable<ProbeType, PriorityBlockingQueue<Pair<Long, ProbeTask>>>();
		testQueue = new LinkedBlockingQueue<TestTask>();
		testResultTable = new Hashtable<Long, FactTestResult>();
		dispatchedTests = new Hashtable<TestType, SourceType>();
		REProbeTaskQueue = new LinkedBlockingQueue<ProbeType>();
		events = new PriorityBlockingQueue<Pair<Long, EventType>>();
		dispatchedEvents = new Hashtable<Pair<ProbeType, String>, SourceType>();

		for (ProbeType pt : ProbeType.values()) 
			probeTaskQueue.put(pt, new PriorityBlockingQueue<Pair<Long, ProbeTask>>() );
		
		startDasuAmModules();
		initializeProbeLimits();
	}
	
	/* 
	 * Initialize each DasuModule available
	 */
	private static void startDasuAmModules() {
		
		LOGGER.info("AMCoordinator: startDasuAmModules()");
		
		probeModules.put(ProbeType.PING, DasuPingModuleV2.getInstance());		
		probeModules.put(ProbeType.TRACEROUTE, DasuTraceRouteModuleV2.getInstance());
		probeModules.put(ProbeType.HTTP, DasuHttpModule.getInstance());
		probeModules.put(ProbeType.NDT, DasuNdtModule.getInstance());
		probeModules.put(ProbeType.NDT2, DasuNdtModule.getInstance());
		probeModules.put(ProbeType.SCRAPE, DasuScrapeModule.getInstance());
		probeModules.put(ProbeType.DNS, DasuDnsModule.getInstance());

		for ( ProbeType pt : ProbeType.values()) {
			if (!probeModules.get(pt).isAlive()) {
				probeModules.get(pt).start();
			}
		}
		
	}
	
	private static void stopDasuAmModules() {
		
		for ( ProbeType pt : ProbeType.values()) {
			if (probeModules != null) {
				if (probeModules.get(pt).isAlive()) { 
					
					if (pt.equals(ProbeType.NDT) && 
							!probeModules.get(pt).getState().equals(Thread.State.TIMED_WAITING))
						probeModules.get(pt).interrupt();
					
					
					probeModules.get(pt).stopModule();
				}
			}
		} 
		
	}

	
	/* Add to the *task* queue */
	
	public static boolean addTestTask(TestType testType) {
		return addTestTask(testType, SourceType.RE, null);
	}
	
	public static boolean addTestTask(TestType testType, Object[] params) {
		return addTestTask(testType, SourceType.RE, params);
	}
	
	public static boolean addTestTask(TestType testType, Object param) {
		return addTestTask(testType, SourceType.RE, new Object[]{param});
	}	

	public static boolean addTestTask(TestType testType, SourceType sourceType, Object[] params) {
		
		updateTestHistory();
		
		if (!testType.equals(TestType.AZUREUS_UPNP) && 
				!testType.equals(TestType.FORCE_SEEDING_THROUGH_PORT) &&
				!DasuConfiguration.getInstance().isEnableAzureusSpeedTest()) {
			
			LOGGER.info("DROP: Azureus Speed test disabled at this time...");

			return false;
		}
		
		if ((sourceType == SourceType.RE) || 
				(shortTermTestTally.get(testType) < maxTestPerMin.get(testType) * shortTermInterval) &&
				(hourlyTestTally.get(testType) < maxTestPerHour.get(testType))) {

			if (params != null)
				return testQueue.add(new TestTask(testType, params, sourceType));
			else
				return testQueue.add(new TestTask(testType, sourceType));

		}
		else {
			LOGGER.info("DROP: Too many tests of type " + testType + ", try again later... " +
						"ShortTerm: " + shortTermTestTally.get(testType) + " hourly: " + hourlyTestTally.get(testType));
		}
		
		return false;
	}
	

	/* Add to the *task* queue */
	public static boolean addProbeTask(ProbeTask pt) {
				
		updateProbeHistory();		
		
		if (Main.DEBUG) {
			for ( ProbeType p : probeTaskQueue.keySet())
				System.out.print(p + " " + probeTaskQueue.get(p).size() + " ");
			System.out.println("\n");
		}
		
		if ((pt.sourceType != SourceType.RE) && 
				((shortTermProbeTally.get(pt.probeType) >= maxProbePerMin.get(pt.probeType) * shortTermInterval) ||
				(hourlyProbeTally.get(pt.probeType) >= maxProbePerHour.get(pt.probeType)))) {
			
			if (Main.DEBUG) 
				System.out.println("DROP: Too many probes of type " + pt.probeType + ", try again later... " +
							"ShortTerm: " + shortTermProbeTally.get(pt.probeType) + " hourly: " + hourlyProbeTally.get(pt.probeType));

			return false;
		}
		
		/*if task is already in the log, don't add it*/
		if (taskLogHistory.containsKey((pt.taskId)))
			return false;
			
		/*if task is in the past, don't add to queue*/
		if (pt.endTime >= Util.currentDasuTime())
			return false;
		
		
		boolean ret = probeTaskQueue.get(pt.probeType).add(new Pair<Long, ProbeTask>(pt.startTime, pt));

		/*
		 *  If and ONLY if the task is of highest priority, add it to the *events* queue
		 *  in order to awake the AMC, otherwise let the AMC deal with it on next 
		 *  scheduled awake 
		 */

		if (pt.priority == -1) {
			synchronized(events) {
				Long id = System.nanoTime();
				if (events.size() != 0)
					id = events.peek().getKey() - 1000000;
				
				events.add(	new Pair<Long, EventType>(id, EventType.RE_REQUEST) );
				events.notifyAll();
			}
		}

		return ret;
	}
	
	private static boolean validateProbe(String ip, ProbeType pt) {
		if ( (pt.equals(ProbeType.PING) || (pt.equals(ProbeType.TRACEROUTE))) && 
				(!Util.isValidIp(ip)) ) {
					
			LOGGER.warning("Invalid IP " + ip);
			return false;

		}	else if ((pt.equals(ProbeType.NDT) || pt.equals(ProbeType.NDT2)) && 
				(ip == null) && DasuNdtModule.getServers().size() == 0) {

			LOGGER.warning("Invalid server name " + ip);

			return false;

		}
		return true;
	}
	
	public static boolean addLowPriorityProbe(String ip, ProbeType pt, SourceType st) {

		if(!validateProbe(ip, pt)) return false;

		ArrayList<String> ipList = new ArrayList<String>();
		ipList.add(ip);

		ProbeTask task = new ProbeTask(-1, -1, 0L, 0L, ipList, pt, TaskType.BULK, st);
		
		task.sourceType = st;
		
		return addProbeTask(task);		
		
	}
	
	/*To be used only by the Detection Rule Engine*/
	public static boolean addPriorityProbe(String ip, ProbeType pt) {
		return addPriorityProbe(ip, pt, null);
	}
	
	/*To be used only by the Detection Rule Engine*/
	public static boolean addPriorityProbe(String ip, ProbeType pt, String target) {
		if(!validateProbe(ip, pt)) return false;

		ArrayList<String> ipList = new ArrayList<String>();
		ipList.add(ip);

		ProbeTask task = new ProbeTask(-1, -1, 0L, 0L, ipList, pt, TaskType.BULK, SourceType.RE);
		task.setTarget(target);

		return addPriorityProbe(task);
	}
	
	/*To be used only by the Detection Rule Engine*/
	public static boolean addPriorityProbe(ProbeTask task) {
		
		boolean ret = false;
		
		//force the source of the task to be the RE, in case the user didn't		
		task.sourceType = SourceType.RE;

		if (REProbeTaskQueue != null) {
			
			REProbeTaskQueue.add(task.probeType);
			ret = addProbeTask(task);
			if(!ret) 
				REProbeTaskQueue.remove(task.probeType);
			
		}
		
		return ret;

	}
	
	/* 
	 * Returns (removes) ONE measurement from List member of the top-most ProbeTask 
	 * in the Queue. Does not remove the ProbeTask from Queue unless the 
	 * ProbeTask List is empty
	 */		
	private static ProbeSubTask popProbeSubTask ( ProbeType pt ) {
				
		PriorityBlockingQueue<Pair<Long, ProbeTask>> taskQueue
			= probeTaskQueue.get(pt);
			
		ProbeSubTask ret = null;
		Pair<Long, ProbeTask> firstTask = null;		
		firstTask = taskQueue.peek();

		while (firstTask != null) {

			/* it's priority task  or endTime = whenever possible */
			if ((firstTask.getValue().taskId == -1) || (firstTask.getKey() == -1)) 
				break;
			
			/* 
			 * endTime already passed, 
			 * remove from queue and try again if there are more tasks in the queue
			 */
			else if (firstTask.getValue().endTime >= Util.currentDasuTime()) {

				taskQueue.poll();
				firstTask = null;
				if (taskQueue.size() == 0)
					break;
			}

			/* it's not time to start the task yet */
			else if (firstTask.getValue().startTime < Util.currentDasuTime()) {
				firstTask = null;
				break;
			}				
		}
		
		
		if (firstTask != null) {
			ProbeTask task = firstTask.getValue();


			if (task.ipsToProbe.size() == 0) {
				taskQueue.poll();
			}
			
			/* if next probe is already dispatched (repeated), do nothing */

			else if (!dispatchedEvents.keySet().contains(
					new Pair<ProbeType, String>(firstTask.getValue().probeType, 
							firstTask.getValue().ipsToProbe.get(0)))) {

				if (task.params != null)
					ret = new ProbeSubTask(task.ipsToProbe.remove(0), task.sourceType, task.taskId, task.params, task.target);
				else
					ret = new ProbeSubTask(task.ipsToProbe.remove(0), task.sourceType, task.taskId, task.target);
				
				if (task.ipsToProbe.size() == 0)
					taskQueue.poll();
				
			}	
		}
		return ret;
	}
				
		
	/*
	 * Returns (removes) the top-most ProbeTask from the Queue 
	 */
	@SuppressWarnings("unused")
	private static Pair<Long, ProbeTask> popProbeTask( ProbeType pt ) {
	
		PriorityBlockingQueue<Pair<Long, ProbeTask>>taskQueue
		= probeTaskQueue.get(pt);

		return taskQueue.poll();
	}
	
	private static void launchNDTProbe(ProbeType pt) {

		ProbeSubTask pst = null;				
		updateProbeHistory();
		
		if (false)
			System.out.println("TALLIES: " + pt + " " + shortTermProbeTally.get(pt) + " " +
				hourlyProbeTally.get(pt));
		
		if ((shortTermProbeTally.get(pt) < maxProbePerMin.get(pt) * shortTermInterval) &&
				(hourlyProbeTally.get(pt) < maxProbePerHour.get(pt))) {
			
			/*TODO: add simultaneous probe support*/
			if (probeModules.get(pt).getState() == Thread.State.TIMED_WAITING) {

				pst = popProbeSubTask(pt);
				
				if (pst != null) {
					
					/*pause all probes*/
					probeModules.get(ProbeType.HTTP).setPause(true);
					probeModules.get(ProbeType.PING).setPause(true);
					probeModules.get(ProbeType.TRACEROUTE).setPause(true);
					probeModules.get(ProbeType.SCRAPE).setPause(true);
					probeModules.get(ProbeType.DNS).setPause(true);
					
					int i = 0;
					while ((probeModules.get(ProbeType.HTTP).getState() != Thread.State.TIMED_WAITING ||
							probeModules.get(ProbeType.PING).getState() != Thread.State.TIMED_WAITING ||
							probeModules.get(ProbeType.TRACEROUTE).getState() != Thread.State.TIMED_WAITING ||
							probeModules.get(ProbeType.SCRAPE).getState() != Thread.State.TIMED_WAITING ||
							probeModules.get(ProbeType.DNS).getState() != Thread.State.TIMED_WAITING) &&
							i++ < 10) {
						
						if (Main.DEBUG)
							System.out.println("Waiting for ProbeModules to pause...");
						
						try {
							Thread.sleep(2*1000);
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
					
					/*pause all torrents*/					
					if (pt.equals(ProbeType.NDT)) {
						if (Main.DEBUG)
							System.out.println("Temporarily stopping all (" + activeTorrents.size()+ ") active torrents");
						
						stopActiveTorrents();
					}
					
					probeModules.get(pt).probe(pst.ipAddress, pst.taskId);
					dispatchedEvents.put(
							new Pair<ProbeType, String>(pt, (pst.ipAddress==null)?"":pst.ipAddress), pst.sourceType);

					//appendToProbeHistory(pt, pst.ipAddress, System.currentTimeMillis());
					appendToTemporaryProbeHistory(pt, pst.ipAddress, System.currentTimeMillis());
					
					if (Main.DEBUG)
						System.out.println("assertToEngine(pst.sourceType, new FactFireAction(\"activeTest\", \"NDT\")");

					assertToEngine(pst.sourceType, new FactFireAction("activeTest", "NDT"));

				}
			}
		}
		else {
			if (Main.DEBUG) 
				System.out.println("WAIT: Too many probes of type " + pt + ", wait for some time... " +
						"ShortTerm: " + shortTermProbeTally.get(pt) + " hourly: " + hourlyProbeTally.get(pt));

		}
		
	}


	/* 
	 * Retrieves a probe of the specified *ProbeType* from the taskQueue 
	 * and, if one is available, passes the target *ip* to the 
	 * corresponding DasuModule for execution
	 */
	private static void launchProbe(ProbeType pt) {

		if (pt == null)
			return;
		
		if (pt.equals(ProbeType.NDT) || pt.equals(ProbeType.NDT2)) {
			launchNDTProbe(pt);
			return;
		}
		
		ProbeSubTask pst = null;				
		updateProbeHistory();
		
		if (false)
			System.out.println("TALLIES: " + pt + " " + shortTermProbeTally.get(pt) + " " +
				hourlyProbeTally.get(pt));
		
		if ((shortTermProbeTally.get(pt) < maxProbePerMin.get(pt) * shortTermInterval) &&
				(hourlyProbeTally.get(pt) < maxProbePerHour.get(pt))) {
			
			/*TODO: add simultaneous probe support*/
			if (probeModules.get(pt).getState() == Thread.State.TIMED_WAITING) {

				//pingTask = popProbeTask(pingTaskQueue);
				//dasuPingModule.pingAll(pingTask.getValue().ipsToProbe);

				pst = popProbeSubTask(pt);
				
				if (pst != null) {
					if (pt.equals(ProbeType.SCRAPE))
						((DasuScrapeModule) probeModules.get(pt)).probe(pst.ipAddress, pst.taskId, pst.params, pst.target);
					else if (pt.equals(ProbeType.DNS))
						((DasuDnsModule) probeModules.get(pt)).probe(pst.ipAddress, pst.taskId, pst.params, pst.target);
					else if ( (pt.equals(ProbeType.TRACEROUTE)) && (pst.sourceType.equals(SourceType.NULL)) )
						((DasuTraceRouteModuleV2) probeModules.get(pt)).probeWithLowPriority(pst.ipAddress, pst.taskId, pst.target);
					else
						probeModules.get(pt).probe(pst.ipAddress, pst.taskId, pst.target);
					
					dispatchedEvents.put(
							new Pair<ProbeType, String>(pt, (pst.ipAddress==null)?"":pst.ipAddress), pst.sourceType);
	
					appendToProbeHistory(pt, pst.ipAddress, System.currentTimeMillis());
				}
			}
		}
		else {
			if (Main.DEBUG) 
				System.out.println("WAIT: Too many probes of type " + pt + ", wait for some time... " +
						"ShortTerm: " + shortTermProbeTally.get(pt) + " hourly: " + hourlyProbeTally.get(pt));

		}
		
	}
		
	private void launchTest() {
		
		updateTestHistory();
		
		if (dispatchedTests.size() == 0) {
			
			TestTask test = testQueue.poll();

			if ((shortTermTestTally.get(test.testType) < maxTestPerMin.get(test.testType) * shortTermInterval) &&
					(hourlyTestTally.get(test.testType) < maxTestPerHour.get(test.testType))) {
			
				if (test.params != null) 
					//appendToTestHistory(test.testType, Integer.toString((Integer) test.params.peek()), System.currentTimeMillis());
					appendToTemporaryTestHistory(test.testType, Integer.toString((Integer) test.params.peek()), System.currentTimeMillis());
				else
					//appendToTestHistory(test.testType, null, System.currentTimeMillis());
					appendToTemporaryTestHistory(test.testType, null, System.currentTimeMillis());				

				
				switch(test.testType) {
				
					case AZUREUS_UPNP:
						if ((test.params != null))
							doUpnpTest( (Integer) test.params.pop());
						else
							doUpnpTest();
						break;
		
					case FORCE_SEEDING_THROUGH_PORT:
						if ((test.params != null) && (test.params.size() > 1)) 
							doForceSeedingThroughPort(
									(Integer) test.params.pop(), 
									(Integer) test.params.pop());
						else
							doForceSeedingThroughPort((Integer) test.params.pop());
						break;
		
					default:
						doSpeedTest(test.testType);
						break;
				
				}
	
				dispatchedTests.put(test.testType, test.sourceType);
			}
			
			/*return to queue*/
			else  {
				testQueue.add(test);
				
				if (Main.DEBUG) 
					System.out.println("WAIT: Too many tests of type " + test.testType + ", wait for some time... " +
							"ShortTerm: " + shortTermProbeTally.get(test.testType) + " hourly: " + hourlyProbeTally.get(test.testType));

			}
			
		}
		
	}
	
	public LinkedList<String> getProbeModules() {
		// Push list of probe module names to DasuData  
		LinkedList<String> l = new LinkedList<String>();
		for(DasuProbeModule pm: probeModules.values()) {
			l.add(pm.getName());
		}
		return l;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Runnable#run()
	 */

	public void run() {
		
		initialize();
		
		try {
			Thread.sleep(5*1000);
		} catch (InterruptedException e1) {
			e1.printStackTrace();
		}
		
		while (isActive()) {
			
			LOGGER.fine("hourlyProbeTally: " + hourlyProbeTally);
			LOGGER.fine("shortTermProbeTally: " + shortTermProbeTally);
			LOGGER.fine("shortTermProbeHistory: " + shortTermProbeHistory);
			LOGGER.fine("hourlyTestTally: " + hourlyTestTally);
			LOGGER.fine("shortTermTestTally: " + shortTermTestTally);
			LOGGER.fine("shortTermTestHistory: " + shortTermTestHistory);
				
			for ( ProbeType p : probeModules.keySet() )
				LOGGER.fine(p + " " + probeModules.get(p).getState());
			
			if (!dispatchedEvents.contains(ProbeType.NDT)) {
			
				if (testQueue.size() != 0)
					launchTest();
				
				if (!dispatchedSpeedTest()) {
					launchNDTProbe(ProbeType.NDT);

					if (!dispatchedEvents.contains(ProbeType.NDT)) {
						launchProbe(ProbeType.PING);
						launchProbe(ProbeType.TRACEROUTE);
						launchProbe(ProbeType.HTTP);
						launchProbe(ProbeType.SCRAPE);
						launchProbe(ProbeType.DNS);
					}
				}
			
				try {
					
					LOGGER.info("AMCOORDINATORV2: sleeping for " + sleepInterval + " seconds...");
					
					Pair<Long, EventType> event = null;					
					synchronized(events) {
												
						if (events.isEmpty())
							events.wait(sleepInterval*1000);
						else
							event = events.poll();
					}
					
					if (event != null) {
						if (event.getValue().equals(EventType.HALT))
							setActive(false);
						else {
							
							try {
								processEvent(event);
							}
							catch (Exception e) {
								e.printStackTrace();
							}
						}
					}
						
					
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
								
			}
		}
	
	}
	
	private void processEvent(Pair<Long, EventType> event) {
		
		switch(event.getValue()) {
			case HPM_RESULT:	
				processPing(event.getKey());
				break;
			case HTRM_RESULT:
				processTraceRoute(event.getKey());
				break;
			case HHM_RESULT:
				processHttp(event.getKey());
				break;
			case HNDT_RESULT:
				processNdt(event.getKey());
				break;
			case UPNP_RESULT:
				processUpnpTest(event.getKey());
				break;
			case SPEEDTEST_RESULT:
				processSpeedTest(event.getKey());
				break;
			case FORCED_SEEDING_THROUGH_PORT_RESULT:
				processForcedSeedingThroughPortTest(event.getKey());
				break;
			case RE_REQUEST:
				if (!dispatchedSpeedTest())
					processAMRequest(event.getKey());
				break;
			case SCRAPE_RESULT:
				processScrape(event.getKey());
				break;
			case DNS_RESULT:
				processDns(event.getKey());
				break;

		}
		
	}

	/* 
	 *  Only used to process measurements submitted by the Detection Engine.
	 *  The parameter eventId is not necessary in this context, but receive it anyway
	 *  in order to keep interface (API) consistent
	 */
	private void processAMRequest(long eventId) {

		ProbeType pt = REProbeTaskQueue.poll();
		launchProbe(pt);
		
	}
	
	private static RuleBasedEngine getEngine(SourceType source) {
		
		RuleBasedEngine eng = null;
		if (source.equals(SourceType.RE)) 
			eng = edu.northwestern.dasu.DasuManager.engine;
		else if (source.equals(SourceType.AME))
			eng = edu.northwestern.dasu.DasuManager.AMEngine;

		return eng;

	}
	
	public static void stop() {
		
		synchronized(events) {
			events.add(new Pair<Long, EventType>(0L, EventType.HALT));
			events.notify();
		}
		
		if (Main.DEBUG) 
			System.out.println("Stopping dasu probe modules");
		
		stopDasuAmModules();
		hourlyHistoryUpdaterTimer.destroy();
		
		self = null;
	}
	
	
	/********* SPECIAL OPERATIONS ********/
	
	// XXX: Main.DEBUGGING CODE -- should be deleted?
	public static void addTorrentDownload() {
	    
		try {
			
			DownloadManager dm = Main.getPluginInterface().getDownloadManager();
			//URL url = new URL("http://torrent.unix-ag.uni-kl.de/torrents/KNOPPIX_V6.2.1CD-2010-01-31-EN.torrent");
			URL url = new URL("http://dl.btjunkie.org/torrent/Redhat-Linux-Enterprise-5-Exclusive-DVD-ISO/40323d496af676894143463d5517b4d91f184964f068/download.torrent");
			TorrentManager tm = Main.getPluginInterface().getTorrentManager();
			TorrentDownloader tdlnd = tm.getURLDownloader(url);
			Torrent torrent = tdlnd.download();
			
			//Torrent torrent = tm.createFromBEncodedFile(torrent_file);
			dm.addDownload(torrent);
	
			Download dlds[] = dm.getDownloads();
			if (Main.DEBUG)
				for (Download d : dlds) {
					System.out.println(d.getName());					
				}
			
							
		} catch (DownloadException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (TorrentException e) {
			e.printStackTrace();
		}
	}
	
	// XXX: Main.DEBUGGING CODE -- should be deleted?
	@SuppressWarnings("deprecation")
	public static void doSpeedTest(final TestType testType) {

		try {
			Thread.sleep(10000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		System.setProperty("azureus.speed.test.challenge.jar.path", "/Users/Mario/Desktop/Azureus2.jar");
		String	explicit_path = System.getProperty( "azureus.speed.test.challenge.jar.path", null );
		
		if (Main.DEBUG)
			System.out.println(explicit_path);

        Thread t =
			new AEThread("SpeedTest Performer") 
			{
				public void 
				runSupport() 
				{
			        AzureusSpeedTest hst = new AzureusSpeedTest();
					hst.runTest(testType);
				}
			};
		
		t.setPriority(Thread.MIN_PRIORITY);
		t.setDaemon(true);
		t.start();

	}
	
	public static void doUpnpTest() {
		doUpnpTest(12345);
	}
	
	public static void doUpnpTest(int portNumber) {
		
		try {
			Thread.sleep(10000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		AzureusUpnpTest ant = new AzureusUpnpTest();
		ant.runTest(portNumber);

	}
	
	public static void doForceSeedingThroughPort(long port) {
		DasuPortThrottlingTest hptt = new DasuPortThrottlingTest(port);
		hptt.start();
	}

	public static void doForceSeedingThroughPort(long port, int sleepInterval) {
		DasuPortThrottlingTest hptt = new DasuPortThrottlingTest(port, sleepInterval);
		hptt.start();
	}
	
	public static void assertToEngine(SourceType sourceType,  FactTemplate fact) {
		
		if (Main.DEBUG)
			System.out.println("Asserting to engine " + sourceType + " fact " + fact);
		
		RuleBasedEngine eng = getEngine(sourceType);
		if (eng != null)
			eng.addFact(fact);
		
	}
	
	/************ PROCESS RESULTS FROM MEASUREMENT/TEST MODULES ***********/
	
	protected static void processUpnpTest(long eventId) {

		FactUpnpTestResult upnpTestResult = 
			(FactUpnpTestResult) testResultTable.remove(eventId);

		boolean result = upnpTestResult.success;
		SourceType sourceEngine = 
			dispatchedTests.remove(TestType.AZUREUS_UPNP);
					
		if (Main.DEBUG) {
			System.out.println(" ********* UPNP TEST RESULT: " + result + 
					" port: " + upnpTestResult.port + " ************");
		}

		/* if test was successful, commit to history, else rollback */
		if (upnpTestResult.error)
			rollbackFromTestHistory(upnpTestResult.testType);
		else
			commitToTestHistory(upnpTestResult.testType);

		
		assertToEngine(sourceEngine, upnpTestResult);

		Statistics.getInstance().addUpnpTestResult(upnpTestResult);
	}

	
	protected static void processSpeedTest(long eventId) {
	
		FactSpeedTestResult speedTestResult = 
			(FactSpeedTestResult) testResultTable.remove(eventId);

		boolean result = speedTestResult.success;
		SourceType sourceEngine = 
			dispatchedTests.remove(speedTestResult.testType);
							
		if (Main.DEBUG)
			System.out.println(" ********* SPEED TEST RESULT: " + result + 
				" dl: " + speedTestResult.downloadSpeed + " " + 
				" ul: " + speedTestResult.uploadSpeed + " " + 
				" time: " + new Date(speedTestResult.testTime) + " " + 
				" success: " + speedTestResult.success + "************");
		 
		
		/* if test was successful, commit to history, else rollback */
		if (!speedTestResult.isSuccess())
			rollbackFromTestHistory(speedTestResult.testType);
		else
			commitToTestHistory(speedTestResult.testType);
		
		assertToEngine(sourceEngine, speedTestResult);

		Statistics.getInstance().addSpeedTestResult(speedTestResult);

	}
		
	protected static void processForcedSeedingThroughPortTest (long eventId) {

		FactForcedSeedingThroughPortTestResult  testResult = 
			(FactForcedSeedingThroughPortTestResult ) testResultTable.remove(eventId);

		boolean result = testResult.success;
		SourceType sourceEngine = 
			dispatchedTests.remove(TestType.FORCE_SEEDING_THROUGH_PORT);
					
		if (Main.DEBUG) {
			System.out.println(" ********* Seeding through port TEST RESULT: " 
					+ result + " ************");
		}

		assertToEngine(sourceEngine, testResult);
			
		//Statistics.getInstance().addUpnpTestResult(upnpTestResult);
	}
	
	
	/* 
	 * Set of functions specific to each DasuProbeModule
	 * used to process results from the corresponding ProbeModule
	 */
	
	private static void processPing(long eventId) {

		DasuPingModuleV2 dasuPingModule = (DasuPingModuleV2) probeModules.get(ProbeType.PING);
		FactPingProbeResult pingResult = 
			dasuPingModule.pingResultTable.remove(eventId);

		if ( pingResult != null ) {
			
			String ipAddressResponse = pingResult.getIp();
						
			SourceType sourceEngine =
				dispatchedEvents.remove(
					new Pair<ProbeType, String>(ProbeType.PING, ipAddressResponse));
	
			assertToEngine(sourceEngine, pingResult);
		}
		
	}
	
	private static void processHttp(long eventId) {

		DasuHttpModule dasuHttpModule = (DasuHttpModule) probeModules.get(ProbeType.HTTP);
		FactHttpProbeResult httpProbeResult = 
			dasuHttpModule.httpResultTable.remove(eventId);

		String url = httpProbeResult.url;					
		SourceType sourceEngine =
			dispatchedEvents.remove(
				new Pair<ProbeType, String>(ProbeType.HTTP, url));

		assertToEngine(sourceEngine, httpProbeResult);
		
	}

	private static void processDns(long eventId) {
		
		DasuDnsModule dasuDnsModule = (DasuDnsModule) probeModules.get(ProbeType.DNS);
		FactDnsProbeResult dnsProbeResult = 
			dasuDnsModule.dnsResultTable.remove(eventId);
		
		String toLookup = dnsProbeResult.getToLookup();					
		SourceType sourceEngine =
			dispatchedEvents.remove(
				new Pair<ProbeType, String>(ProbeType.DNS, toLookup));

		assertToEngine(sourceEngine, dnsProbeResult);
	}
	
	private static void processScrape(long eventId) {
		
		DasuScrapeModule dasuScrapepModule = (DasuScrapeModule) probeModules.get(ProbeType.SCRAPE);
		FactScrapeProbeResult scrapeProbeResult = 
			dasuScrapepModule.scrapeResultTable.remove(eventId);
		
		String hash = scrapeProbeResult.hash;					
		SourceType sourceEngine =
			dispatchedEvents.remove(
				new Pair<ProbeType, String>(ProbeType.SCRAPE, hash));

		assertToEngine(sourceEngine, scrapeProbeResult);
	}
	
	private static void processTraceRoute(long eventId) {

		DasuTraceRouteModuleV2 dasuTraceRouteModule = 
			(DasuTraceRouteModuleV2) probeModules.get(ProbeType.TRACEROUTE);
		FactTraceResult trResult = 
			dasuTraceRouteModule.trResultTable.remove(eventId);

		String ipAddressResponse = trResult.dest;
					
		SourceType sourceEngine =
			dispatchedEvents.remove(
				new Pair<ProbeType, String>(ProbeType.TRACEROUTE, ipAddressResponse));

		assertToEngine(sourceEngine, trResult);
		
	}
	
	private static void processNdt(long eventId) {

		if (Main.DEBUG)
			System.out.println("processNdt");
		
		DasuNdtModule dasuNdtModule = (DasuNdtModule) probeModules.get(ProbeType.NDT);
		FactNdtProbeResult ndtResult = 
			dasuNdtModule.ndtResultTable.remove(eventId);

		String server = ndtResult.server;

		if ( (!dispatchedEvents.containsKey(new Pair<ProbeType, String>(ProbeType.NDT, server))) &&
			(!dispatchedEvents.containsKey(new Pair<ProbeType, String>(ProbeType.NDT2, server))) ) {
			server = "";
		}
		
		ProbeType pt = ProbeType.NDT;
		if (!dispatchedEvents.containsKey(new Pair<ProbeType, String>(pt, server)))
			pt = ProbeType.NDT2;
		
		SourceType sourceEngine =
			dispatchedEvents.remove(
				new Pair<ProbeType, String>(pt, server));
		
		
		if (Main.DEBUG)
			System.out.println("temporaryShortTermProbeHistory: " + temporaryShortTermProbeHistory);
		
		/* if test was successful, commit to history, else rollback */
		if (ndtResult.failed) {
			if (!rollbackFromProbeHistory(pt, ndtResult.server))
				rollbackFromProbeHistory(pt, null);
		}
		else {
			if (!commitToProbeHistory(pt, ndtResult.server))
				commitToProbeHistory(pt, null);
		}
		
		if (Main.DEBUG)
			System.out.println("temporaryShortTermProbeHistory: " + temporaryShortTermProbeHistory);
		
		assertToEngine(sourceEngine, ndtResult);
		
		/*restart all probes*/
		probeModules.get(ProbeType.HTTP).setPause(false);
		probeModules.get(ProbeType.PING).setPause(false);
		probeModules.get(ProbeType.TRACEROUTE).setPause(false);
		probeModules.get(ProbeType.SCRAPE).setPause(false);
		probeModules.get(ProbeType.DNS).setPause(false);
		
		/*restart all torrents*/
		if (pt.equals(ProbeType.NDT)) {
			restartActiveTorrents();
		}
				
	}
	
	public static String[] getPropertiesFromServer(String url) {
		HashMap<String, String> map = Main.getPropertiesFromServer(url);
		return (String[]) map.values().toArray(new String[map.size()]);
	}

	
	/****** CRONJOB SCHEDULER INTERFACE ********/
	
	public static void removeCronJob(String id, Integer[] hour, 
			Integer[] day, String commandToExecute) {
		removeCronJob(id, hour, day, commandToExecute, CronjobType.CUSTOM);
	}

	public static void removeCronJob(String id, String commandToExecute, CronjobType cronjobType) {
		removeCronJob(id, new Integer[]{-1}, new Integer[]{-1}, commandToExecute, cronjobType);
	}
	
	private static void removeCronJob(String id, Integer[] hour, 
			Integer[] day, String commandToExecute, CronjobType cronjobType) {
		
		ArrayList<Integer> hours = new ArrayList<Integer>(Arrays.asList(hour));
		ArrayList<Integer> days = new ArrayList<Integer>(Arrays.asList(day));

		if ((hours == null) || (days == null))
			return;
		
		if ((hours.size() == 1) && (hours.contains(-1))) {
			hours.clear();
			for (int i=0; i < 24; i++) {
				hours.add(i);
			}
		}
		
		if ((days.size() == 1) && (days.contains(-1))) {
			days.clear();
			for (int i=0; i < 7; i++) {
				days.add(i);
			}
		}
		
		CronjobEntry job = 	new CronjobEntry(id, new HashSet<Integer>(hours), 
				new HashSet<Integer>(days),commandToExecute, cronjobType);
 			
		DasuCronScheduler.getInstance().removeCronJob(job);
	}

	public static void addCronJob(String id, Integer[] hour, 
			Integer[] day, String commandToExecute) {
		addCronJob(id, hour, day, commandToExecute, CronjobType.CUSTOM);
	}

	public static void addCronJob(String id, String commandToExecute, CronjobType cronjobType) {
		addCronJob(id, new Integer[]{-1}, new Integer[]{-1}, commandToExecute, cronjobType);
	}

	private static void addCronJob(String id, Integer[] hour, 
			Integer[] day, String commandToExecute, CronjobType cronjobType) {
		
		ArrayList<Integer> hours = new ArrayList<Integer>(Arrays.asList(hour));
		ArrayList<Integer> days = new ArrayList<Integer>(Arrays.asList(day));

		if ((hours == null) || (days == null))
			return;
		
		if ((hours.size() == 1) && (hours.contains(-1))) {
			hours.clear();
			for (int i=0; i < 24; i++) {
				hours.add(i);
			}
		}
		
		if ((days.size() == 1) && (days.contains(-1))) {
			days.clear();
			for (int i=0; i < 7; i++) {
				days.add(i);
			}
		}
		
		CronjobEntry job = 	new CronjobEntry(id, new HashSet<Integer>(hours), 
				new HashSet<Integer>(days),commandToExecute, cronjobType);
 			
		DasuCronScheduler.getInstance().registerCronJob(job);
	}
	
	private static void launchHourlyProbeHistoryUpdater() {
	
		hourlyHistoryUpdaterTimer = Main.createTimer("AMCHourlyProbeHistoryUpdaterTimer");
		hourlyHistoryUpdaterTimer.addPeriodicEvent(60*60*1000, 
				new ITimerEventPerformer(){
	
					public void perform(ITimerEvent event) {
						
						for (ProbeType pt : hourlyProbeTally.keySet())
							hourlyProbeTally.put(pt, 0);
						for (TestType tt : hourlyTestTally.keySet())
							hourlyTestTally.put(tt, 0);
					}
		});
		
	}
	
	
	/* clean shortTerm history if necessary
	 * It only iterates over the oldest entries
	 * from the top, if no entries qualify, stop
	 */

	private static void updateProbeHistory() {
	
		synchronized(shortTermProbeHistory) {

			Long currTime = System.currentTimeMillis();
			Iterator<Entry<Pair<ProbeType, String>, Long>> iter = shortTermProbeHistory.entrySet().iterator();
			while (iter.hasNext()) {
				Entry<Pair<ProbeType, String>, Long> entry = iter.next();
				ProbeType probetype = entry.getKey().getKey();
				if (currTime - entry.getValue() > shortTermInterval*1000*60) {
					iter.remove();
					if (shortTermProbeTally.get(probetype) > 0)
						shortTermProbeTally.put(probetype, shortTermProbeTally.get(probetype)-1);	
				}
				else {
					break;
				}
			}
		}
	}
		
	private static void updateTestHistory() {
		
		synchronized(shortTermTestHistory) {

			Long currTime = System.currentTimeMillis();
			Iterator<Entry<Pair<TestType, String>, Long>> iter = shortTermTestHistory.entrySet().iterator();
			while (iter.hasNext()) {
				Entry<Pair<TestType, String>, Long> entry = iter.next();
				TestType testType = entry.getKey().getKey();
				
				if (Main.DEBUG)
					System.out.println(testType + " " + (currTime - entry.getValue())/(1000*1000));
				
				if (currTime - entry.getValue() > shortTermInterval*1000*60) {
					iter.remove();
					shortTermTestTally.put(testType, shortTermTestTally.get(testType)-1);	
				}
				else {
					break;
				}
			}
		}
	}
	
	private static void appendToProbeHistory(ProbeType pt, String ipAddress, long timeStamp) {
		
		/* then add to queue */
		synchronized(shortTermProbeHistory) {
			shortTermProbeHistory.put(new Pair<ProbeType, String>(pt, ipAddress), System.currentTimeMillis());
		}
		shortTermProbeTally.put(pt, shortTermProbeTally.get(pt)+1);
		hourlyProbeTally.put(pt, hourlyProbeTally.get(pt)+1);
			
	}
	
	private static void appendToTemporaryProbeHistory(ProbeType pt, String ipAddress, long timeStamp) {
		
		synchronized(temporaryShortTermProbeHistory) {
			temporaryShortTermProbeHistory.put(new Pair<ProbeType, String>(pt, ipAddress), System.currentTimeMillis());
		}
			
	}
	
	private static boolean rollbackFromProbeHistory(ProbeType pt, String ipAddress) {
		
		Long timeStamp = null; 
		synchronized(temporaryShortTermProbeHistory) {		
			timeStamp = temporaryShortTermProbeHistory.remove(new Pair<ProbeType, String>(pt, ipAddress));
			
		}
		
		return (timeStamp==null)?false:true;
	}

	private static boolean commitToProbeHistory(ProbeType pt, String ipAddress) {
		
		Long timeStamp = null;
		synchronized(temporaryShortTermProbeHistory) {
			timeStamp = temporaryShortTermProbeHistory.remove(new Pair<ProbeType, String>(pt, ipAddress));
		}
		
		if (timeStamp != null) {
			appendToProbeHistory(pt, ipAddress, timeStamp);
			return true;
		}
		
		return false;
		
	}
	
	private static void appendToTestHistory(TestType tt, String val, long timeSpan) {
		
		/* then add to queue */
		synchronized(shortTermTestHistory) {
			shortTermTestHistory.put(new Pair<TestType, String>(tt, val), timeSpan);
		}
		shortTermTestTally.put(tt, shortTermTestTally.get(tt)+1);
		hourlyTestTally.put(tt, hourlyTestTally.get(tt)+1);
			
	}
	
	private static void appendToTemporaryTestHistory(TestType tt, String val, long timeStamp) {
		
		synchronized(temporaryShortTermTestHistory) {
			temporaryShortTermTestHistory.put(tt, new Pair<String, Long>(val, timeStamp));
		}
			
	}
	
	private static boolean rollbackFromTestHistory(TestType tt) {
		
		Pair<String, Long> values = null; 
		synchronized(temporaryShortTermTestHistory) {		
			values = temporaryShortTermTestHistory.remove(tt);
		}
		
		return (values==null)?false:true;
	}
	
	private static boolean commitToTestHistory(TestType tt) {
		
		Pair<String, Long> values = null; 
		synchronized(temporaryShortTermTestHistory) {
			values = temporaryShortTermTestHistory.remove(tt);
		}
		
		if (values != null) {
			appendToTestHistory(tt, values.getKey(), values.getValue());
			return true;
		}
		
		return false;
		
	}
	
	private static boolean dispatchedSpeedTest() {
		
		return ( (dispatchedTests.size() != 0) && 
				(dispatchedTests.contains(TestType.AZUREUS_DL) ||
						dispatchedTests.contains(TestType.AZUREUS_UL) ||
						dispatchedTests.contains(TestType.AZUREUS_DL_ENCRYPTION) ||
						dispatchedTests.contains(TestType.AZUREUS_UL_ENCRYPTION) ) ); 

	}

	public synchronized static void stopActiveTorrents() {
		
		for (Download d : Main.getPluginInterface().getDownloadManager().getDownloads()) {
			
			if (d.getState() != Download.ST_STOPPED && d.getState() != Download.ST_STOPPING) {
				
				try {
					d.stop();
					activeTorrents.add(d);
				} catch (DownloadException e) {
					e.printStackTrace();
				}
			}
		}
		
		if (activeTorrents.size() > 0)
			Main.logPlugin("Temporarily stopping all (" + activeTorrents.size()+ ") active torrents");
		
	}
	
	public static ArrayList<ArrayList<String>> getMyDnsServers() {
		return ((DasuDnsModule) probeModules.get(ProbeType.DNS)).getMyDnsServers();
	}
	
	public static String getMyPublicIp() {
		return Main.getPublicIpAddress();
	}
	
	public synchronized static void restartActiveTorrents() {

		if (Main.DEBUG)
			System.out.println("sendToLog(\"Restarting all (\" + activeTorrents.size()+ \") active torrents\"))");
		
		if (activeTorrents.size() > 0)
			Main.logPlugin("Restarting all (" + activeTorrents.size()+ ") active torrents");
		
		Iterator<Download> iter = activeTorrents.iterator();
		while (iter.hasNext()) {
			try {
				Download d = iter.next();
				d.restart();
				iter.remove();
			} catch (DownloadException e) {
				e.printStackTrace();
			}
		}
	}		

	/******************
	/** sessionSamples
	/******************/
	
	public static Hashtable<String, SessionSignalHolder> getSessionSample() {
		return sessionSamples;
	}
	
	public static boolean sessionSampleExists(String name) {
		return sessionSamples.containsKey(name);
	}
	
	public static boolean sessionSampleCreate(String name, int size) {			
		
		if (!sessionSampleExists(name)) {
			
			try {
				sessionSamples.put(name, new SessionSignalHolder(size));
			} catch(Exception e) {
				e.printStackTrace();				
				return false;
			}
		}
		
		if (Main.DEBUG)
			sessionSamplePrint(name);
		
		return true;
			
	}
	
	public static boolean sessionSampleDestroy(String name) {		
		
		if (sessionSampleExists(name))
			sessionSamples.remove(name);		
		return true;
	}
	
	public static boolean sessionSampleSetValue(String name, Object value) {
		
		if (sessionSampleExists(name)) {
			sessionSamples.get(name).set(value);
			return true;
		}		
		return false;
	}
	
	public static void sessionSamplePrint(String name) {
		
		if (sessionSampleExists(name)) {
			System.out.println(sessionSamples.get(name));
		}
		
	}
	
	public static Collection<Object> sessionSampleGetValues(String name) {

		if (sessionSampleExists(name)) {
			return sessionSamples.get(name).getValues();
		}

		return null;
	}
	
	public static Collection<Long> sessionSampleGetTimeStamps(String name) {

		if (historicSampleExists(name)) {
			return sessionSamples.get(name).getTime();
		}

		return null;
	}
	
	/******************
	/** historicSamples
	/******************/

	public static Hashtable<String, HistoricSignalHolder> getHistoricSample() {
		return historicSamples;
	}
	
	public static boolean historicSampleExists(String name) {
		return historicSamples.containsKey(name);
	}
	
	public static boolean historicSampleCreate(String name, TimeType timeSpanType, 
			TimeType intervalType, int intervalLength) {
		
		if (!historicSampleExists(name) && (historicSamples.size() < DasuConfiguration.getInstance().getMaxLogEntries())) {
			
			try {
				historicSamples.put(name, new HistoricSignalHolder(name, timeSpanType, intervalType, intervalLength));
			} catch(Exception e) {
				e.printStackTrace();				
				return false;
			}
		}
		
		if (Main.DEBUG)
			historicSamplePrint(name);
		
		return true;
			
	}
	
	public static boolean historicSampleDestroy(String name) {
		
		if (historicSampleExists(name))
			historicSamples.remove(name);
		
		return true;

	}
	
	public static float historicSampleGetValue(String name, String signalName, long timeRecorded) {
		
		if (historicSampleExists(name))
			return historicSamples.get(name).get(signalName, timeRecorded);
		
		return -1;
	}

	public static long historicSampleGetTimeStamp(String name, long timeRecorded) {
		
		if (historicSampleExists(name))
			return historicSamples.get(name).getTimeStamp(timeRecorded);
		
		return -1;
	}
	
	public static boolean historicSampleSetValue(String name, String signalName, long timeRecorded, float value, String unit) {
		
		if (historicSampleExists(name)) {
			historicSamples.get(name).set(signalName, timeRecorded, value, unit);
			return true;
		}
		
		return false;
	}
	
	public static void historicSamplePrint(String name) {
		
		if (historicSampleExists(name)) {
			System.out.println(historicSamples.get(name));
		}
		
	}
	
	public static Collection<Float> historicSampleGetValues(String name, String signalName) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getValuesHash(signalName).values();
		}

		return null;
	}
	
	public static Collection<Long> historicSampleGetTimeStamps(String name) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getTimeStampsHash().values();
		}
		
		return null;
	}
	
	public static Collection<Float> historicSampleGetValuesSubsetPast(String name, String signalName, long timeRecorded, 
			int numIntervals) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getValuesSubsetPast(signalName, timeRecorded, numIntervals);
		}

		return null;
	}
	
	public static Collection<Long> historicSampleGetTimeStampsSubsetPast(String name, String signalName, long timeRecorded, 
			int numIntervals) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getTimeStampSubsetPast(timeRecorded, numIntervals);
		}

		return null;
	}
	
	/******************/
		
	public static void commitInterference(String type, Long timestamp) {
		Statistics.getInstance().commitInterference(type, timestamp);
	}
	
	public static void commitFeedback(String type, String name, String email, String message, String dump) {
		Main.getReporter().insertFeedback(type, name, email, message, dump, System.currentTimeMillis(),
				Util.currentGMTTime());		
	}
	
	public static void commitUserReports(String dids, String comment) {
		Statistics.getInstance().commitUserComments(dids, comment);		
	}
	
	public static void commitIspInfo(String prefix, String ispname) {
		Statistics.getInstance().commitIspInfo(prefix, ispname);	
	}

	public static void commitInterferenceEvent(String type, long timestamp, long downloadId, long droppedTo, float mean) {
		Statistics.getInstance().commitInterferenceEvent(type, (long) timestamp, (long) downloadId, (long) droppedTo, mean);
	}
	
	public static String getNewRandomDasuUrl() {
		return DasuManager.getInstance().getNewRandomDasuUrl();
	}
	
	public static float getPluginVersion() {
		return Float.parseFloat(Main.getPluginInterface().getPluginVersion());
	}
	
	public static boolean isActive() {
		return active;
	}

	public static boolean isPause() {
		return pause;
	}

	public static void setActive(boolean active) {
		AMCoordinatorV2.active = active;
	}
	
	public static void addPopup(String msg) {
		DasuMessageManager.getInstance().addMessage("NA", msg, -1, true);
	}
	
	public static void addPopup(String msg, String id, int timeout, boolean repeatable) {
		DasuMessageManager.getInstance().addMessage(id, msg, timeout, repeatable);
	}


	
}
