/**
 * Dasu Project
 *
 * File:         DasuHttpModule.java
 * RCS:          $Id$
 * Description:  DasuHttpModule class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 15, 2010 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;

import java.io.IOException;
import java.net.InetAddress;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.xbill.DNS.DClass;
import org.xbill.DNS.ExtendedFlags;
import org.xbill.DNS.Message;
import org.xbill.DNS.Name;
import org.xbill.DNS.Record;
import org.xbill.DNS.ReverseMap;
import org.xbill.DNS.Section;
import org.xbill.DNS.SimpleResolver;
import org.xbill.DNS.TSIG;
import org.xbill.DNS.TextParseException;
import org.xbill.DNS.Type; 

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactDidEvent;
import edu.northwestern.dasu.drools.FactDnsProbeResult;
import edu.northwestern.dasu.drools.FactDnsTimeout;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuScrapeModule class manages the set of 
 *         Scrapes performed by the peer.
 */

public class DasuDnsModule extends DasuProbeModule {	
		
	private class TaskEntry {
		
		public Integer taskId;
		public String toLookup;
		public String server = null;
		public String options;
		public String target;
		
		public TaskEntry(String toLookup, Integer taskId, Map params, String target) {
			this.taskId = taskId;
			this.toLookup = toLookup;
			this.target = target;
						
			if (params != null) {
				
				if (params.containsKey("server"))
					this.server = (String) params.get("server");
				
				if (params.containsKey("options"))
					this.options = (String) params.get("options");				

			}
			
		}
	}
	
	private final static Logger LOGGER = Logger.getLogger(DasuDnsModule.class.getName());
	private static boolean DEBUG = false;
	
	ArrayList<ArrayList<String>> myDns =  null;
	public Hashtable<Long, FactDnsProbeResult> dnsResultTable;
	private ArrayList<TaskEntry> toDoLookupList = new ArrayList<TaskEntry>();
	private static DasuDnsModule self;
	private Hashtable<String, Long> timeoutIsRunning;
	
	public DasuDnsModule() {
		super("DasuDnsModule");
		this.pi = Main.getPluginInterface();
		dnsResultTable = new Hashtable<Long, FactDnsProbeResult>();
		timeoutIsRunning = new Hashtable<String, Long>();

		active = true;
	}
	
	public synchronized static DasuDnsModule getInstance() {
		if (self == null) {
			self = new DasuDnsModule();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}

	@Override
	public void run() {
		
	  while (!Main.isShuttingDown() && isActive()) {
		 
			TaskEntry taskEntry = null;
			synchronized (toDoLookupList) {
				while ((toDoLookupList.size()==0) && !Main.isShuttingDown()) {
					try {
						toDoLookupList.wait(sleepInterval*1000);
					} catch (InterruptedException e) {
						self = null;
						return;
					}
				}
				
				if (toDoLookupList.size() > 0)
					taskEntry = toDoLookupList.remove(0);
				
				LOGGER.fine("toLookup " + taskEntry.toLookup);
			}
			
			if (taskEntry.toLookup.equals("HALT")) 
				return;			
			
			probeIp(taskEntry.toLookup, taskEntry.server, taskEntry.options, taskEntry.taskId, taskEntry.target);


	  } // end while active
		
	  self = null;
	  
	}
		
	@Override
	public void probeAll(List<String> list, Integer taskId, String target) {

		synchronized (toDoLookupList) {
			for (String toLookup : list)
				toDoLookupList.add(new TaskEntry(toLookup, taskId, null, target));
			toDoLookupList.notifyAll();
		}
	}
	
	/*default probe uses default DNS server*/
	
	@Override
	public void probeWithPriority(String toLookup, Integer taskId, String target) {
		
		synchronized (toDoLookupList) {
			toDoLookupList.add(new TaskEntry(toLookup, taskId, null, target));
			toDoLookupList.notifyAll();
		}
	}

	public void probe(String toLookup, Integer taskId, Map params, String target) {
		
		LOGGER.fine("probe() " + toLookup + " " + taskId + " " + params + " " + target);
		
		synchronized (toDoLookupList) {
			toDoLookupList.add(new TaskEntry(toLookup, taskId, params, target));
			toDoLookupList.notifyAll();
		}

	}
	
	/*default probe uses default DNS server*/
	
	@Override
	public void probe(String toLookup, Integer taskId, String target) {

		synchronized (toDoLookupList) {
			toDoLookupList.add(new TaskEntry(toLookup, taskId, null, target));
			toDoLookupList.notifyAll();
		}
	}
	
	/*default probe uses default DNS server*/
	
	@Override
	protected Object probeIp(String toLookup, Integer taskId, String target) {
		
		return probeIp(toLookup, null, null, taskId, target);
		
	}
	
	public String getType(String options) {
		
		String type = "A";
		
		if (options != null) {
			int arg = 0;
			String[] argv = options.split(" ");
			while (argv.length > arg && argv[arg].startsWith("-") && (argv[arg].length() > 1)) {
	
				switch (argv[arg].charAt(1)) {
	
				case 't':
					
					String typeString;
					
					if (argv[arg].length() > 2) {
						typeString = argv[arg].substring(2);
					} else {
						typeString = argv[++arg];
					}
	
					type = typeString;
					
					break;
				
				case 'x':
					
					type = "PTR";
					break;
				}
				
				arg++;
			}
		}
		
		return type;
		
	}
	
	protected Object probeIp(String toLookup, String server, String options, Integer taskId, String target) {
	
		Long requestTime = Util.currentGMTTime();
		Message msg = doLookup(toLookup, server, options, taskId);
		FactDnsProbeResult ret = null;
		
		LOGGER.fine("msg: " + msg);

		try {
			if (( msg != null ) && ( msg.getHeader().getID() != -1)) {
				
				ret = new FactDnsProbeResult(server, toLookup, options,
						requestTime, msg, Util.currentGMTTime(), getType(options), taskId);
				ret.setTarget(target);
				if ( ret.getServer() == null ) {
					
					org.xbill.DNS.ResolverConfig DNSConfig = new org.xbill.DNS.ResolverConfig();
					server = DNSConfig.server();
					FactDnsProbeResult fact =  new FactDnsProbeResult(DNSConfig.server(), toLookup, options,
							requestTime, msg, Util.currentGMTTime(), getType(options), taskId);
					Statistics.getInstance().addDnsResult(fact);

				}
				else 
					Statistics.getInstance().addDnsResult(ret);
				
				
			}
			else {
				
				ret = new FactDnsProbeResult(server, toLookup, options,
						requestTime, new Message(), Util.currentGMTTime(), getType(options), taskId);		
				ret.setFailed(true);		
				ret.setTarget(target);
								
				if ( msg != null &&  msg.getHeader().getID() == -1)
					ret.setTimeout(true);
				
			}
		}

		catch (Exception e) {
			e.printStackTrace();

			ret = new FactDnsProbeResult(server, toLookup, options,
					requestTime, new Message(), Util.currentGMTTime(), getType(options), taskId);		
			ret.setFailed(true);
			ret.setTarget(target);
			
			if ( msg.getHeader().getID() == -1)
				ret.setTimeout(true);			
		}
		
		updateLookupResultQueue(ret);
		return null;

	}
	
	// TODO: merge with Util.doLookup.
	// JSO: not sure how this code differs from Util.doLookup, but a diff shows that it does.
	public Message doLookup(String nameString, String server, String args, Integer taskId) {

		String[] argv = new String[]{};
		if (args != null)
			argv = args.split(" ");
		
		org.xbill.DNS.Name name = null;
		int type = org.xbill.DNS.Type.A;
		int dclass = DClass.IN;
		int arg;
		Message query;
		Message response = null;
		Record rec;
		SimpleResolver res = null;
		boolean printQuery = false;
		
		if ( server == null ) {
			org.xbill.DNS.ResolverConfig DNSConfig = new org.xbill.DNS.ResolverConfig();
			server = DNSConfig.server();
		}
		
		if (timeoutIsRunning.containsKey(server))
			return new Message(-1);

		//server = "4.4.4.2";
		
		if (!Statistics.dnsMap.containsKey(server))
			Statistics.dnsMap.put(server, Statistics.dnsMap.size());		
		
		try {
			
			res = new SimpleResolver(server);
			
			arg = 0;
			name = Name.fromString(nameString, Name.root);

			if ( argv.length > 0) {
				
				while (argv.length > arg && argv[arg].startsWith("-") && (argv[arg].length() > 1)) {
					
					switch (argv[arg].charAt(1)) {

					case 'x':
						name = ReverseMap.fromAddress(nameString);
						type = Type.PTR;
						dclass = DClass.IN;
						break;
						
					case 'p':
	
						String portStr;
						int port;
	
						if (argv[arg].length() > 2) {
							portStr = argv[arg].substring(2);
						} else {
							portStr = argv[++arg];
						}
	
						port = Integer.parseInt(portStr);
	
						if ((port < 0) || (port > 65536)) {
							System.out.println("Invalid port");
	
							return null;
						}
	
						res.setPort(port);
	
						break;
	
					case 'b':
	
						String addrStr;
	
						if (argv[arg].length() > 2) {
							addrStr = argv[arg].substring(2);
						} else {
							addrStr = argv[++arg];
						}
	
						InetAddress addr;
	
						try {
							addr = InetAddress.getByName(addrStr);
						} catch (Exception e) {
							System.out.println("Invalid address");
	
							return null;
						}
	
						res.setLocalAddress(addr);
	
						break;
	
					case 'k':
	
						String key;
	
						if (argv[arg].length() > 2) {
							key = argv[arg].substring(2);
						} else {
							key = argv[++arg];
						}
	
						res.setTSIGKey(TSIG.fromString(key));
	
						break;
	
					case 'T':
	
						res.setTCP(true);
						break;
						
					case 't':
						
						String typeString;
						
						if (argv[arg].length() > 2) {
							typeString = argv[arg].substring(2);
						} else {
							typeString = argv[++arg];
						}
	
						type = Type.value(typeString);
						
						break;
	
					case 'i':
						res.setIgnoreTruncation(true);
	
						break;
	
					case 'e':
	
						String ednsStr;
						int edns;
	
						if (argv[arg].length() > 2) {
							ednsStr = argv[arg].substring(2);
						} else {
							ednsStr = argv[++arg];
						}
	
						edns = Integer.parseInt(ednsStr);
	
						if ((edns < 0) || (edns > 1)) {
							System.out.println("Unsupported " + "EDNS level: "
									+ edns);
	
							return null;
						}
	
						res.setEDNS(edns);
	
						break;
	
					case 'd':
						res.setEDNS(0, 0, ExtendedFlags.DO, null);
	
						break;
	
					case 'q':
						printQuery = true;
	
						break;
	
					default:
						System.out.print("Invalid option: ");
						System.out.println(argv[arg]);
					}
	
					arg++;
				}
			}
			
			if (name == null) {
				return null;
			}
			
			rec = Record.newRecord(name, type, dclass);
			query = Message.newQuery(rec);

			if (printQuery) {
				System.out.println(query);
			}

			response = res.send(query);
			
		} catch (ArrayIndexOutOfBoundsException e) {
			LOGGER.severe(e.toString());
			return null;
		} catch (TextParseException e) {
			LOGGER.severe(e.toString());
			return null;
		} catch (IOException e) {
			
			if (e instanceof SocketTimeoutException) {
				
				if (DasuConfiguration.getInstance().getDnsTimeout() != 0) {
					
					 launchDnsTimeout(server);
				}
			}
			
			LOGGER.fine(e.toString());
			return null;
		}

		return response;		
	}
	
	public ArrayList<ArrayList<String>> getMyDnsServers() {		
		return getMyDns(Main.getPublicIpAddress());
	}
	
	private ArrayList<ArrayList<String>> getMyDns(String myIp) {
		
		if (myDns != null)
			return myDns;
		
		String publicDnsIp, publicDnsIpAna;
		int publicDnsId, publicDnsIdAna, dnsId;
		myDns = new ArrayList<ArrayList<String>>();	
		org.xbill.DNS.ResolverConfig DNSConfig = new org.xbill.DNS.ResolverConfig();
		int order = 0;
		String type = "";
	
		for (String n : DNSConfig.servers()) {

			order++;
			publicDnsIp = null;
			publicDnsIpAna = null;
			publicDnsId = -1;
			publicDnsIdAna = -1;
			
			//dnsId = Main.getReporter().makeId("dnsips", "ip", n);
			dnsId = Statistics.getInstance().getDnsId(n);
						
			if (DasuConfiguration.getInstance().isDoAnaLookup()) {
				 publicDnsIpAna = getPublicDnsIpAna(n, myIp);
				 type = "A";
			}
			else {
				publicDnsIp = getMyPublicDnsIp( n, myIp );
				type = "P";
			}

			LOGGER.fine("DNSIPS: " + n + " " + publicDnsIpAna + " " + order);
			
			if ( (!publicDnsIpAna.equals(n) || (publicDnsIpAna.equals(n) && (!isPrivateIp(n)))) ) {
				publicDnsIdAna = Statistics.getInstance().getDnsId(publicDnsIpAna);
				Main.getReporter().insertDnsMapping(Statistics.getInstance().getPeerId(myIp), publicDnsIdAna, dnsId, "A", order, Util.currentGMTTime());

				//publicDnsIdAna = Main.getReporter().makeId("dnsips", "ip", publicDnsIpAna);
//				if (!Statistics.dnsMap.containsKey(publicDnsIpAna)) {
//					publicDnsIdAna = Main.getReporter().makeId("dnsips", "ip", publicDnsIpAna);						
//					if (publicDnsIdAna == -1)
//						Main.getReporter().insertDnsMapping(Statistics.getInstance().getPeerId(myIp), publicDnsIdAna, dnsId, "A", order);
//					
//					//pingDnsServers(publicDnsIpAna);
//					//traceRouteDnsServers(publicDnsIpAna);				
//					Statistics.dnsMap.put(publicDnsIpAna, Statistics.dnsMap.size());
//				}

			}

			else {
				
			  if ( isPrivateIp(n)) {
				  publicDnsId = Statistics.getInstance().getDnsId(publicDnsIp);
				  Main.getReporter().insertDnsMapping(Statistics.getInstance().getPeerId(myIp), publicDnsId, dnsId, "P", order, Util.currentGMTTime());

//				publicDnsId = Main.getReporter().makeId("dnsips", "ip", publicDnsIp);
//				if (!Statistics.dnsMap.containsKey(publicDnsIp)) {
//					Main.getReporter().insertDnsMapping(Statistics.getInstance().getPeerId(myIp), publicDnsId, dnsId, "P", order);
//					if (publicDnsId == -1)
//						publicDnsId = Main.getReporter().makeId("dnsips", "ip", publicDnsIp);
//
//					//pingDnsServers(publicDnsIp);
//					//traceRouteDnsServers(publicDnsIp);
//					Statistics.dnsMap.put(publicDnsIp, Statistics.dnsMap.size());
//				}
			  }
			  			  
			}
			
			ArrayList<String> in = new ArrayList<String>();
			in.add(n);
			in.add(type.equals("A")?publicDnsIpAna:publicDnsIp);
			in.add(type);
			myDns.add(in);			
			
			if (!Statistics.dnsMap.containsKey(n))
				Statistics.dnsMap.put(n, Statistics.dnsMap.size());
		}
		
		return myDns;
	}
	
	public String getPublicDnsIpAna (String dnsIp, String myIp) {
	
		String returnDnsIp = dnsIp;
		Message msg = doLookup(myIp+".ana-aqualab.cs.northwestern.edu", dnsIp, null, -1);
		if ( msg != null ) {
			Record[] records = msg.getSectionArray(Section.ANSWER);
			if (records.length != 0) {
				//System.out.println("PublicDnsIpAna: " + records[0].rdataToString());				
				returnDnsIp = records[0].rdataToString();	
			}
		}
		
		return returnDnsIp; 
	}
	
	public String getMyPublicDnsIp (String dnsIp, String myIp) {

		String returnDnsIp = dnsIp;
		
		Message msg = doLookup(myIp, null, "-x", -1);
		if ( msg != null ) {
			
			Record[] records = msg.getSectionArray(Section.ANSWER);
			if (records!=null && records.length != 0) {
				String ipAddressResponse = records[0].rdataToString();
				//System.out.println("MyPublicDns: " + ipAddressResponse);
				ipAddressResponse = 
					ipAddressResponse.substring(ipAddressResponse.indexOf("."));
				//System.out.println("MyPublicDns: " + ipAddressResponse);
				msg = doLookup(ipAddressResponse.substring(1), null, "-t NS", -1);
				records = msg.getSectionArray(Section.ANSWER);

				if (records.length != 0) {
					//System.out.println("MyPublicDns: " + records[0].rdataToString());
					returnDnsIp = records[0].rdataToString();
					
					msg = doLookup(records[0].rdataToString(), null, null, -1);
					records = msg.getSectionArray(Section.ANSWER);
					
					if (records.length != 0) {
						returnDnsIp = records[0].rdataToString();
						//System.out.println("LOOKUP " + records[0].rdataToString());
					}
				}
			}
		}
		
		if (!isValidIp(returnDnsIp))
			return dnsIp;
		
		return returnDnsIp;
	}
	
	private boolean isPrivateIp(String ip) {

		 if (ip.startsWith("127.") || ip.startsWith("192.168") || ip.startsWith("10.") || 
				 ip.matches("172.1[6-9].[\\d]{1,3}\\.[\\d]{1,3}") || ip.matches("172.2[0-9].[\\d]{1,3}\\.[\\d]{1,3}") ||
				 ip.matches("172.3[0-1].[\\d]{1,3}\\.[\\d]{1,3}"))
 			 return true;
		 
		 return false;  

	 }   
	
	public void launchDnsTimeout(final String dnsIp)
	{
	  
		if (DEBUG)
			System.out.println("Timeout!!!" + timeoutIsRunning);
		
	  if (!timeoutIsRunning.containsKey(dnsIp)) {

		final String myIp = Main.getPublicIpAddress();
		timeoutIsRunning.put(dnsIp, Util.currentGMTTime());
		
	  	Main.createThread("TimeoutDigger-" + dnsIp, new Runnable(){

			public void run() {
				org.xbill.DNS.Name name = null;
				int type = org.xbill.DNS.Type.A;
				int dclass = DClass.IN;
				boolean again = true;
				boolean giveUp = false;

				Message query;
				Record rec;
				SimpleResolver res = null;
				int timeout = DasuConfiguration.getInstance().getDnsTimeout();
				
				long b4 = Util.currentGMTTime();
				int i = 1;
				while (again) {
				  try {
					name = Name.fromString("google.com", Name.root);
					res = new SimpleResolver(dnsIp);
					rec = Record.newRecord(name, type, dclass);
					query = Message.newQuery(rec);
					res.send(query);
					again = false;
				  
				  } catch (IOException e1) {
					  
					  if (DEBUG)
						  System.out.println(dnsIp + " dns TIMEOUT: try again in " + timeout + " seconds...");
					  
					  if (Util.currentGMTTime() - b4 >= i*5*60*1000) {
						  timeout *= 2;
						  i++;
						  
						  if (DEBUG)
							  System.out.println(dnsIp + " timeout: " + timeout);
					  }
						  
					  //try again until timeout >= 10 mins, then give up...
					  if ( timeout >= 10*60 ) {
						  
						  if (DEBUG)
							  System.out.println(dnsIp + " dns TIMEOUT: giving up...");
						  
						  again = false;
						  giveUp = true;
					  }
					  else {
				  
						  try {
								Thread.sleep(timeout*1000);
							  } catch (InterruptedException e) {
								e.printStackTrace();
							  }
							  again = true;	
					  }
				  }
				  
				} //while
				
				long after = Util.currentGMTTime();
				if (DEBUG) {
					System.out.println(dnsIp + " dns TIMEOUT: back online...");
					System.out.println(dnsIp + ": " + Long.toString(after - b4));
				}
				
				Statistics.getInstance().addDnsTimeout(dnsIp, myIp, b4, after, giveUp);
				timeoutIsRunning.remove(dnsIp);		
				
				AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDnsTimeout(dnsIp, b4, after, giveUp));
				
			} //run
		     	
	  	 });


	  }	// if

	}
	
	private void updateLookupResultQueue(FactDnsProbeResult result) {
		
		Long id = System.nanoTime();
		dnsResultTable.put(id, result);
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(
				new Pair<Long, EventType>(id, EventType.DNS_RESULT));
			AMCoordinatorV2.events.notifyAll();
		}
	}
	
	@Override
	public void stopModule() {
		
		setActive(false);
		synchronized(toDoLookupList) {
			toDoLookupList.add(new TaskEntry("HALT", -1, null, null));
			toDoLookupList.notify();
		}
	
		self = null;
	}
	
	
}
