/**
 * Dasu Project
 *
 * File:         DasuTraceRouteModuleV2.java
 * RCS:          $Id$
 * Description:  DasuTraceRouteModuleV2 class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:24:54 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.measurement;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.logging.Logger;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactTraceResult;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.HashSetCache;
import edu.northwestern.dasu.util.Pair;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The HamecaTraceRouteModule class manages the set of 
 *         trace route measurements performed by the peer.
 */


public class DasuTraceRouteModuleV2 extends DasuProbeModule {	
	
	private final static Logger LOGGER = Logger.getLogger(DasuTraceRouteModuleV2.class.getName());
	
	public Hashtable<Long, FactTraceResult> trResultTable;
	private static DasuTraceRouteModuleV2 self;

    private HashSetCache<TraceRouteRun> pendingRuns;
    private HashSetCache<TraceRouteRun> pendingRunsLowPrio;
    private Object lossSync = new Object();
    private double lossSum = 0;
    private double lossCount = 0;
    //private Integer currentRuns;
    private Hashtable<String, TraceRouteRun> currentRuns;
	
	public DasuTraceRouteModuleV2() {
		super("DasuTraceRouteModule");
		this.pi = Main.getPluginInterface();
		self = this;
		active = true;
    	pendingRuns = new HashSetCache<TraceRouteRun>(200);
    	pendingRunsLowPrio = new HashSetCache<TraceRouteRun>(200);
    	trResultTable = new Hashtable<Long, FactTraceResult>();
        currentRuns = new Hashtable<String, TraceRouteRun>();

	}
	
	public synchronized static DasuTraceRouteModuleV2 getInstance() {
		if (self == null) {
			self = new DasuTraceRouteModuleV2();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}
		
	public void probeWithLowPriority(String ip, Integer taskId, String target) {
		
		synchronized(pendingRunsLowPrio){
			pendingRunsLowPrio.add(new TraceRouteRun(ip, taskId, target));
		}

	}

	@Override
	public void probe(String ip, Integer taskId, String target) {
		
		synchronized(pendingRuns){
			pendingRuns.add(new TraceRouteRun(ip, taskId, target));
			pendingRuns.notifyAll();
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void run() {

		//TraceRouteRun trr = null;
		while (!Main.isShuttingDown() && isActive()){
			
			synchronized(pendingRuns){
				while (pendingRuns.size()==0 && pendingRunsLowPrio.size()==0 
						&& !Main.isShuttingDown()){
					try {
						pendingRuns.wait(sleepInterval*1000);
					} catch (InterruptedException e) {
						setActive(false);
						self = null;
						return;
					}
				}
			}
				
			synchronized(currentRuns) {
				while (currentRuns.size() >= 5) {
					try {
						LOGGER.warning("too many concurrent runs (" + currentRuns.size() + "), " + "wait for 5 secnods... ");
						LOGGER.warning("pedingRuns: " + pendingRuns.size() + " pendingRunsLowPrio: " + pendingRunsLowPrio.size());
						
						currentRuns.wait(5*1000);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}

			TraceRouteRun trr = null;

			synchronized(pendingRuns) {
				
				Iterator<TraceRouteRun> it;
				if (pendingRuns.size()>0){
					
					it = pendingRuns.iterator();
					trr = it.next();					
					it.remove();
					
					LOGGER.info("High priority traceroute to "+trr.ip + ", taskId: " + trr.taskId);
				}
				
			}
			
			if (trr == null) { 
				synchronized(pendingRunsLowPrio) {
						
					Iterator<TraceRouteRun> it;
					if (pendingRunsLowPrio.size()>0){
						
						it = pendingRunsLowPrio.iterator();
						trr = it.next();
						it.remove();
						
						LOGGER.info("Low priority traceroute to "+trr.ip + ", taskId: " + trr.taskId);
					}
				}
			}
				
			if (trr != null) {

				if (trr.ip.equals("HALT")) 
					return;

				Main.createThread("traceRoute " + trr.ip,  trr);
				currentRuns.put(trr.ip, trr);

				
			}
        
		} // end while active

		self = null;
	}
	
	@Override
	protected Object probeIp(String ipAddress, Integer taskId, String target) {
		
		TraceRouteRun trr = new TraceRouteRun(ipAddress, taskId, target);
		Main.createThread("traceRoute " + trr.ip,  trr);
		return null;
	}
	

	private void recordInDb(TraceRouteRun trr, String dest, ArrayList<TraceEntry> data) {
		
		
        if (trr.recordInDB){
        	
        	LOGGER.fine("Traceroute Result for ip " + trr.ip + "(" + data.size() +"), taskId: " + trr.taskId);
        	
        	FactTraceResult tr = new FactTraceResult();
        	tr.setTarget(trr.target);
        	
            tr.dest = dest;
            tr.source = Main.getPublicIpAddress();
            tr.entries = data;
            tr.taskId = trr.taskId;
            
            updateTrResultQueue(tr);
            Statistics.getInstance().addTraceRouteResult(tr);
			
            synchronized(lossSync){
            	double dropped = 0;
            	double count = 0;
            	for (TraceEntry te : data){
            		int myDropped = 0;
            		boolean valid = false;
            		for (int i = 0; i < te.rtt.length; i++){
            			if (te.rtt[i]>0) valid = true;
            			else myDropped++;
            		} // end each entry
            		if (valid) {
            			dropped+=myDropped;
            			count+=te.rtt.length;
            		}
            	} // end all tr data
            	if (count>0){
	            	lossSum+=dropped/count;
	            	lossCount++;
            	}	
            } // end sync
            
        }
	        
	}

	private synchronized void updateTrResultQueue(FactTraceResult ftr) {

		Long id = System.nanoTime();
		trResultTable.put(id, ftr);
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(
					new Pair<Long, EventType>(id, EventType.HTRM_RESULT));
			AMCoordinatorV2.events.notifyAll();
		}
		
		currentRuns.remove(ftr.dest);
	}

	
	private class TraceRouteRun implements Runnable {
		public String ip;
		public boolean recordInDB = true;
		public int taskId = -1;
		public String target;

		Process p;
	    BufferedReader in;
		ArrayList<TraceEntry> data = new ArrayList<TraceEntry>();
		
		public TraceRouteRun(String ip, String target){
			this.ip = ip;
			this.target = target;
		}
		
		public TraceRouteRun(String ip, boolean record, String target){
			this.ip = ip;
			recordInDB = record;
			this.target = target;
		} 

		public TraceRouteRun(String ip, int taskId, String target) {
			this.ip = ip;
			this.taskId = taskId;
			this.target = target;
		}
		
		@Override
		public boolean equals(Object obj) {

			try {
				return obj!=null && obj instanceof TraceRouteRun && ip.equals(((TraceRouteRun)obj).ip) && target.equals(((TraceRouteRun)obj).target);
			}
			catch (Exception e) { 
				return false;
			}
		}
		@Override
		public int hashCode() {

			return ip.hashCode();
		}

		/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run() {

			String line = null;
			
			try {
				
				long now = System.currentTimeMillis(); 
				long timeoutInMillis = 1000L * DasuConfiguration.getInstance().getTracerouteTimeout(); 
				long finish = now + timeoutInMillis; 
	            
				if (pi.getUtilities().isWindows()) {

					p = Runtime.getRuntime().exec("tracert -d -w 3000 " + ip);
					
	                in = new BufferedReader(new InputStreamReader(
	                            p.getInputStream()));              
	                
	                while ((line = in.readLine()) != null) {                	
	                    // parse data
	                	String[] entries = line.split("[\\s]+");
	                	if (entries.length<=1) continue;
	                	try {
	                		Integer.parseInt(entries[1]);
	                	} catch (NumberFormatException e){
	                		continue;
	                	}
	                	TraceEntry te = new TraceEntry();
	                	int i = 2;
	                	int numEntries = 0;
	                	while (i < entries.length-1){
	                        try{
	                		te.rtt[numEntries] = getValue(entries[i]);
	                        } catch (NumberFormatException e){
	                            // failure during traceroute
	                            te.router[te.numRouters++] = getRouterIp(entries[i]);
	                            break;
	                        }
	                		if (te.rtt[numEntries++]<0) i++;
	                		else i+=2;
	                		if (numEntries==te.rtt.length) break;
	                		
	                	}
	                	
	                	te.router[te.numRouters++] = getRouterIp(entries[entries.length-1]);  
	                	data.add(te);
	                	if (data.size()>=30) 
	                		break;
	                	
					    if ( System.currentTimeMillis() > finish ) {

					    	LOGGER.info("Traceroute to " + ip +" look longer than " + DasuConfiguration.getInstance().getTracerouteTimeout() + " seconds, aborted...");
					    	break;
					    }              
					    	
	                }

	                // then destroy
	                if (p!=null) 
	                	p.destroy();         
	                
	                
	            } else if (pi.getUtilities().isLinux() || pi.getUtilities().isOSX() 
	            		|| pi.getUtilities().isUnix() || pi.getUtilities().isFreeBSD()) {
	                
	            	p = Runtime.getRuntime().exec("traceroute -n -w 3 " + ip);
	                
	                in = new BufferedReader(new InputStreamReader(
	                            p.getInputStream()));

	                while (in!=null && (line = in.readLine()) != null) {                	
	                    // parse data
	                	
	                	String[] entries = line.split("[\\s]+");
	                	if (entries.length<=1) continue;
	                	int index = 0;
	                	while (entries[index].equals("")) index++;
	                	try {
	                		Integer.parseInt(entries[index]);
	                	} catch (NumberFormatException e){
	                		// not an entry we care about
	                		continue;
	                	}
	                	TraceEntry te = new TraceEntry();
	                	int rttCount = 0;
	                	for (int i = index+1; i < entries.length; i++){
	                		if (entries[i].contains("*")){
	                			te.router[te.numRouters++] = getRouterIp(entries[i]);
	                			te.rtt[rttCount++] = -1f;
	                			//System.out.println("Found asterisk!");
	                		}
	                		else {
	                			if(entries[i].matches("([0-9]{1,3}\\.){3}[0-9]{1,3}")){
	                				// if code reaches this point, this is an ip
	                				te.router[te.numRouters++] = getRouterIp(entries[i]);
	                				//System.out.println("Found router!");
	                			}
	                			else{
	                				// not an ip, so it must be a value
	                				try {
		                				te.rtt[rttCount] = Float.parseFloat(entries[i++]);
		                				rttCount++;
	                				} catch (NumberFormatException e ){
	                					
	                					continue; // ignore the garbage
	                				}
	                					                				
	                			}
	                		}
	                	}
	                 
	                	data.add(te);
	                	if (data.size()>30) 
	                		break;
	                	
					    if ( System.currentTimeMillis() > finish ) {
					    						    	
					    	LOGGER.info("Traceroute to " + ip +" took longer than " + DasuConfiguration.getInstance().getTracerouteTimeout() + " seconds, aborted...");
						    break;
					    }
					    

	                }

	                // then destroy
	                if (p!=null) 
	                	p.destroy();
	                
	            }
				
	        } catch (IOException e) {
	        	
	        } catch (Exception e){

	        	if (isActive()){
	        		Main.getReporter().reportError(e);
	        		Main.getReporter().reportError(new 
		            		RuntimeException("Error input was: \n"+line));
		        }
	        }	
	        
			recordInDb(this, ip, data);	        
		 	
		}
		
		public void stop() {
			
			if (p!=null)
				p.destroy();
			
			try {
				finalize();
			} catch (Throwable e) {
				e.printStackTrace();
			}
		}
	}
	
	private float getValue(String val) {
		if (val.contains("<"))return 0;
		else if (val.contains("*")) return -1;
    	else return Float.parseFloat(val);
	}
	
	public double getLossSample() {
		synchronized(lossSync){
			double toReturn = lossCount==0?0:lossSum/lossCount;
			lossSum = 0;
			lossCount = 0;
			return toReturn;
		}
	}

	private String getRouterIp(String maybeIp) {
		if(maybeIp.matches("([0-9]{1,3}\\.){3}[0-9]{1,3}")){
			// if code reaches this point, this is an ip	
			Statistics.getInstance().addRouterForLookup(maybeIp);
			return maybeIp;
			//System.out.println("Found router!");
		}
		Statistics.getInstance().addRouterForLookup("unknown");
		return "unknown";
	}
		

	@Override
	public void stopModule(){
		
		setActive(false);
		
		synchronized(pendingRuns) {
			pendingRuns.clear();
			pendingRuns.add(new TraceRouteRun("HALT", null));
			pendingRuns.notify();
		}
		
		synchronized(currentRuns) {
			for (String s : currentRuns.keySet())
				currentRuns.get(s).stop();
			
			currentRuns.clear();
			currentRuns.notify();
		}
				
		if (self==null || 
				(self.pendingRuns==null && self.pendingRunsLowPrio==null)) return;

		if (self != null && self.pendingRuns != null){
			synchronized(self.pendingRuns){
				self.pendingRuns.notifyAll();
			}
		}		
		
		self = null;
		pi = null;
	}
	
	
	public synchronized Integer getCurrentRuns() {
		return currentRuns.size();
	}

}
