/**
 * Dasu Project
 *
 * File:         ResetMonitor.java
 * RCS:          $Id: ResetMonitor.java,v 1.0 Mar 5, 2011 4:19:20 PM john Exp $
 * Description:  ResetMonitor class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Mar 5, 2011 at 4:19:20 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;

import java.util.Date;

import org.gudy.azureus2.core3.logging.ILogEventListener;
import org.gudy.azureus2.core3.logging.LogEvent;
import org.gudy.azureus2.core3.logging.LogIDs;
import org.gudy.azureus2.core3.logging.LogRelation;
import org.gudy.azureus2.core3.logging.Logger;
import org.gudy.azureus2.plugins.utils.UTTimer;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;

import com.azureus.plugins.aznetmon.main.RSTPacketCountPerformer;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.download.DownloadMonitor;
import edu.northwestern.dasu.stats.RSTStat;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Util;

/**
 * Dasu Project
 *
 * File:         ResetMonitor.java
 * RCS:          $Id: ResetMonitor.java,v 1.0 Mar 5, 2011 4:19:20 PM john Exp $
 * Description:  ResetMonitor class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Mar 5, 2011 at 4:19:20 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

/**
 * @author John Otto
 *
 * The ResetMonitor class ...
 * 	JSO: I think there was redundancy in reset monitoring...
 * 		perf and utTimer were created/destroyed in Main, but updated manually in the run loop of DasuManager
 * 	*** IN ADDITION, another reset listener was created in DasuManager...
 * 
 * THIS CLASS DOES BOTH THINGS!
 */
public class ResetMonitor implements ILogEventListener {

	private final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger(ResetMonitor.class.getName());
	
	/** timer for RST observations */
	protected RSTPacketCountPerformer perf;
	private UTTimer utTimer = null;
	
	private DownloadMonitor dlm;
	
	public ResetMonitor(DownloadMonitor dlm) {
		this.dlm = dlm;
		perf = new RSTPacketCountPerformer();
		// ADD as a listener!!!
		
		Logger.addListener(this);
	}
	
	public void update() {
		perf.perform(new UTTimerEvent() { public void cancel() {} });
	}
	
	public void runOnce(String name, long interval) {
		utTimer = Main.getPluginInterface().getUtilities().createTimer(name, false);
		utTimer.addEvent(interval, perf);
	}
	
	public void destroy() {
		if(utTimer != null) {
			utTimer.destroy();
		}
		
		Logger.removeListener(this);
	}

	/* (non-Javadoc)
	 * @see org.gudy.azureus2.core3.logging.ILogEventListener#log(org.gudy.azureus2.core3.logging.LogEvent)
	 */
	public void log(LogEvent event) {
		int type = -1;

		if (event.logID == LogIDs.PEER) {
		
			if (event.text.indexOf("Connection reset by peer") != -1)
				type = 0;
			
			else if (event.text.indexOf("An existing connection was forcibly closed by the remote host") != -1) 
				type = 1;
			
			else if (event.text.indexOf("Established") != -1) 
				type = 2;

		}
		else if ((event.logID == LogIDs.PLUGIN)) {
			return;
		}

		if (type != -1) {

			LOGGER.finest("************" + event.text + "********");


			for (Object obj : event.relatedTo) {
				
				if (obj instanceof LogRelation) {
					String relatedTo = ((LogRelation) obj)
							.getRelationText();
					
					//if (DEBUG) {
					//	System.out.println("obj: " + obj.getClass());
					//	System.out.println("relatedTo obj: " + relatedTo);
					//}
					
					org.gudy.azureus2.core3.peer.impl.transport.PEPeerTransportProtocol pe = 
						(org.gudy.azureus2.core3.peer.impl.transport.PEPeerTransportProtocol) obj;
					
					long timeStamp = Util.currentGMTTime();

					String tdlm = pe.getManager().getDisplayName();
					String peerip = pe.getIp();
					String bitTorrentClient = pe.getClient();
					int port = pe.getPort();
					boolean isSeed = pe.isSeed();
					boolean amSeed = pe.getManager().isSeeding();
					String source = pe.getPeerSource();
					int numSeeds = pe.getManager().getNbSeeds();
					int numPeers = pe.getManager().getNbPeers();
					String encryption = pe.getEncryption();
					
					//boolean isCryptoRequired = com.aelitis.azureus.core.networkmanager.NetworkManager.getCryptoRequired(pe.getManager().getAdapter().getCryptoLevel());
					
					// If we're not tracking this download...return.
					if (!dlm.hasId(tdlm)) return;
					long did = dlm.getId(tdlm);

					LOGGER.finest("tdlm: " + tdlm);
					LOGGER.finest("peerip: " + peerip);
					LOGGER.finest("client: " + bitTorrentClient);
					LOGGER.finest("port: " + port);
					LOGGER.finest("isSeed: " + isSeed);
					LOGGER.finest("amSeed?: " + amSeed);
					LOGGER.finest("source: " + source);
					LOGGER.finest("numSeeds: " + numSeeds);
					LOGGER.finest("numPeers: " + numPeers);
					LOGGER.finest("encryption: " + encryption);
					
					Main.logPlugin(event.text + " (" + peerip + ")");
					
					if ( type == 0 || type ==1 ) {
					
						RSTStat rstStat = new RSTStat(did, peerip, timeStamp, 1, type, 
								port, bitTorrentClient, source, isSeed, amSeed,
								numSeeds, numPeers, encryption);

						/*if the download changed status recently ignore reset*/							
						if (DasuConfiguration.getInstance().getIgnoreStateChangeRST()) {
							Long lastState = dlm.getLastStateChange(tdlm);
							if (lastState != null) {
								
								if (System.currentTimeMillis() - lastState < DasuConfiguration.getInstance().getStateChangeRSTInterval()*1000 ) {
									LOGGER.fine("****Ignore reset... due to download state change****");
								
									return;
								}
							}
						}
						
					
						Statistics.getInstance().addRSTStatResult(rstStat);
						
						dlm.updateRstFacts(did, rstStat);
					}
//					else if (type == 2) {
//						
//						RemotePeerConnectionStat rpcStat = new RemotePeerConnectionStat(peerip, timeStamp, source, 
//								port, bitTorrentClient, isSeed, amSeed, numSeeds, numPeers, encryption);
//						
//					
//						Statistics.getInstance().addRpcStatResult(
//								rpcStat);
//						
//					}

				}
			}		
		}
	}
}
