/**
 * Dasu Project
 *
 * File:         HistoryManager.java
 * RCS:          $Id$
 * Description:  HistoryManager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 2, 2010 at 8:45:04 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;

import java.io.Serializable;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;

import edu.northwestern.dasu.DasuConfiguration;


/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The SignalSample class ...
 */
public class SignalSample implements Serializable {

	private static final long serialVersionUID = 1221642255540144145L;
	
	private String name;
	private TimeType timeSpanType;
	private TimeType intervalType;
	private int intervalLength;
	private LinkedHashMap<Integer, Float> values;
	private LinkedHashMap<Integer, Long> timeStamps;
	private static boolean DEBUG = false; 
	private String units;
	
	public SignalSample(String signalName, TimeType timeSpanType, TimeType intervalType, int intervalLength, String units) throws Exception {
		
		if (DEBUG) {
			System.out.println("DEBUG: New SignalSample(): ");
			System.out.println("timeSpanType: " + timeSpanType);
			System.out.println("intervalType: " + intervalType);
			System.out.println("intervalLength: " + intervalLength);
		}
		
		this.name = signalName;
		if (timeSpanType.getVal() <= intervalType.getVal()) 
			throw new Exception("Interval length must be great than intervalType");
		
		if (intervalType.getVal() * intervalLength > timeSpanType.getVal())
			throw new Exception("The number of intervals of the specified type exceed the specified time span");
		
		if (timeSpanType.getVal() / (intervalType.getVal() * intervalLength) 
				> DasuConfiguration.getInstance().getMaxNumEntriesPerHistoricSignal())
			throw new Exception("The number of samples per timeSpan is too large");
		
		this.timeSpanType = timeSpanType;
		this.intervalType = intervalType;
		this.intervalLength = intervalLength;
		this.units = units;
		values = new LinkedHashMap<Integer, Float>();
		timeStamps = new LinkedHashMap<Integer, Long>();
		
		if (timeSpanType.equals(TimeType.YEAR)) { 
			
			if (intervalType.equals(TimeType.MONTH)) {
					
				for (int i=0; i <= Calendar.getInstance().getMaximum(Calendar.MONTH); i += intervalLength) {
					values.put(i/intervalLength, -1F);
					timeStamps.put(i/intervalLength, -1L);
				}				
			}
		
			else if (intervalType.equals(TimeType.WEEK)) {

				/*53 weeks in a year?*/
				for (int i=0; i <= Calendar.getInstance().getMaximum(Calendar.WEEK_OF_YEAR); i += intervalLength) {
					values.put(i/intervalLength, -1F);
					timeStamps.put(i/intervalLength, -1L);
				}				
			
			}
		
			else if (intervalType.equals(TimeType.DAY)) {
	
				/*365 days in a year?*/
				for (int i=0; i <= Calendar.getInstance().getMaximum(Calendar.DAY_OF_YEAR); i += intervalLength) {
					values.put(i/intervalLength, -1F);
					timeStamps.put(i/intervalLength, -1L);
				}				
			}
			else {
				throw new Exception("The number of samples per timeSpan is too large");
			}
		}
		
		else {
			for (int i=0; i < timeSpanType.getVal(); i+=intervalLength*intervalType.getVal() ) {
				values.put(i/(intervalLength*intervalType.getVal()), -1F);
				timeStamps.put(i/(intervalLength*intervalType.getVal()), -1L);
			}
		}
		
		
		for (int k: values.keySet()) {
			System.out.println("index: " + k);
		}
		
	}
	
	public String getName() {
		return name;
	}
	
	public String getUnits() {
		return units;
	}
	
	public long getTimeStamp(long time) {
		
		return timeStamps.get(toInterval(time));
	}
	
	public HashMap<Integer, Long> getTimeStampsHash() {
		return timeStamps;
	}
	
	public HashMap<Integer, Float> getValuesHash() {
		return values;
	}
	
	public float get(long time) {		
		return values.get(toInterval(time));
	}

	@Override
	public String toString() {
		
		String ret = "SingnalSample: " + name + " timeSpanType: " + timeSpanType + " intervalType: " +
			intervalType + " invervalLength: " + intervalLength + " units: " + units + "\n";
		ret += values + "\n";
		ret += timeStamps + "\n";
		return ret;
	
	}
	
	public boolean set(long time, float val) {
		
		int key = toInterval(time);
		if (!values.containsKey(key)) {
			System.out.println("There's something wrong with the timeline!!");
			return false;
		}
		values.put(key, val);
		timeStamps.put(key, time);
		return true;
	}
	
	public boolean isSet(long time) {
		
		int key = toInterval(time);
		return ((values.containsKey(key)) && values.get(key) != -1);
			
	}
	
	private int toInterval(long timeRecorded) {
		
		Calendar c = Calendar.getInstance();
		c.setTimeInMillis(timeRecorded);
		//System.out.println(new Date(timeRecorded));
		
		int interval = 0;

		/*treat specially if it's a year, try to avoid leap year*/

		if (timeSpanType.equals(TimeType.YEAR)) {
			
			
			if (intervalType.equals(TimeType.MONTH)) {
				Double result  = Math.floor( (double) (c.get(Calendar.MONTH)) / intervalLength) ;
				return result.intValue();
			}
			
			else if (intervalType.equals(TimeType.WEEK)) {
				//System.out.println(c.get(Calendar.WEEK_OF_YEAR));
				Double result  = Math.floor( (double) (c.get(Calendar.WEEK_OF_YEAR)-1) / intervalLength);
				return result.intValue();
			}
			
			else if (intervalType.equals(TimeType.DAY)) {
				Double result  = Math.floor( (double) (c.get(Calendar.DAY_OF_YEAR)-1) / intervalLength);
				return result.intValue();
			}
		
		}
		
		if (timeSpanType.equals(TimeType.MONTH)) {

			if (intervalType.equals(TimeType.WEEK)) {
				int week = c.get(Calendar.WEEK_OF_MONTH)-1;
				interval = week*7*24*60*60;
			}
			
			else if (intervalType.equals(TimeType.DAY)) {
				int day = c.get(Calendar.DAY_OF_MONTH)-1;
				interval = day*24*60*60;
			}
			
			else if (intervalType.equals(TimeType.HOUR)) {
				int day = c.get(Calendar.DAY_OF_MONTH)-1;
				int hour = c.get(Calendar.HOUR_OF_DAY);
				interval = day*24*60*60 + hour*60*60;
			}
			
			else if (intervalType.equals(TimeType.MINUTE)) {
				int day = c.get(Calendar.DAY_OF_MONTH)-1;
				int hour = c.get(Calendar.HOUR_OF_DAY);
				int min = c.get(Calendar.MINUTE);
				
				interval = day*24*60*60 + hour*60*60 + min*60;
			}
			
			else if (intervalType.equals(TimeType.SECOND)) {
				int week = c.get(Calendar.WEEK_OF_MONTH)-1;
				int day = c.get(Calendar.DAY_OF_WEEK)-1;
				int hour = c.get(Calendar.HOUR_OF_DAY);
				int min = c.get(Calendar.MINUTE);
				int sec = c.get(Calendar.SECOND);
				
				interval = week*7*24*60*60 + day*24*60*60 + hour*60*60 + min * 60 + sec;
			}
			
		}
		else if (timeSpanType.equals(TimeType.WEEK)) {

			if (intervalType.equals(TimeType.DAY)) {
				int day = c.get(Calendar.DAY_OF_WEEK)-1;
				interval = day*24*60*60;
			}
			
			else if (intervalType.equals(TimeType.HOUR)) {
				int day = c.get(Calendar.DAY_OF_WEEK)-1;
				int hour = c.get(Calendar.HOUR_OF_DAY);
				interval = day*24*60*60 + hour*60*60;
			}
			
			else if (intervalType.equals(TimeType.MINUTE)) {
				int day = c.get(Calendar.DAY_OF_WEEK)-1;
				int hour = c.get(Calendar.HOUR_OF_DAY);
				int min = c.get(Calendar.MINUTE);
				
				interval = day*24*60*60 + hour*60*60 + min*60;
			}
			
			else if (intervalType.equals(TimeType.SECOND)) {
				int day = c.get(Calendar.DAY_OF_WEEK)-1;
				int hour = c.get(Calendar.HOUR_OF_DAY);
				int min = c.get(Calendar.MINUTE);
				int sec = c.get(Calendar.SECOND);
				
				interval = day*24*60*60 + hour * 60 *60 + min * 60 + sec;
			}
			
		}
		
		else if (timeSpanType.equals(TimeType.DAY)) {
			
			if (intervalType.equals(TimeType.HOUR)) {
				
				int hour = c.get(Calendar.HOUR_OF_DAY);
				interval = hour*60*60;
			}
			
			else if (intervalType.equals(TimeType.MINUTE)) {
				int hour = c.get(Calendar.HOUR_OF_DAY);
				int min = c.get(Calendar.MINUTE);
				
				interval = hour*60*60 + min*60;
			}
			
			else if (intervalType.equals(TimeType.SECOND)) {
				int hour = c.get(Calendar.HOUR_OF_DAY);
				int min = c.get(Calendar.MINUTE);
				int sec = c.get(Calendar.SECOND);
				
				interval = hour*60*60 + min * 60 + sec;
			}
			
		}
		
		else if (timeSpanType.equals(TimeType.HOUR)) {
			
			if (intervalType.equals(TimeType.MINUTE)) {
				int min = c.get(Calendar.MINUTE);
				
				interval = min*60;
			}
			
			else if (intervalType.equals(TimeType.SECOND)) {
				int min = c.get(Calendar.MINUTE);
				int sec = c.get(Calendar.SECOND);
				
				interval = min*60 + sec;
			}
			
		}
		
		else if (timeSpanType.equals(TimeType.MINUTE)) {
			
			if (intervalType.equals(TimeType.SECOND)) {
				int sec = c.get(Calendar.SECOND);
				
				interval = sec;
			}
			
		}
		
		if (DEBUG) {
			System.out.println("interval: " + interval);
			System.out.println("divisor: " + intervalLength*intervalType.getVal());
		}
		
		Double result = Math.floor((double) interval / (double) (intervalLength*intervalType.getVal())) /** intervalLength* intervalType.getVal()*/;
		//System.out.println(result.intValue());
		return result.intValue();
	}
	
}
	

