/**
 * Dasu Project
 *
 * File:         NetstatBWStats.java
 * RCS:          $Id$
 * Description:  NetstatBWStats class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Sep 30, 2010 at 12:15:52 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.measurement;

import com.azureus.plugins.aznetmon.main.DataListener;
import edu.northwestern.dasu.util.Util;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

import org.apache.commons.math.stat.univariate.rank.Percentile;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The NetstatBWStats class ...
 */

public class UPnPBWStats {


    public static long INTERVAL = 1000 * 30 * 1; //show a stat every 30 seconds.
    public static int MAX_NUM_ENTRIES = 100;

    //Some constants for other packages to use.
    public static final String PLUGIN_MAP_NAME = "plugin.versionserver.data";

    
    private static UPnPBWStats instance = new UPnPBWStats(INTERVAL);

    //data
    private List<UPnPBWData> bwMeasurementDataList;
    private List dataListeners; //<DataListener>
    private UPnPBWData baseBWData = null;

    private String status;


    public synchronized static UPnPBWStats getInstance() {
        return instance;
    }

    private UPnPBWStats(long interval)
    {
        bwMeasurementDataList = new ArrayList<UPnPBWData>(); //<BWData>
        dataListeners = new ArrayList<DataListener>(); //<DataListener>
        status = "Initial data is zero. Collection is 10 min interval.";
    }

    /**
     * What is the status of the application.
     * @param _status -
     */
    public void setStatus(String _status){
        status = _status;
    }

    public String getStatus(){
        return status;
    }

    
    public void add(UPnPBWData d){
    	
    	bwMeasurementDataList.add(0,d);
    	
    	if (baseBWData == null)
    		baseBWData = new UPnPBWData(d.timeRecorded, d.totalBytesSent, d.totalBytesReceived, d.numberOfActiveConnections, 
    				d.totalPacketsSent, d.totalPacketsReceived, d.uptime);
    	

        if (bwMeasurementDataList.size()>MAX_NUM_ENTRIES) 
        	bwMeasurementDataList.remove(bwMeasurementDataList.size()-1);
        
        //inform any listeners of this new data.
        update();

        //update the status
        setStatus( "Last measurement at "+ Util.currentGMTTime() + ". Next update in " + INTERVAL/1000 + " minutes" );

    }//add


    public UPnPBWData getMostRecent(){

        if( bwMeasurementDataList.size()>0 ){
            return (UPnPBWData) bwMeasurementDataList.get(0);
        }else{
            return null;
        }
    }

    public UPnPBWData getBaseBWData() {
    	return baseBWData;
    }
    
    
    public UPnPBWData getValAroundTime(long time) {

    	UPnPBWData bwData = null;
    	if( bwMeasurementDataList.size()>0 ){
    		for (int i = bwMeasurementDataList.size()-1; i >= 0; i--) {
    			if (Math.abs(( (UPnPBWData) bwMeasurementDataList.get(i)).timeRecorded - time )/1000 <= 5) {
    				bwData = (UPnPBWData) bwMeasurementDataList.get(i);
    				break;
    			}
    				
    			//return (NetstatBWData) bwMeasurementDataList.get(0);
    		}
    	}
    	
    	return bwData;
    	
    }
    

    /**
     * Get upto n most recent results.
     * @param n -
     * @return -
     */
    public UPnPBWData[] getMostRecent(int n){

        int size = bwMeasurementDataList.size();
        if( n<size ) size = n;

        if(size==0){
        	UPnPBWData[] def = new UPnPBWData[1];
            def[0] = new UPnPBWData();
            return def;
        }

        UPnPBWData[] retVal = new UPnPBWData[size];
        for(int i=0; i<size; i++){

            retVal[i] = (UPnPBWData) bwMeasurementDataList.get(i);

        }//for
        
        return retVal;
    }

    private ArrayList<UPnPBWData> getDataBetween(long first, long last) {
    	
    	ArrayList<UPnPBWData> retVal = new ArrayList<UPnPBWData>();
    	
    	Iterator<UPnPBWData> iter = bwMeasurementDataList.iterator();
    	while (iter.hasNext()) {
    		UPnPBWData data = iter.next();
    		if ((data.timeRecorded >= first) && (data.timeRecorded <= last))
    			retVal.add(data);
    	}

    	return retVal;
    }
    
    /* returns values in BytesPerSecond */
    public Double getPercentileUploadBetween(long first, long last, float percentile) {
		
    	ArrayList<UPnPBWData> vals = getDataBetween(first, last);
    	
    	double[] v = new double[vals.size()];
    	Double currSent = 0d;
    	Long currTimeSpan = 1l;
    	
    	for (int i=0; i < vals.size()-1; i++) {
			currSent = (double) (vals.get(i).totalBytesSent - vals.get(i+1).totalBytesSent);
			currTimeSpan = (vals.get(i).timeRecorded - vals.get(i+1).timeRecorded);
			v[i] = currSent/currTimeSpan;
    	}
    	
    	Percentile pt = new Percentile();
    	pt.setPercentile(percentile);
    	return pt.evaluate(v)*1000;
    	
    }
    
    /* returns values in BytesPerSecond */
    public Double getPercentileDownloadBetween(long first, long last, float percentile) {
		
    	ArrayList<UPnPBWData> vals = getDataBetween(first, last);
    	
    	double[] v = new double[vals.size()];
    	Double currReceived = 0d;
    	Long currTimeSpan = 1l;
    	
    	for (int i=0; i < vals.size()-1; i++) {
			currReceived = (double) (vals.get(i).totalBytesReceived - vals.get(i+1).totalBytesReceived);
			currTimeSpan = (vals.get(i).timeRecorded - vals.get(i+1).timeRecorded);
			v[i] = currReceived/currTimeSpan;
    	}
    	
    	Percentile pt = new Percentile();
    	pt.setPercentile(percentile);
    	return pt.evaluate(v)*1000;
    	
    }
    
    /* returns values in BytesPerSecond */
    public Double getFastestUploadBetween(long first, long last) {
    	
    	ArrayList<UPnPBWData> vals = getDataBetween(first, last);
    	Long timeSpan = 1l;
    	Double currSent = 0d;
    	Long currTimeSpan = 1l;
    	Double ret = new Double(-1);
    	
    	for (int i=0; i < vals.size()-1; i++) {

			currSent = (double) (vals.get(i).totalBytesSent - vals.get(i+1).totalBytesSent);
			currTimeSpan = (vals.get(i).timeRecorded - vals.get(i+1).timeRecorded);

			if (currTimeSpan != 0 && timeSpan != 0) {
				if (currSent/currTimeSpan > ret)
					ret = currSent/currTimeSpan;
			}
    	}
    	
    	return ret*1000;
    }
    
    /* returns values in BytesPerSecond */
  
    public Double getFastestDownloadBetween(long first, long last) {
    	    	
    	ArrayList<UPnPBWData> vals = getDataBetween(first, last);
    	Long timeSpan = 1l;
    	Double currReceived = 0d;
    	Long currTimeSpan = 1l;
    	Double ret = new Double(-1);
    	
    	for (int i=0; i < vals.size()-1; i++) {

    		currReceived = (double) (vals.get(i).totalBytesReceived - vals.get(i+1).totalBytesReceived);
			currTimeSpan = (vals.get(i).timeRecorded - vals.get(i+1).timeRecorded);
			
			if (currTimeSpan != 0 && timeSpan != 0) {
				if (currReceived/currTimeSpan > ret)
					ret = currReceived/currTimeSpan;
			}
    	}
    	
    	return ret*1000;
    }
    
    public synchronized void update(){

        int size = dataListeners.size();
        for(int i=0; i<size; i++ ){
            DataListener dl = (DataListener) dataListeners.get(i);
            dl.update();
        }

    }//update

    @Override
    public String toString() {
    	
    	String ret = new String();
    	Iterator<UPnPBWData> iter = bwMeasurementDataList.iterator();
    	while (iter.hasNext()) {
    		UPnPBWData data = iter.next();
    		ret = ret.concat(data.toString());
    	}
    	
    	return ret;
    }
    
    
    /**
     * data-type.
     */
    public static class UPnPBWData{

    	public long timeRecorded=0;
        public String msg = "";

        public long totalBytesSent=-1;
        public long totalBytesReceived=-1;
    	public long numberOfActiveConnections=-1;
    	public long totalPacketsSent=-1;
    	public long totalPacketsReceived=-1;
    	public long uptime=-1;
        
        public long deltaTime=0;
        public long deltaBytesSent=0;
        public long deltaBytesReceived=0;

        //Note: This is used for debugging only.
        public boolean isDebugMode = false;
        public String[] debugLines = null;
        
        public UPnPBWData() {
        	timeRecorded = Util.currentGMTTime();
        }
        
        public UPnPBWData(long sent, long received, long numberOfActiveConnections, long totalPacketsSent,
        		long totalPacketsReceived, long uptime ){
        	this.timeRecorded = Util.currentGMTTime();
        	this.totalBytesSent = sent;
        	this.totalBytesReceived = received;
        	this.numberOfActiveConnections = numberOfActiveConnections;
        	this.totalPacketsSent = totalPacketsSent;
        	this.totalPacketsReceived = totalPacketsReceived;
        	this.uptime = uptime;
        }
        
        public UPnPBWData(long timeRecorded, long sent, long received, long numberOfActiveConnections, long totalPacketsSent,
        		long totalPacketsReceived, long uptime){
        	this.timeRecorded = timeRecorded;
        	this.totalBytesSent = sent;
        	this.totalBytesReceived = received;
        	this.totalBytesReceived = received;
        	this.numberOfActiveConnections = numberOfActiveConnections;
        	this.totalPacketsSent = totalPacketsSent;
        	this.totalPacketsReceived = totalPacketsReceived;
        	this.uptime = uptime;
        }
 

        public UPnPBWData(String _msg){
            msg = _msg;
        }

        /**
         * Calculate percentages and deltas based on the raw data.
         * @param prev -
         */
        public void calculate(UPnPBWData prev){

            deltaTime = timeRecorded - prev.timeRecorded;
            deltaBytesSent = totalBytesSent - prev.totalBytesSent;
            deltaBytesReceived = totalBytesReceived - prev.totalBytesReceived;

        }
        
        @Override
        public String toString() {
        	
        	return "timeRecorded: " + timeRecorded + " bytesSent: " + totalBytesSent +
        		" bytesReceived: " + totalBytesReceived + "\n";
        }
        
    }//static class UPnPBWData


}//class
