/**
 * Dasu Project
 *
 * File:         NetstatBanwidthPerformer.java
 * RCS:          $Id$
 * Description:  NetstatBanwidthPerformer class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Sep 30, 2010 at 11:57:52 AM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.measurement;

import org.gudy.azureus2.plugins.utils.UTTimer;
import org.gudy.azureus2.plugins.utils.UTTimerEventPerformer;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;
import org.gudy.azureus2.plugins.PluginInterface;

import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.measurement.NetstatBWStats.NetstatBWData;
import edu.northwestern.dasu.measurement.UPnPBWStats.UPnPBWData;
import edu.northwestern.dasu.upnp.UPNPDeviceHandler;

import java.util.*;
import java.util.logging.Logger;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The NetstatBanwidthPerformer class ...
 */

public class UPnPBanwidthPerformer implements UTTimerEventPerformer {

	private final static Logger LOGGER = Logger.getLogger(UPnPBanwidthPerformer.class.getName());

    private String platform;
    private String eth;
    private String name;

    public UTTimer timer = null;
    UTTimerEvent timerEvent = null;
    
    //eth is the interface name I want to use
    public UPnPBanwidthPerformer(String eth, String name){

    	this.name = name;
        this.eth = eth;
        this.platform = Main.getPlatform();
        		
    }
    
    public void loop(long interval) {
    	timer = Main.getPluginInterface().getUtilities().createTimer(name, true); // assume "lightweight" task
    	timer.addPeriodicEvent(interval, this);
    }
    
    public void runOnce(long interval) {
    	timer = Main.getPluginInterface().getUtilities().createTimer(name, true); // assume "lightweight" task
    	timer.addEvent(interval, this);
    }

    public void cancelTimer() {
    			
    	LOGGER.info("Cancelling timeEvent!");
    	
    	if (timerEvent != null) {    		
    		timerEvent.cancel();
    	}
    	if (timerEvent != null) {    		
    		timerEvent.cancel();
    	}
    	
    }
    
    public void perform(UTTimerEvent utTimerEvent)
    {
    	
   		timerEvent = utTimerEvent;
   		
   		LOGGER.info("UPNP PERFORM!: " + utTimerEvent.toString());
   		   		
   		update();

    }//perform
    
    public UPnPBWData getData()
    {
    	HashMap<String, String> res = UPNPDeviceHandler.getUPNPDynamicInfo(eth);
        
        if( res != null ){

	   		long totalBytesSent = res.containsKey("TotalBytesSent") ? Long.parseLong(res.get("TotalBytesSent")) : -1;
	   		long totalBytesReceived = res.containsKey("TotalBytesReceived") ? Long.parseLong(res.get("TotalBytesReceived")) : -1;
	   		long numberOfActiveConnections = res.containsKey("NumberOfActiveConnections") ? Long.parseLong(res.get("NumberOfActiveConnections")) : -1;
	   		long totalPacketsSent = res.containsKey("TotalPacketsSent") ? Long.parseLong(res.get("TotalPacketsSent")) : -1;
	   		long totalPacketsReceived = res.containsKey("TotalPacketsReceived") ? Long.parseLong(res.get("TotalPacketsReceived")) : -1;
	   		long uptime = res.containsKey("Uptime") ? Long.parseLong(res.get("Uptime")) : -1;
	   		
   			UPnPBWData bwData = new UPnPBWData(totalBytesSent, totalBytesReceived, 
   					numberOfActiveConnections, totalPacketsSent, totalPacketsReceived, uptime);
   			return bwData;

        }//if
        
        return null;

    }//perform
    
    public void update() {    
    	
		UPnPBWData bwData = getData();
   		UPnPBWStats stats = UPnPBWStats.getInstance();   		

   		if (bwData != null ) {

	   		try  {
	   			
	            //Get the deltas.
	   			UPnPBWData recent = stats.getMostRecent();
	
	            if( recent != null ){
	                bwData.calculate( recent );
	            }
	
	        	stats.add( bwData );
	        	
	   		}
   			catch(Exception e) {
	   			e.printStackTrace();
	   		}
    	}
   		else {   			
   			stats.setStatus("network device not upnp-capable");
   		}
    	
    }


    
    ///*** testing ***/

    /**
     * For testing.
     * @param args
     */
    public static void main(String[] args)
    {

    	String eth = "en1";
    	UPnPBanwidthPerformer bwperf = new UPnPBanwidthPerformer(null, eth);

        UTTimerEvent empty = new UTTimerEvent(){

            public void cancel() {

            }
        };

        bwperf.perform(empty);
        
        try{Thread.sleep(5 * 1000);}catch(Exception e){}

        bwperf.perform(empty);

        System.out.println(NetstatBWStats.getInstance().getMostRecent().bytesReceived);

        
    }//main


}//class

