/**
 * Dasu Project
 *
 * File:         PeerStats.java
 * RCS:          $Id$
 * Description:  PeerStats class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 10, 2007 at 12:38:43 PM
 * Language:     Java
 * Package:      edu.northwestern.news.stats
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2007, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.stats;

import java.util.HashSet;
import java.util.Set;
import java.util.Map.Entry;

import org.gudy.azureus2.core3.config.COConfigurationManager;
import org.gudy.azureus2.core3.config.impl.TransferSpeedValidator;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadManager;
import org.gudy.azureus2.plugins.peers.Peer;

import com.azureus.plugins.aznetmon.main.RSTPacketCountPerformer;
import com.azureus.plugins.aznetmon.main.RSTPacketStats;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.measurement.AMCoordinatorV2;
import edu.northwestern.dasu.measurement.NetstatBWStats;
import edu.northwestern.dasu.measurement.NetstatBanwidthPerformer;
import edu.northwestern.dasu.measurement.ProbeType;
import edu.northwestern.dasu.measurement.SourceType;
import edu.northwestern.dasu.measurement.NetstatBWStats.NetstatBWData;
import edu.northwestern.dasu.measurement.UPnPBWStats;
import edu.northwestern.dasu.measurement.UPnPBWStats.UPnPBWData;
import edu.northwestern.dasu.measurement.UPnPBanwidthPerformer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.util.HashMapCache;
import edu.northwestern.dasu.util.Util;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The PeerStats class represents general data collected for a peer.
 */
public class PeerStatistics implements ITimerEventPerformer{
	
	/** Download rate update interval, in ms */
	private static int RATE_UPDATE_INTERVAL = 5*1000;
	
	/** Peer statistics update interval, in ms */
	private static int PEER_UPDATE_INTERVAL = 30*1000;

	/** max number of dynamic local peer stat entries to cache */
	public static final int MAX_DLPS_ENTRIES = 100;

	/** max number of dynamic (remote) peer stat entries to cache */
	public static final int MAX_DPS_ENTRIES = 100;

	private static final int MAX_PEER_ENTRIES = 200;

	/** max time between reporting peer data */
	private static final long PEER_REPORT_INTERVAL = 30*60*1000;

	/** max time between reporting static data */
	private static final long STATIC_DATA_FLUSH_INTERVAL = 60*60*1000;
	
	/** 
	 * The PeerStaticData class represents data that needs 
	 * to be reported only once during a peer's lifetime.
	 */
	
	private static PeerStatistics self;
	private static HashMapCache<String, RemotePeerStatStatic> peerData;
	/** local peer stats */
	private static StaticLocalPeerStat myStats;
	/** last update time */
	private static long lastUpdateTime = -1;
	private static long lastStaticFlushTime = -1;
	private static long lastFlushTime;
	
	private Long firstBytesSentNetstat = null;
	private Long firstBytesReceivedNetstat = null;
	private Integer firstnConnReset = null; 
	private Integer firstnCurrentOpen = null;
	private Integer firstnPassiveOpens = null;
	private Integer firstnActiveOpens = null;
	private Integer firstnFailedConnAttempt = null;
	
	private Long firstBytesSentUPnP = null;
	private Long firstBytesReceivedUPnP = null;
	private Long firstPacketsSentUPnP = null;
	private Long firstPacketsReceivedUPnP = null;

	
	/**
	 * Creates a new PeerStatistics object
	 *
	 */
	private PeerStatistics(){
		peerData = new HashMapCache<String, RemotePeerStatStatic>(400);
		myStats = new StaticLocalPeerStat();	
	}
	
	/**
	 * Gets a PeerStatistics object.
	 * @return
	 */
	public static PeerStatistics getInstance(){
		if (self == null){
			self = new PeerStatistics();
		}
		return self;
	}
	
	public static StaticLocalPeerStat getLocalPeerStatStatic() {
		return myStats;
	}
	
	/**
	 * Retrieves stat data for the specified peer
	 * @param p the peer to get data from
	 * @return
	 */
	private static RemotePeerStatStatic getPeerData(Peer p){
		
		RemotePeerStatStatic psd;
		psd = peerData.get(p.getIp());
		if (psd==null) {
			psd = new RemotePeerStatStatic(p);
			synchronized(peerData) {
				peerData.put(p.getIp(), psd);
			}
		}
		
		if (peerData.size()>MAX_PEER_ENTRIES || System.currentTimeMillis()-lastFlushTime > PEER_REPORT_INTERVAL){
			synchronized (peerData){
				
				Set<String> toRemove = new HashSet<String>();
				
				// get ids in one batch
				for (Entry<String, RemotePeerStatStatic> ent : peerData.entrySet()){
					Statistics.getInstance().addPeerForLookup(ent.getKey());
					Statistics.getInstance().addPeerClientForLookup(ent.getValue().peer.getClient());
				}
	
				for (Entry<String, RemotePeerStatStatic> sps : peerData.entrySet()){
					if (sps.getValue().isDone()){
						reportStaticPeerStat(sps.getValue());
						
						toRemove.add(sps.getKey());
					}
					
				}
				
				for (String key: toRemove) peerData.remove(key);
			}
			lastFlushTime=System.currentTimeMillis();
			Main.getReporter().disconnect();
		}
		
		
		return psd;
	}

	/**
	 * @param sps
	 */
	private static void reportStaticPeerStat(RemotePeerStatStatic sps) {
		// one last try (3 is the "" string)
		if (sps.peerClient == -1 || sps.peerClient == 3 ) {
			sps.peerClient = Statistics.getInstance().getPeerClientId(sps.peer.getClient());
		}
				
		if (sps.peerId==-1){
			sps.peerId = Statistics.getInstance().getPeerId(sps.peer.getIp());
		}

		if (DasuConfiguration.getInstance().isRecordStats()) {
			Main.getReporter().insertRemotePeerStatStatic(Statistics.getInstance().getSelfId(), 
					sps.peerId, 				
					sps.peerClient, sps.currentLifetime, sps.tcpPort, 
					sps.udpPort, sps.sameCluster, sps.type, sps.encryption, sps.pieceSize);
		}

		sps.flush();

		// do another traceroute if at least 5 minutes passed
		if (Main.isShuttingDown()) return;
		if (Util.currentGMTTime()-sps.firstSeen > 5*60*1000){
			AMCoordinatorV2.addLowPriorityProbe(sps.peer.getIp(), ProbeType.TRACEROUTE, SourceType.NULL);
		}
	}

	/**
	 * Implements the periodic action, gets peer data.
	 */
	public void perform(ITimerEvent event) {
				
		if (Main.isShuttingDown()){			
			if (event!=null) event.cancel();
			return;
		}
		
		if (!DasuConfiguration.getInstance().isRecordStats()) 
			return;
		
		// TODO this won't work w/ more than two intervals--must be craftier about 
		// "last update time".
		if (Util.currentGMTTime()-lastUpdateTime >= RATE_UPDATE_INTERVAL){
			myStats.update(); // note that this updates only, does not add data			
		}
		
		if (Util.currentGMTTime()-lastStaticFlushTime >= STATIC_DATA_FLUSH_INTERVAL) {
			myStats.flushStatic();
			myStats.configChanged = false;
			lastStaticFlushTime = Util.currentGMTTime();
			// while here, also flush remote peer stats
			
			String keys[] = peerData.keySet().toArray(new String[]{});
			synchronized(peerData){
				
				// get ids in one batch
				for (Entry<String, RemotePeerStatStatic> ent : peerData.entrySet()){
					Statistics.getInstance().addPeerForLookup(ent.getKey());
					Statistics.getInstance().addPeerClientForLookup(ent.getValue().peer.getClient());
				}
				
				for (String peerIp : keys){
					RemotePeerStatStatic sps = peerData.get(peerIp);
					if (sps!=null && sps.isDone()){
						peerData.remove(peerIp);
						reportStaticPeerStat(sps);
						
					}
				}
			}
			
		}

		if (Util.currentGMTTime()-lastUpdateTime >= 
				DasuConfiguration.getInstance().getPeerUpdateIntervalSec()*1000){
		
			//collect stats about peer running plugin
			int totalUploadRate=0;
			int totalDownloadRate=0;
			int totalConnections=0;
			int maxDownloadRate = -1;
			int maxUploadRate = -1;
			boolean autoSpeedActive = false;
			int numTorrents = 0;
			int numSeedingTorrents = 0;
			int numLeechingTorrents = 0;
			int numConnectedSeeds = 0;
			int numConnectedLeechers = 0;
			long bytesSentLpsd = 0;
			long bytesReceivedLpsd = 0;
			long swarmPeerSpeed = 0;
			
			// TODO limits on number of peers to collect data from
			DownloadManager dm = Main.getPluginInterface().getDownloadManager();
			totalUploadRate = dm.getStats().getDataSendRate();
			totalDownloadRate = dm.getStats().getDataReceiveRate();	
			bytesSentLpsd = dm.getStats().getDataBytesSent();
			bytesReceivedLpsd = dm.getStats().getDataBytesReceived();
			swarmPeerSpeed = Main.getGlobalManager().getStats().getTotalSwarmsPeerRate(true, false);

			try {
				maxDownloadRate = COConfigurationManager
					.getIntParameter("Max Download Speed KBs");
				maxUploadRate = COConfigurationManager
					.getIntParameter("Max Upload Speed KBs");
				
				autoSpeedActive = checkAutoSpeed();

			} catch (Exception e){
			 // some problem w/ getting parameters...
			} catch (Error e ){
				// same thing, particularly if classes change under us...
			}
			
			Download dls[] = dm.getDownloads();

			numTorrents = dm.getDownloads().length;
			for (Download d : dls){		

				if (d.getState() == Download.ST_SEEDING)
					numSeedingTorrents++;
				if (d.getState() == Download.ST_DOWNLOADING)
					numLeechingTorrents++;
				
				if (d.getPeerManager() != null) {
					numConnectedSeeds += d.getPeerManager().getStats().getConnectedSeeds();
					numConnectedLeechers += d.getPeerManager().getStats().getConnectedLeechers();
				}
				
				if (!(d.getState()==Download.ST_DOWNLOADING 
						|| d.getState() == Download.ST_SEEDING)) continue;
				if (d.getPeerManager()==null) continue;
				Peer peers[] = d.getPeerManager().getPeers();
				totalConnections += peers.length;
				
				
				/*Update/Create RemotePeer Statistics*/
				for (Peer p : peers){
					getPeerData(p).updateStats();
				}
				
				Main.getReporter().disconnect();
			}
			
			myStats.uptime = dm.getStats().getSessionUptimeSeconds();
			DynamicLocalPeerStat dlps = new DynamicLocalPeerStat();
			dlps.downloadRate = totalDownloadRate;
			dlps.uploadRate = totalUploadRate;
			dlps.maxUploadRate = maxUploadRate;
			dlps.maxDownloadRate = maxDownloadRate;
			dlps.autoSpeed = autoSpeedActive;
			dlps.numTorrents = numTorrents;
			dlps.numSeedingTorrents = numSeedingTorrents;
			dlps.numLeechingTorrents = numLeechingTorrents;
			dlps.numConnectedSeeds = numConnectedSeeds;
			dlps.numConnectedLeechers = numConnectedLeechers;	
			dlps.bytesSentLpsd = bytesSentLpsd;
			dlps.bytesReceivedLpsd = bytesReceivedLpsd;
			dlps.swarmPeerSpeed = swarmPeerSpeed;
			
			try {				
				
				if (Main.getPluginInterface().getUtilities().isWindows() || Main.getPluginInterface().getUtilities().isOSX() ||
						Main.getPluginInterface().getUtilities().isLinux()) {
					
					String platform = Main.getPlatform();
					
					if ( RSTPacketStats.getInstance().getBaseRSTData() != null) {
					
						RSTPacketStats.RSTPacketData rstData = new RSTPacketCountPerformer(null).getData(platform);
						
						if (firstnConnReset == null) {
							firstnConnReset = rstData.nConnReset; 
							firstnCurrentOpen = rstData.nCurrentOpen;
							firstnPassiveOpens = rstData.nPassiveOpens;
							firstnActiveOpens = rstData.nActiveOpens;
							firstnFailedConnAttempt = rstData.nFailedConnAttempt;
						}
						
						dlps.rst = rstData.nConnReset - firstnConnReset; 
						dlps.numConnections = rstData.nCurrentOpen - firstnCurrentOpen;
						dlps.passiveOpens = rstData.nPassiveOpens - firstnPassiveOpens;
						dlps.activeOpens = rstData.nActiveOpens - firstnActiveOpens;
						dlps.failedConns = rstData.nFailedConnAttempt - firstnFailedConnAttempt;
					}
					
					if (NetstatBWStats.getInstance().getBaseBWData() != null) {
					
						NetstatBWData bwNetstat = new NetstatBanwidthPerformer(Main.getInterfaceName(), "PeerStatistics").getData(platform);
						
						if (firstBytesSentNetstat == null) {
							firstBytesSentNetstat = bwNetstat.bytesSent;
							firstBytesReceivedNetstat = bwNetstat.bytesReceived;
						}
							
						dlps.bytesSentNetstat =  bwNetstat.bytesSent - firstBytesSentNetstat;  	
						dlps.bytesReceivedNetstat =  bwNetstat.bytesReceived - firstBytesReceivedNetstat;
					}
				}
				
				
				if (UPnPBWStats.getInstance().getBaseBWData() != null) {
					
					UPnPBWData bwUPnP = new UPnPBanwidthPerformer(Main.getInterfaceName(), "PeerStatistics").getData();
					
					if (firstBytesSentUPnP == null) {
						firstBytesSentUPnP = bwUPnP.totalBytesSent;
						firstBytesReceivedUPnP = bwUPnP.totalBytesReceived;
						firstPacketsSentUPnP = bwUPnP.totalPacketsSent;
						firstPacketsReceivedUPnP = bwUPnP.totalPacketsReceived;
					}
						
					dlps.UPnPTotalBytesSent =  bwUPnP.totalBytesSent - firstBytesSentUPnP;
					dlps.UPnPTotalBytesReceived =  bwUPnP.totalBytesReceived - firstBytesReceivedUPnP;
					dlps.UPnPTotalPacketsSent =  bwUPnP.totalPacketsSent - firstPacketsSentUPnP;
					dlps.UPnPTotalPacketsReceived =  bwUPnP.totalPacketsReceived - firstPacketsReceivedUPnP;
					dlps.UPnPNumberOfActiveConnections = bwUPnP.numberOfActiveConnections;
					dlps.UPnPUptime = bwUPnP.uptime;
				}
			
				
			} catch (Exception e){
				// do nothing
				dlps.rst = -1;
				dlps.numConnections = -1;
				dlps.passiveOpens = -1;
				dlps.activeOpens = -1;
				dlps.failedConns = -1;
				dlps.bytesSentNetstat = -1;
				dlps.bytesReceivedNetstat = -1;
			}
		
			
			if (!Main.isShuttingDown()){
				try {
					myStats.add(dlps);
					lastUpdateTime = Util.currentGMTTime();
				}
				catch(Exception e) {					
				}
			}

		}
		
		
	}

	private boolean checkAutoSpeed() {
		String config = "";
		if ( COConfigurationManager.getBooleanParameter(TransferSpeedValidator.AUTO_UPLOAD_ENABLED_CONFIGKEY)){
    		
			config = ( TransferSpeedValidator.AUTO_UPLOAD_ENABLED_CONFIGKEY );
    	} else {
    	
	    	if ( Main.getPluginInterface().getDownloadManager().isSeedingOnly()){
	        	
	    		config = ( TransferSpeedValidator.AUTO_UPLOAD_SEEDING_ENABLED_CONFIGKEY );
	        	
	      	}else{
	      		
	      		config = ( TransferSpeedValidator.AUTO_UPLOAD_ENABLED_CONFIGKEY );
	      	}
    	}
    	
    	return COConfigurationManager.getBooleanParameter(config);
	}

	/**
	 * 
	 * @return time, in ms, between executions of the perform method
	 */
	public static long getInterval() {
		return Math.min(PEER_UPDATE_INTERVAL, RATE_UPDATE_INTERVAL);
	}

	/**
	 * Sends data back to the db.
	 *
	 */
	public void flush() {
		perform(null);
		
		// flush local data
		myStats.flush();
		
		
		// flush other peer data
		synchronized (peerData){
			String keys[] = peerData.keySet().toArray(new String[]{});
			
			// get ids in one batch
			for (Entry<String, RemotePeerStatStatic> ent : peerData.entrySet()){
				Statistics.getInstance().addPeerForLookup(ent.getKey());
				Statistics.getInstance().addPeerClientForLookup(ent.getValue().peer.getClient());
			}

			for (String peerIp : keys){
				RemotePeerStatStatic sps = peerData.get(peerIp);
				sps.flush();
				peerData.remove(sps);
			}
		}
		
	}
	
	/**
	 * Indicates that "static" data should be reported.
	 *
	 */
	public void stop(){
		
		myStats.stop();
		
		// flush other peer data
		String keys[] = peerData.keySet().toArray(new String[]{});
		
		// get ids in one batch
		for (Entry<String, RemotePeerStatStatic> ent : peerData.entrySet()){
			Statistics.getInstance().addPeerForLookup(ent.getKey());
			Statistics.getInstance().addPeerClientForLookup(ent.getValue().peer.getClient());
		}
		
		for (String peerIp : keys){
			RemotePeerStatStatic sps = peerData.get(peerIp);
			if (sps!=null) sps.stop();
			peerData.remove(peerIp);
			reportStaticPeerStat(sps);
		}
		
		synchronized (peerData) {
			peerData.clear();
		}
		peerData = null;
		myStats = null;
		self = null;
	}

	public static void setRecordingInterval(int interval) {
		PEER_UPDATE_INTERVAL = interval;
		
	}


}
