/**
 * Dasu Project
 *
 * File:         StaticLocalPeerStat.java
 * RCS:          $Id$
 * Description:  StaticLocalPeerStat class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 17, 2010 at 7:42:41 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.stats
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.stats;

import java.util.ArrayList;
import java.util.LinkedList;

import org.gudy.azureus2.core3.config.COConfigurationListener;
import org.gudy.azureus2.core3.config.COConfigurationManager;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.Main;


/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * Local Peer Statistics class ...
 */

public class StaticLocalPeerStat {
	
	/** in seconds */
	public long uptime; 
	public int maxUploadRate = -1;
	public int maxDownloadRate = -1;

	public int maxPeerConnections;
	public int maxPeerConnectionsPerTorrent;
	public int maxPeerConnectionsPerTorrentSeeding = -1; /* -1 means not enabled, so value should be ignored */
	public int maxUploadRateSeeding = -1; /* -1 means not enabled, so value should be ignored */
	public int maxSeedsPerTorrent;
	public boolean lazyBitfield;
	public boolean requireEncryption;
	public String encryptionMinLevel;
	public boolean allowNonEncryptedOut;
	public boolean allowNonEncryptedIn;
	public boolean useCryptoPortExtension;
	protected Boolean configChanged = false;
	
	public int tcpPortIn;
	public int udpPortIn;
	
	
	ArrayList<DynamicLocalPeerStat> stats = new ArrayList<DynamicLocalPeerStat>();
	
	public StaticLocalPeerStat() {
		
		COConfigurationManager.addListener(new COConfigurationListener() {

			public void configurationSaved() {
				configChanged = true;
				update();
			}
			
		});
		
	}
	
	public void update() {
		
		maxUploadRate = COConfigurationManager.getIntParameter("Max Download Speed KBs");
		maxDownloadRate = COConfigurationManager.getIntParameter("Max Upload Speed KBs");
		
		uptime = Main.getPluginInterface().getDownloadManager().getStats().getSessionUptimeSeconds();
		
		maxPeerConnections = COConfigurationManager.getIntParameter("Max.Peer.Connections.Total");
		maxPeerConnectionsPerTorrent = COConfigurationManager.getIntParameter("Max.Peer.Connections.Per.Torrent");
		if (COConfigurationManager.getBooleanParameter("Max.Peer.Connections.Per.Torrent.When.Seeding.Enable"))
			maxPeerConnectionsPerTorrentSeeding = COConfigurationManager.getIntParameter("Max.Peer.Connections.Per.Torrent.When.Seeding");
		if (COConfigurationManager.getBooleanParameter("enable.seedingonly.maxuploads")) 
			maxUploadRateSeeding = COConfigurationManager.getIntParameter("Max Uploads Seeding");	
		maxSeedsPerTorrent = COConfigurationManager.getIntParameter("Max Seeds Per Torrent");	
		lazyBitfield = COConfigurationManager.getBooleanParameter("Use Lazy Bitfield");
		
		requireEncryption = COConfigurationManager.getBooleanParameter("network.transport.encrypted.require");
		encryptionMinLevel = COConfigurationManager.getStringParameter("network.transport.encrypted.min_level");
		allowNonEncryptedOut = COConfigurationManager.getBooleanParameter("network.transport.encrypted.fallback.outgoing");
		allowNonEncryptedIn = COConfigurationManager.getBooleanParameter("network.transport.encrypted.fallback.incoming");
		useCryptoPortExtension = COConfigurationManager.getBooleanParameter("network.transport.encrypted.use.crypto.port");
		tcpPortIn = COConfigurationManager.getIntParameter("TCP.Listen.Port");
		udpPortIn = COConfigurationManager.getIntParameter("UDP.Listen.Port");
		
	}

	/**
	 * Periodic flushing of accumulated data.
	 *
	 */
	public void flush() {			
		
		synchronized (stats){
			LinkedList<DynamicLocalPeerStat> toRemove = new LinkedList<DynamicLocalPeerStat>();
			for (DynamicLocalPeerStat dlps : stats){
				if (DasuConfiguration.getInstance().isRecordStats()){
					Main.getReporter().insertLocalPeerStatDynamic(
						Statistics.getInstance().getSelfId(), 
						dlps.currentTime, dlps.uploadRate, dlps.downloadRate, 
						dlps.maxDownloadRate, dlps.maxUploadRate, dlps.autoSpeed, dlps.rst, 
						dlps.numConnections, dlps.passiveOpens, dlps.activeOpens, dlps.failedConns,
						dlps.numTorrents, dlps.numSeedingTorrents, dlps.numLeechingTorrents, dlps.numConnectedSeeds, dlps.numConnectedLeechers,
						dlps.bytesSentLpsd, dlps.bytesReceivedLpsd, dlps.bytesSentNetstat, dlps.bytesReceivedNetstat, dlps.swarmPeerSpeed,
						dlps.UPnPNumberOfActiveConnections, dlps.UPnPTotalBytesSent, dlps.UPnPTotalBytesReceived, dlps.UPnPTotalPacketsSent,
						dlps.UPnPTotalPacketsReceived, dlps.UPnPUptime);
				}
			}
			stats.clear();
		}
		
		
	}

	/**
	 * Reporting of data that is updated once per session.
	 *
	 */
	public void stop() {
		// TODO check that this reports correctly when app is closed
		if (DasuConfiguration.getInstance().isRecordStats()) {
			long sessionTime = -1;
			try {
				sessionTime = Main.getPluginInterface().getDownloadManager().getStats().getSessionUptimeSeconds();
			} catch (Exception e){
				e.printStackTrace();
			}
			Main.getReporter().insertLocalPeerStatStatic(Statistics.getInstance().getSelfId(), 			
				sessionTime, 
				maxUploadRate, maxDownloadRate, maxPeerConnections, maxPeerConnectionsPerTorrent, 
				maxPeerConnectionsPerTorrentSeeding, maxUploadRateSeeding, maxSeedsPerTorrent, lazyBitfield?1:0, 
				requireEncryption?1:0, encryptionMinLevel, allowNonEncryptedOut?1:0, allowNonEncryptedIn?1:0,
				useCryptoPortExtension?1:0, tcpPortIn, udpPortIn);
		}
		flush();
		//stats = null;
		
	}

	public void add(DynamicLocalPeerStat dlps) {			
		if (stats.size()>=PeerStatistics.getInstance().MAX_DLPS_ENTRIES){
			flush();
			Main.getReporter().disconnect();
		}
		synchronized(stats){
			if (stats.size()>PeerStatistics.getInstance().MAX_DLPS_ENTRIES) stats.remove(0);
			stats.add(dlps);
		}
	}

	public void flushStatic() {
		if (DasuConfiguration.getInstance().isRecordStats()) {
			Main.getReporter().insertLocalPeerStatStatic(Statistics.getInstance().getSelfId(), 			
					Main.getPluginInterface().getDownloadManager().getStats().getSessionUptimeSeconds(), 
				maxUploadRate, maxDownloadRate, maxPeerConnections, maxPeerConnectionsPerTorrent, 
				maxPeerConnectionsPerTorrentSeeding, maxUploadRateSeeding, maxSeedsPerTorrent, lazyBitfield?1:0, 
				requireEncryption?1:0, encryptionMinLevel, allowNonEncryptedOut?1:0, allowNonEncryptedIn?1:0,
				useCryptoPortExtension?1:0, tcpPortIn, udpPortIn);
		}

	}
}