/**
 * Dasu Project
 *
 * File:         Statistics.java
 * RCS:          $Id$
 * Description:  Statistics class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Jun 14, 2006 at 7:51:32 AM
 * Language:     Java
 * Package:      edu.northwestern.news.stats
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2006, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.stats;

import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.database.DatabaseReporter;
import edu.northwestern.dasu.drools.FactDnsProbeResult;
import edu.northwestern.dasu.drools.FactHttpProbeResult;
import edu.northwestern.dasu.drools.FactNdtProbeResult;
import edu.northwestern.dasu.drools.FactScrapeProbeResult;
import edu.northwestern.dasu.drools.FactSpeedTestResult;
import edu.northwestern.dasu.drools.FactTraceResult;
import edu.northwestern.dasu.drools.FactUpnpTestResult;
import edu.northwestern.dasu.measurement.TestType;
import edu.northwestern.dasu.measurement.TraceEntry;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.util.HashSetCache;
import edu.northwestern.dasu.util.Util;


/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 * 
 *         The Statistics class manages statistics collected by this plugin.
 */
public class Statistics implements ITimerEventPerformer {
	
	private static Statistics self;
	
	public static Hashtable<String, Integer> peerMap = null; 
	public static Hashtable<Integer, Integer> peerIds = new Hashtable<Integer, Integer>();;
	public static Hashtable<String, Integer> peerClientMap = null;
	public static Hashtable<Integer, Integer> peerClientIds = new Hashtable<Integer, Integer>();;
	public static Hashtable<String, Integer> routerMap = null;
	public static Hashtable<Integer, Integer> routerIds = new Hashtable<Integer, Integer>();;
	public static Hashtable<String, Integer> urlMap = null;
	public static Hashtable<Integer, Integer> urlIds = new Hashtable<Integer, Integer>();;
	public static Hashtable<String, Integer> dnsMap = null;
	public static Hashtable<Integer, Integer> dnsIds = new Hashtable<Integer, Integer>();;
	private static final int MAX_TRACE_RESULTS = 100;
	private static final int MAX_PING_RESULTS = 100;
	private static final int MAX_HTTP_RESULTS = 10;
	private static final int MAX_TEST_RESULTS = 1;
	private static final int MAX_RSTSTAT_RESULTS = 50;
	private static final int MAX_RPCSTAT_RESULTS = 50;
	private final long UPDATE_INTERVAL;
	private boolean DEBUG = false;
	
	public static final Object sessionRestartMonitor = new Object();
	
	/**
	 * List of things that can be reported.
	 * 
	 */
	
	/** list of all ping results */
	private static HashSetCache<PingResultDasu> pingResultsDasu;
	/** list of all http results */
	private static HashSetCache<FactHttpProbeResult> httpResultsDasu;
	/** list of all speedTest results */
	private static HashSetCache<FactSpeedTestResult> speedTestResults;
	/** list of all upnpTestResults results */
	private static HashSetCache<FactUpnpTestResult> upnpTestResults;
	/** list of all data reset events */
	private static HashSetCache<RSTStat> RSTStatResults;
	/** list of all connection establishment events */
	private static HashSetCache<RemotePeerConnectionStat> RpcStatResults;
	/** list of all peers */
	private static HashSet<String> peers;
	/** list of all trace route results */
	private static HashSetCache<FactTraceResult> traceResults;
	/** list of all scrape results */
	private static HashSetCache<FactScrapeProbeResult> scrapeResults;
	/** list of all dns results */
	private static HashSetCache<FactDnsProbeResult> dnsResults;
	
	/**
	 * Stuff for the database
	 */
	static private DatabaseReporter db = null;
	
	/** random object for jitter */
	static private Random r = new Random();
	
	private HashSetCache<Exception> errors;
	private int selfId = -1;
	private ITimer timer;
	
	private Statistics() {

		restartSession();
		
		if (r == null) {
			r = new Random();
		}
		peers = new HashSet<String>();
		traceResults = new HashSetCache<FactTraceResult>(MAX_TRACE_RESULTS);
		pingResultsDasu = new HashSetCache<PingResultDasu>(MAX_PING_RESULTS);
		scrapeResults = new HashSetCache<FactScrapeProbeResult>(MAX_TRACE_RESULTS);
		dnsResults = new HashSetCache<FactDnsProbeResult>(MAX_TRACE_RESULTS);
		httpResultsDasu = new HashSetCache<FactHttpProbeResult>(
				MAX_HTTP_RESULTS);
		speedTestResults = new HashSetCache<FactSpeedTestResult>(
				MAX_TEST_RESULTS);
		upnpTestResults = new HashSetCache<FactUpnpTestResult>(MAX_TEST_RESULTS);
		RSTStatResults = new HashSetCache<RSTStat>(MAX_RSTSTAT_RESULTS);
		RpcStatResults = new HashSetCache<RemotePeerConnectionStat>(MAX_RPCSTAT_RESULTS);

		errors = new HashSetCache<Exception>();
		UPDATE_INTERVAL = DasuConfiguration.getInstance()
				.getDbUpdateFreqSec() * 1000;

		timer = Main.createTimer("DasuStatistics");
		timer.addPeriodicEvent(UPDATE_INTERVAL, this);
	}

	public static synchronized Statistics getInstance() {
		if (self == null) {
			self = new Statistics();
		}

		return self;
	}
	
	public void restartSession() {
		
		synchronized(sessionRestartMonitor) {
			
			Main.sessionStartTime = System.nanoTime();
			
			DasuManager.getPersistentDataManager().getDbIdMaps().put(Main.sessionStartTime, new Hashtable<String, Hashtable<String, Integer>>());
			DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).put("peerMap", new Hashtable<String, Integer>());
			DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).put("routerMap", new Hashtable<String, Integer>());
			DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).put("dnsMap", new Hashtable<String, Integer>());
			DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).put("peerClientMap", new Hashtable<String, Integer>());
			DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).put("urlMap", new Hashtable<String, Integer>());
			
			peerMap = DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).get("peerMap");
			routerMap = DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).get("routerMap");
			dnsMap = DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).get("dnsMap");
			peerClientMap = DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).get("peerClientMap");
			urlMap = DasuManager.getPersistentDataManager().getDbIdMaps().get(Main.sessionStartTime).get("urlMap");
		}
		
	}

	private boolean addPeerClient(String peerClient) {
		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect())
				return false;
		}

		synchronized (peerClientMap) {
			if (peerClientMap.get(peerClient) == null) {
				peerClientMap.put(peerClient, peerClientMap.size());
			}
		}

		return true;
	}
	
	protected boolean addPeer(String peer) {
		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect())
				return false;
		}

		synchronized (peerMap) {
			if (peerMap.get(peer) == null) {
				peerMap.put(peer, peerMap.size());
			}
		}

		return peers.add(peer);
	}
	
	private boolean addDns(String server) {
		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect())
				return false;
		}

		if (server == null )
			server = "";

		synchronized (dnsMap) {
			if (dnsMap.get(server) == null) {
				dnsMap.put(server, dnsMap.size());
			}
		}
		
		return true;
		
	}
	
	private boolean addUrl(String url) {
		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect())
				return false;
		}

		int p;

		synchronized (urlMap) {
			if (urlMap.get(url) == null) {
				urlMap.put(url, urlMap.size());
			}			
		}
		
		return true;
		
	}

	public void addDnsTimeout(String dnsIp, String myIp, long startTime, long endTime, boolean giveUp) {
		  		
		int myDnsId = -1;
    	int myId = -1;
    	    	
    	myDnsId = dnsMap.get(dnsIp);
 		myId = peerMap.get(myIp);
  
 		Main.getReporter().insertDnsTimeout(myDnsId, myId, startTime, endTime, boolToInt(giveUp));
        
    }

	/**
	 * Called right before plugin unloads.
	 */
	public void flush() {
		if (DasuConfiguration.getInstance().isRecordStats()
				&& DasuConfiguration.getInstance().isSendToDatabase()) {

			Main.getReporter().setExtendedConnection(true);
			
			commitPingsDasu();
			commitHttpDasu();
			commitDnsDasu();
			commitUpnpTest();
			commitSpeedTest();
			commitRSTStat();
			commitRemotePeerConnections();
			reportTraces();
			reportScrapes();
			commitErrors();
			
			Main.getReporter().setExtendedConnection(false);

		}
		Main.getReporter().disconnect();
	}

	public void addPingResultDasu(String sourceIp, String destIp, Double rtt,
			long timeInMillis, int destType, int taskId) {

		PingResultDasu pr = new PingResultDasu(sourceIp, destIp, rtt,
				timeInMillis, destType, taskId);

		if (DEBUG) {
			System.out.println("AddPingResultDasu " + "destIp: " + destIp
					+ " sourceIp: " + sourceIp + " rtt: " + rtt + " time: "
					+ timeInMillis + ";" + " destType: " + destType + ";"
					+ "taskId: " + taskId);
		}

		synchronized (pingResultsDasu) {
			pingResultsDasu.add(pr);
		}
		synchronized (pingResultsDasu) {
			if (pingResultsDasu.size() >= MAX_PING_RESULTS) {
				commitPingsDasu();
			}
		}

	}

	public void addHttpResultDasu(FactHttpProbeResult httpResult) {

		if (DEBUG) {
			System.out
					.println("AddHttpResultDasu " + "url: " + httpResult.url
							+ " httpResult.downloadDuration: "
							+ httpResult.downloadDuration
							+ " httpResult.time: " + httpResult.time + ";"
							+ " downloadSpeed: " + httpResult.downloadSpeed
							+ ";" + "noBytes: " + httpResult.noBytes + ";"
							+ "taskId: " + httpResult.taskId);
		}
		
		addUrl(httpResult.url);

		synchronized (httpResultsDasu) {
			httpResultsDasu.add(httpResult);
		}
		synchronized (httpResultsDasu) {
			if (httpResultsDasu.size() >= MAX_HTTP_RESULTS) {
				commitHttpDasu();
			}
		}

	}
	
	public void addNdtResultDasu(FactNdtProbeResult ndtResult) {
		commitNdtDasu(ndtResult);
	}

	public void addSpeedTestResult(FactSpeedTestResult result) {

		if (DEBUG) {

			System.out.println("AddSpeedTestResult " + " dl: "
					+ result.downloadSpeed + " " + " ul: " + result.uploadSpeed
					+ " " + " time: " + new Date(result.testTime) + " "
					+ " success: " + result.success + "************");
		}

		synchronized (speedTestResults) {
			speedTestResults.add(result);

			if (speedTestResults.size() >= MAX_TEST_RESULTS)
				commitSpeedTest();
			
		}

	}

	public void addUpnpTestResult(FactUpnpTestResult result) {

		if (DEBUG) {

			System.out.println("UpnpTestResult " + " port: " + result.port
					+ " " + " protocol: " + result.protocol + " " + " time: "
					+ new Date(result.testTime) + " " + " success: "
					+ result.success + "************");
		}

		synchronized (upnpTestResults) {
			upnpTestResults.add(result);
		}
		synchronized (upnpTestResults) {
			if (upnpTestResults.size() >= MAX_TEST_RESULTS) {
				commitUpnpTest();
			}
		}

	}

	public void addRSTStatResult(RSTStat result) {

		if (DEBUG) {

			System.out.println("addConnectionReset " + " did: "
					+ result.downloadId + " " + " numRst: " + result.numRst
					+ " " + " time: " + new Date(result.timeRecorded) + " "
					+ " PeerIp: " + result.peerIp + "************");
		}

		synchronized (Statistics.peerMap) {
			if (!Statistics.peerMap.containsKey(result.peerIp)) {
				Statistics.peerMap.put(result.peerIp, peerMap.size() + 1);
			}
		}

		synchronized (RSTStatResults) {
			RSTStatResults.add(result);
		}
		synchronized (RSTStatResults) {
			if (RSTStatResults.size() >= MAX_RSTSTAT_RESULTS) {
				commitRSTStat();
			}
		}

	}
	
	public void addRpcStatResult(RemotePeerConnectionStat result) {

		if (DEBUG) {

			System.out.println("addRemoteConnection " + " did: "
					+ " " + " time: " + new Date(result.timeRecorded) + " "
					+ " PeerIp: " + result.peerIp + "************");
		}

		synchronized (Statistics.peerMap) {
			if (!Statistics.peerMap.containsKey(result.peerIp)) {
				Statistics.peerMap.put(result.peerIp, peerMap.size() + 1);
			}
		}

		synchronized (RpcStatResults) {
			RpcStatResults.add(result);
		}
		synchronized (RpcStatResults) {
			if (RpcStatResults.size() >= MAX_RPCSTAT_RESULTS) {
				commitRemotePeerConnections();
			}
		}

	}
	
	public void addScrapeResult(FactScrapeProbeResult result) {

		if (DEBUG) {

			System.out.println("addScrapeResult " + " type " 
					+ result.type + " time: " + new Date(result.time) + " "
					+ " numLeechers: " + result.getNumLeechers()
					+ " numSeeders: "  + result.getNumSeeds() + "************");
		}
		
		
		synchronized (scrapeResults) {
			if (scrapeResults.size() > MAX_TRACE_RESULTS) {
				reportScrapes();
			}
			scrapeResults.add(result);
		}

	}
	
	public void addDnsResult(FactDnsProbeResult result) {

		if (DEBUG) {

			System.out.println("addDnsResult " + " server " 
					+ result.getServer() + " toLookup: " + result.getToLookup() + " "
					+ " response: " + result.getSimpleResponse() + " ************");
		}

		addDns(result.getServer());
		synchronized (dnsResults) {
			if (dnsResults.size() > MAX_TRACE_RESULTS) {
				commitDnsDasu();
			}
			dnsResults.add(result);
		}

	}

	private void commitDnsDasu() {

		if (DEBUG)
			System.out.println("commitDnsDasu()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}
		
		LinkedList<FactDnsProbeResult> dnsToRemove = new LinkedList<FactDnsProbeResult>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;
		
		synchronized (dnsResults) {
			for (FactDnsProbeResult h : dnsResults) {
				if (db == null)
					db = Main.getReporter();
				
				if (h.getServer() == null)
					h.setServer("");
				
				int dnsId = dnsMap.get(h.getServer());

				if (Main.getReporter().insertDnsResult(myId,
						dnsId, h.getRequestTime(), h.getResponseTime(), h.getToLookup(),
						h.getType(), h.getOptions(), h.getSimpleResponse(), h.isFailed(),
						h.getTaskId()))
					dnsToRemove.add(h);
			}
		}

		synchronized (dnsResults) {
			dnsResults.removeAll(dnsToRemove);
		}
		Main.getReporter().disconnect();

	}
	
	public void reportScrapes() {

		if (DEBUG)
			System.out.println("reportScrapesDasu()");
		
		if (db == null) {
			db = Main.getReporter();
		}
		if (!Main.getReporter().connect()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return;
		
		while (scrapeResults.size() > 0) {
			FactScrapeProbeResult scrape = null;
			// update map of ids from database
			synchronized (scrapeResults) {
				Iterator<FactScrapeProbeResult> it = scrapeResults.iterator();
				scrape = it.next();
				it.remove();
			}
		
			for (String ip : scrape.getIps())
				addPeerForLookup(ip);
			
			int myId = updateIdMap();
	
			synchronized (Main.getReporter()) {
				// make top-level entry
				long entryId = Main.getReporter().insertScrapeEntry(
						scrape.type, myId, scrape.tracker, scrape.hash, scrape.getIps().size(), 
						scrape.numSeeds, scrape.numLeechers, scrape.taskId, scrape.time);
	
				// if valid, insert scrape entries
				if (entryId > -1) {
					for (int i=0; i < scrape.getIps().size(); i++) {
			
						String ip = scrape.getIps().get(i);
			
						// get mappings for storing entry
						addPeer(ip);
						int peerid = peerMap.get(ip);
					
						// insert data
						Main.getReporter().insertScrapeEntryData(entryId,
								peerid, scrape.getPorts().get(i), scrape.getUdpPorts().get(i));
							
					}
				} // end case valid entry
				else {
					scrapeResults.add(scrape);
					break; // exit for now
				}
				
			}// end synchronized part
		}
		
		Main.getReporter().disconnect();

	}

	/**
    *
    */
	private void commitPingsDasu() {

		if (DEBUG)
			System.out.println("commitPingsDasu()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<PingResultDasu> pingsToRemove = new LinkedList<PingResultDasu>();

		synchronized (pingResultsDasu) {
			
			for (PingResultDasu p : pingResultsDasu) {
				if (db == null) {
					db = Main.getReporter();
				}

				p.sourceId = peerMap.get(p.sourceIp);

				if (p.destType == 1)
					p.destId = peerMap.get(p.destIp);
				
				if (Main.getReporter().insertPingResultDasu(
						p.sourceId, p.destId, p.rtt, p.time, p.destType,
						p.taskId))
					pingsToRemove.add(p);

			}
		}

		synchronized (pingResultsDasu) {
			pingResultsDasu.removeAll(pingsToRemove);
		}
		Main.getReporter().disconnect();

	}

	public int makeUUIDid(String uuid) {
		return Main.getReporter().makeId("uuids", "uuid", uuid);
	}
	
	/**
    *
    */
	private void commitHttpDasu() {

		if (DEBUG)
			System.out.println("commitHttpDasu()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}
		
		LinkedList<FactHttpProbeResult> httpsToRemove = new LinkedList<FactHttpProbeResult>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;
		
		synchronized (httpResultsDasu) {
			for (FactHttpProbeResult h : httpResultsDasu) {
				if (db == null)
					db = Main.getReporter();
				
				int urlId = urlMap.get(h.url);

				if (Main.getReporter().insertHttpResultDasu(myId,
						urlId, h.downloadDuration, h.downloadSpeed, h.time,
						h.noBytes, h.taskId))
					httpsToRemove.add(h);

			}
		}

		synchronized (httpResultsDasu) {
			httpResultsDasu.removeAll(httpsToRemove);
		}
		Main.getReporter().disconnect();

	}

	/**
    *
    */
	public void commitNdtDasu(FactNdtProbeResult ndt) {

		if (DEBUG)
			System.out.println("commitNdtDasu()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;
		
//		if (ndt.failed == true)
//			return;

		Main.getReporter().insertNdtResultDasu(myId,
				ndt.osName, ndt.osArch, ndt.osVer, ndt.interfaceName, ndt.javaVer, 
				ndt.server, ndt.c2sRate, ndt.s2cRate, ndt.mylink, ndt.mismatch, ndt.avgrtt, 
				ndt.packetSize, ndt.connectionStalledTimes, ndt.connectionIdlePercentage, 
				boolToInt(ndt.c2sPacketQueuing), boolToInt(ndt.s2cPacketQueuing), 
				ndt.s2cThroughputPacketQueueingDetectedPercentage, 
				ndt.connectionReceiverLimitedPercentage, ndt.connectionSenderLimitedPercentage,
				ndt.connectionNetworkLimitedPercentage, boolToInt(ndt.RFC2018), boolToInt(ndt.RFC896), 
				boolToInt(ndt.RFC3168), boolToInt(ndt.RFC1323TimeStamping), boolToInt(ndt.RFC1323WindowScaling), 
				ndt.serverScalingFactor, ndt.clientScalingFactor, boolToInt(ndt.serverBehindFirewall), 
				boolToInt(ndt.clientBehindFirewall), ndt.theoreticalNetworkLimit, ndt.serverBuffer, 
				ndt.serverThroughputLimit, ndt.clientBuffer, ndt.clientThroughputLimit, 
				ndt.time, boolToInt(ndt.failed), ndt.errorMsg, ndt.c2sRateNetstat, ndt.s2cRateNetstat, 
				ndt.networkFlowControlThroughputLimit, boolToInt(ndt.congestion), ndt.congestionSignals, ndt.fastRetransmit, ndt.timeouts,
				ndt.outOfOrderPercentage, ndt.pktsRetrans, ndt.dupAcksIn, ndt.sACKsRcvd, ndt.packetLoss);

		Main.getReporter().disconnect();

	}
	
	
	/**
    *
    */
	public void commitSpeedTest() {

		if (DEBUG)
			System.out.println("commitSpeedTest()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<FactSpeedTestResult> speedTestsToRemove = new LinkedList<FactSpeedTestResult>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (speedTestResults) {
			for (FactSpeedTestResult s : speedTestResults) {
				if (db == null)
					db = Main.getReporter();

				if (!s.isSuccess()) {
					speedTestsToRemove.add(s);
					continue;
				}
				
				if (Main.getReporter().insertSpeedTestResult(myId,
						s.testTime, s.downloadSpeed, s.uploadSpeed,
						boolToInt(s.encryption), testTypeToString(s.testType), 
						s.downloadSpeedNetstat, s.uploadSpeedNetstat))
					speedTestsToRemove.add(s);

			}
		}

		synchronized (speedTestResults) {
			speedTestResults.removeAll(speedTestsToRemove);
		}
		Main.getReporter().disconnect();

	}

	private char testTypeToString(TestType testType) {
		char ret = 'u';
		if ((testType.compareTo(TestType.AZUREUS_DL) == 0)
				|| (testType.compareTo(TestType.AZUREUS_DL_ENCRYPTION)) == 0)
			ret = 'd';

		return ret;
	}


	public int updateIdMap() {

		String myIp = Main.getPublicIpAddress();

		if (!peerMap.containsKey(myIp))			
			peerMap.put(myIp, peerMap.size());

		return peerMap.get(myIp);
		
	}
	/**
    *
    */
	public void commitUpnpTest() {

		if (DEBUG)
			System.out.println("commitUpnpTest()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<FactUpnpTestResult> upnpTestsToRemove = new LinkedList<FactUpnpTestResult>();

//		batchProcessIds();		
//		String myIp = Main.getPublicIpAddress();
//		int myId = -1;
//		// update map of ids from database
//		if (!peerIds.containsKey(peerMap.get(myIp))) {
//			myId = Main.getReporter().makeId("peerips", "ip", myIp);
//
//			if (myId == -1)
//				return;
//
//			peerIds.put(peerMap.get(myIp), myId);
//		} else
//			myId = peerIds.get(peerMap.get(myIp));
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (upnpTestResults) {
			for (FactUpnpTestResult u : upnpTestResults) {
				if (db == null)
					db = Main.getReporter();

				if (Main.getReporter().insertUpnpTestResult(myId,
						u.testTime, u.port, u.protocol, boolToInt(u.success)))
					upnpTestsToRemove.add(u);

			}
		}

		synchronized (upnpTestResults) {
			upnpTestResults.removeAll(upnpTestsToRemove);
		}
		Main.getReporter().disconnect();

	}

	private int boolToInt(boolean b) {
		if (b)
			return 1;

		return 0;
	}

	/**
    *
    */
	private void commitRSTStat() {

		if (DEBUG)
			System.out.println("commitRSTStat()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<RSTStat> RSTStatsToRemove = new LinkedList<RSTStat>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (RSTStatResults) {
			for (RSTStat r : RSTStatResults) {
				if (db == null)
					db = Main.getReporter();

				int peerId = -1;

				peerId = peerMap.get(r.peerIp);

				if (Main.getReporter().insertRSTStat(peerId, myId, r.downloadId, r.numRst, r.timeRecorded,
					r.type, r.port, r.bitTorrentClient, r.source, boolToInt(r.isSeed), boolToInt(r.amSeed), 
					r.numSeeds, r.numPeers, r.encryption))
					RSTStatsToRemove.add(r);

			}
		}

		synchronized (RSTStatResults) {
			RSTStatResults.removeAll(RSTStatsToRemove);
		}
		Main.getReporter().disconnect();

	}
	
	/**
    *
    */
	private void commitRemotePeerConnections() {

		if (DEBUG)
			System.out.println("commitRemotePeerConnections()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}


		LinkedList<RemotePeerConnectionStat> RemotePeerRemove = new LinkedList<RemotePeerConnectionStat>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (RpcStatResults) {
			for (RemotePeerConnectionStat r : RpcStatResults) {
				if (db == null)
					db = Main.getReporter();

				int peerId = -1;

				peerId = peerMap.get(r.peerIp);

				if (Main.getReporter().insertRemotePeerConnectionStat(peerId, myId, r.timeRecorded,
					r.type, r.port, r.bitTorrentClient, boolToInt(r.isSeed), boolToInt(r.amSeed), 
					r.numSeeds, r.numPeers, r.encryption))
					RemotePeerRemove.add(r);

			}
		}

		synchronized (RpcStatResults) {
			RpcStatResults.removeAll(RemotePeerRemove);
		}
		Main.getReporter().disconnect();

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.gudy.azureus2.plugins.utils.UTTimerEventPerformer#perform(org.gudy
	 * .azureus2.plugins.utils.UTTimerEvent)
	 */
	public void perform(ITimerEvent event) {
		try {
			if (DasuConfiguration.getInstance().isRecordStats()
					&& DasuConfiguration.getInstance().isSendToDatabase()) {
				Main.getReporter().setExtendedConnection(true);

				reportTraces();
				traceResults.clear();
				reportScrapes();
				scrapeResults.clear();
				commitErrors();
				errors.clear();
				commitPingsDasu();
				pingResultsDasu.clear();
				commitHttpDasu();
				httpResultsDasu.clear();
				commitDnsDasu();
				dnsResults.clear();
				commitSpeedTest();
				speedTestResults.clear();
				commitUpnpTest();
				upnpTestResults.clear();
				commitRSTStat();
				RSTStatResults.clear();
				commitRemotePeerConnections();
				RpcStatResults.clear();

				Main.getReporter().setExtendedConnection(false);
				

			} else {
				errors.clear();
				traceResults.clear();
				pingResultsDasu.clear();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		Main.getReporter().disconnect();
	}

	private void commitErrors() {
		if (db == null) {
			db = Main.getReporter();
		}
		if (!Main.getReporter().connect())
			return;
		while (errors.size() > 0) {
			Exception e = errors.iterator().next();
			Main.getReporter().reportError(e);
			errors.remove(e);
		}

		Main.getReporter().disconnect();

	}

	public void addTraceRouteResult(FactTraceResult tr) {

		addPeerForLookup(tr.dest);
		
		for (int i=0; i< tr.entries.size(); i++) {
			TraceEntry te = tr.entries.get(i);
			
			for (int j = 0; j < te.router.length; j++) {
				String router;

				if (te.numRouters - 1 < j) {
					router = te.router[0];
				} else {
					router = te.router[j];
				}

				addRouterForLookup(router);
			}
		}
		
		synchronized (traceResults) {
			if (traceResults.size() > MAX_TRACE_RESULTS) {
				reportTraces();
			}
			traceResults.add(tr);
		}
		
	}
	
	private void reportTraces() {
		int sourceId = -1;
		int destId = -1;

		if (db == null) {
			db = Main.getReporter();
		}
		
		while (traceResults.size() > 0) {
			FactTraceResult tr = null;
			// update map of ids from database
			synchronized (traceResults) {
				Iterator<FactTraceResult> it = traceResults.iterator();
				tr = it.next();
				it.remove();
			}

			// get mappings for storing entry
			addPeer(tr.source);
			sourceId = peerMap.get(tr.source);

			addPeer(tr.dest);			
			destId = peerMap.get(tr.dest);

			int lastIndex = tr.entries.size() - 1;
			Float rtts[];
			for (int i = tr.entries.size() - 1; i >= 0; i--) {
				rtts = tr.entries.get(i).rtt;
				if (rtts[0] + rtts[1] + rtts[2] == -3) {
					lastIndex = i;
				} else {
					break;
				}
			}

			int taskId = tr.taskId;

			synchronized (Main.getReporter()) {
				// make top-level entry
				long entryId = Main.getReporter().insertTraceEntry(
						sourceId, destId, tr.timestamp, lastIndex, taskId);

				// if valid, insert trace route entries
				if (entryId > -1) {

					TraceEntry te;
					for (int j = 0; j <= lastIndex; j++) {
						te = tr.entries.get(j);
						int myRouterIds[] = new int[3];
						for (int i = 0; i < te.router.length; i++) {
							String router;

							if (te.numRouters - 1 < i) {
								router = te.router[0];
							} else {
								router = te.router[i];
							}

							if (!routerMap.containsKey(router)) {
								routerMap.put(router, routerMap.size());
							}
							myRouterIds[i] = routerMap.get(router);
						}
						// insert data
						Main.getReporter().insertTraceEntryData(entryId,
								myRouterIds, te);
					} // end for trace entries

				} // end case valid entry
				else {
					traceResults.add(tr);
					break; // exit for now
				}
			} // end synchronized part

		}

		Main.getReporter().disconnect();

	}

	public int getPeerId(String ip) {
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return -1;
		
		if (peerMap.get(ip) == null) {
			if (!addPeer(ip))
				return -1;
		}

		return (peerMap.get(ip));
	}
	
	public int getDnsId(String ip) {
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return -1;
		
		if (dnsMap.get(ip) == null) {
			if (!addDns(ip))
				return -1;
		}

		return dnsMap.get(ip);
	}


	public int getPeerClientId(String client) {
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return -1;
		if (client.matches("(.*)\\[(.*)\\]") || client.contains("Unknown")
				|| client.contains("Inconnu") || client.contains("Ukjent")
				|| client.contains("Unbekannt")
				|| client.contains("Sconosciuto")) {
			if (client.toLowerCase().contains("exeem"))
				client = "Exeem";
			else
				client = "Unknown";
		}

		if (peerClientMap.get(client) == null) {
			if (!addPeerClient(client))
				return -1;
		}

		return peerClientMap.get(client);
	}

	public int getSelfId() {
		// if (selfId == -1){
		selfId = getPeerId(Main.getPublicIpAddress());
		// }
		return selfId;
	}

	public void stop() {

		if (timer != null) {
			timer.destroy();
		}
		Main.getReporter().stop();
		
		self = null;
		db = null;
		traceResults.clear();
		scrapeResults.clear();

		r = null;
	}
	
	/**
	 * Add peer only for lookup, don't connect to DB to get id
	 * 
	 * @param peer
	 * @return
	 */
	public void addPeerForLookup(String peer) {

		synchronized (peerMap) {
			if (peerMap.get(peer) == null) {
				peerMap.put(peer, peerMap.size());
			}

		}
		peers.add(peer);
	}

	/**
	 * Register peer client name for DB lookup
	 * 
	 * @param peerClient
	 */
	protected void addPeerClientForLookup(String peerClient) {
		synchronized (peerClientMap) {
			if (peerClientMap.get(peerClient) == null) {
				peerClientMap.put(peerClient, peerClientMap.size());
			}
		}

	}

	/**
	 * Register peer client name for DB lookup
	 * 
	 * @param peerClient
	 */
	public void addRouterForLookup(String routerIp) {
		synchronized (routerMap) {
			if (routerMap.get(routerIp) == null) {
				routerMap.put(routerIp, routerMap.size());
			}
		}

	}

	public void reportIpChange(String oldIp, String newIp, String localIp,
			long currentGMTTime) {
		int pub = getPeerId(oldIp);
		int nouvel = getPeerId(newIp);
		int local = getPeerId(localIp);
		Main.getReporter().reportIpChange(pub, nouvel, local,
				currentGMTTime);

	}
	
	/**
    *
    */
	public void commitUserComments(String dids, String comment) {

		if (DEBUG)
			System.out.println("commitUserComments()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		Main.getReporter().insertUserReports(myId, dids, comment);

		Main.getReporter().disconnect();

	}
	
	/**
    *
    */
	public void commitTask(int taskId) {

		if (DEBUG)
			System.out.println("commitTask()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		Main.getReporter().insertTask(myId, Util.currentGMTTime(), taskId);

		Main.getReporter().disconnect();

	}

	/**
    *
    */
	public void commitIspInfo(String prefix, String ispname) {

		if (DEBUG)
			System.out.println("commitIspInfo()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		Main.getReporter().insertIspInfo(myId, prefix, ispname, Util.currentGMTTime());

		Main.getReporter().disconnect();

	}
	
	/**
    *
    */
	public void commitStaticUPnPInfo(HashMap<String, String> vals) {

		if (DEBUG)
			System.out.println("commitStaticUPnPInfo()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		String UpstreamMaxBitRate = vals.containsKey("UpstreamMaxBitRate") && vals.get("UpstreamMaxBitRate") != null ? vals.get("UpstreamMaxBitRate") : "";
		String DownstreamMaxBitRate = vals.containsKey("DownstreamMaxBitRate") && vals.get("DownstreamMaxBitRate") != null ? vals.get("DownstreamMaxBitRate") : "";
		String NATEnabled = vals.containsKey("NATEnabled") && vals.get("NATEnabled") != null ? vals.get("NATEnabled") : "";
		String Layer1UpstreamMaxBitRate = vals.containsKey("Layer1UpstreamMaxBitRate") && vals.get("Layer1UpstreamMaxBitRate") != null ? vals.get("Layer1UpstreamMaxBitRate") : "";
		String Layer1DownstreamMaxBitRate = vals.containsKey("Layer1DownstreamMaxBitRate") && vals.get("Layer1DownstreamMaxBitRate") != null ? vals.get("Layer1DownstreamMaxBitRate") : "";
		String PortMappingEnabled = vals.containsKey("PortMappingEnabled") && vals.get("PortMappingEnabled") != null ? vals.get("PortMappingEnabled") : "";
		String WANAccessType = vals.containsKey("WANAccessType") && vals.get("WANAccessType") != null ? vals.get("WANAccessType") : "";
		String MaximumActiveConnections = vals.containsKey("MaximumActiveConnections") && vals.get("MaximumActiveConnections") != null ? vals.get("MaximumActiveConnections") : "";
		String ExternalIPAddress = vals.containsKey("ExternalIPAddress") && vals.get("ExternalIPAddress") != null  ? vals.get("ExternalIPAddress") : "";

		Integer sourceId = -1;
		if ((ExternalIPAddress.length() > 0) && 
				ExternalIPAddress.matches("^[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}$")) {			
			
			addPeer(ExternalIPAddress);
			sourceId = peerMap.get(ExternalIPAddress);
			
		}
		
		if(NATEnabled != null){
			NATEnabled = NATEnabled.toLowerCase().equals("true") ? "1" : "0";
		} else {
			NATEnabled = "-1";
		}
		
		if(PortMappingEnabled != null){
			PortMappingEnabled = PortMappingEnabled.toLowerCase().equals("true") ? "1" : "0";
		} else {
			PortMappingEnabled = "-1";
		}
		
		int myId = updateIdMap();

		Main.getReporter().insertStaticUPnPInfo(myId, UpstreamMaxBitRate, DownstreamMaxBitRate, NATEnabled,
				Layer1UpstreamMaxBitRate, Layer1DownstreamMaxBitRate, PortMappingEnabled, WANAccessType, MaximumActiveConnections,
				sourceId.toString());

		Main.getReporter().disconnect();

	}
	
	/**
    *
    */
    public void commitInterference(String type, Long timestamp) {

		if (DEBUG)
			System.out.println("commitInterference()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		Main.getReporter().insertInterference(myId, type, timestamp);

		Main.getReporter().disconnect();

	}
    
   	/**
    *
    */
	public void commitInterferenceEvent(String type, Long timestamp, Long downloadId, Long droppedTo, float mean) {

		if (DEBUG)
			System.out.println("commitInterferenceEvent()");

		if (db == null) {
			db = Main.getReporter();
			if (!Main.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		Main.getReporter().insertInterferenceEvent(myId, type, timestamp, downloadId, droppedTo, mean);

		Main.getReporter().disconnect();

	}


}
