/**
 * Dasu Project
 *
 * File:         DasuCronScheduler.java
 * RCS:          $Id$
 * Description:  DasuCronScheduler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 4:55:53 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.timer
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.timer;

import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.concurrent.TimeUnit;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.drools.FactFireAction;
import edu.northwestern.dasu.measurement.AMCoordinatorV2;
import edu.northwestern.dasu.measurement.SourceType;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The DasuCronScheduler class ...
 */

public class DasuCronScheduler implements Runnable {

	private static DasuCronScheduler self = null;
	private static int schedulerInterval = 
		DasuConfiguration.getInstance().getSchedulerInterval();

	private static Hashtable<String, CronjobEntry> schedule = 
		DasuManager.getPersistentDataManager().getCronjobSchedule(); 
	
	private Boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;
	private static boolean DEBUG = false;

	private DasuCronScheduler() {
		isActive = true;
	}
	
    public static DasuCronScheduler getInstance() {
    	if (self == null)
    		self = new DasuCronScheduler();
	    return self;
    }
    
    @SuppressWarnings("unchecked")
	private void initialize() {
    	
    	if (schedule != null) {
    	
	    	/* Mark every entry as inActive */
    		Iterator iter = schedule.entrySet().iterator();
    		while (iter.hasNext()) {
    			Entry<String, CronjobEntry> ent = (Entry<String, CronjobEntry>) iter.next();
    			ent.getValue().setActive(false);
    			
    		}
    	}
    	
    	if (DEBUG)
    		schedulerInterval = 15;
    }
    
	public void registerCronJob (CronjobEntry cronJob) {
		
		if (DEBUG)
			System.out.println(cronJob);
		
		/* if job already exists, don't replace, mark as active */
		if (schedule.containsKey(cronJob.getId())) {
			if (schedule.get(cronJob.getId()).equals(cronJob)) {
				
				CronjobEntry cEntry = schedule.get(cronJob.getId());
				cEntry.setActive(true);
				schedule.put(cEntry.getId(), cEntry);
			}
			else
				schedule.put(cronJob.getId(), cronJob);
		}
		else 
			schedule.put(cronJob.getId(), cronJob);	
		
		if (DEBUG)
			System.out.println(schedule);
		
	}
	
	public void removeCronJob (CronjobEntry cronJob) {
		
		if (schedule.containsKey(cronJob.getId())) {
			if (schedule.get(cronJob.getId()).equals(cronJob))				
				schedule.remove(cronJob.getId());
		}
		
	}
	
	private long getDifference(Calendar a, Calendar b, TimeUnit units) {
		return 
		  units.convert(b.getTimeInMillis()- a.getTimeInMillis(), TimeUnit.MILLISECONDS);
	}
	
	public void run() {
		
		initialize();
		
		while (isActive) {
					
			if (DEBUG)
				System.out.println("DasuCronScheduler(): I'm awake!!");
			
			for ( Entry<String, CronjobEntry> entry: schedule.entrySet() ) {
				CronjobEntry event = entry.getValue();
				
				if (DEBUG)
					System.out.println(event.toString());
				
				if (event.isActive()) {
				
					boolean launchTest = false;
					long currTime = Util.currentGMTTime() + Util.dasuTimeOffset;
					Calendar currCal = Calendar.getInstance();
					Calendar lastCal = Calendar.getInstance();
					currCal.setTime(new Date(currTime));
					lastCal.setTime(new Date(event.getLastExecuted()));

					switch(event.getCronjobType()) {

						case ONCE_PER_MONTH:
							if (lastCal.get(Calendar.MONTH) - currCal.get(Calendar.MONTH) >= 1)
								launchTest = true;
							break;
							
						case ONCE_PER_WEEK:
							if (getDifference(lastCal, currCal, TimeUnit.DAYS) >= 7)
								launchTest = true;
							break;
							
						case ONCE_PER_DAY:
							if (getDifference(lastCal, currCal, TimeUnit.DAYS) >= 1)
								launchTest = true;
							break;
							
						case ONCE_PER_HOUR:
							if (getDifference(lastCal, currCal, TimeUnit.HOURS) >= 1)
								launchTest = true;
							break;
							
						case TWICE_PER_MONTH:
							if (getDifference(lastCal, currCal, TimeUnit.DAYS) >= 15)
								launchTest = true;
							break;

						case TWICE_PER_WEEK:
							if (getDifference(lastCal, currCal, TimeUnit.HOURS) >= 84)
								launchTest = true;
							break;
							
						case TWICE_PER_DAY:
							if (getDifference(lastCal, currCal, TimeUnit.HOURS) >= 12)
								launchTest = true;
							break;
							
						case TWICE_PER_HOUR:
							if (getDifference(lastCal, currCal, TimeUnit.MINUTES) >= 30)
								launchTest = true;
							break;

						default: 
							if ((event.getHours().contains(currCal.get(Calendar.HOUR_OF_DAY))) &&
									event.getDays().contains(currCal.get(Calendar.DAY_OF_WEEK))) {

								if ( (currCal.get(Calendar.HOUR_OF_DAY) != lastCal.get(Calendar.HOUR_OF_DAY)) ||
									(currCal.get(Calendar.DAY_OF_WEEK) != lastCal.get(Calendar.DAY_OF_WEEK))) {
										
									launchTest = true;
								}
							}
							
					}
										
					if (launchTest) {
						if (DEBUG)
							System.out.println("Launch Cronb Job: " + event.getCommandToExecute());
						
						event.setLastExecuted(currTime);
						AMCoordinatorV2.assertToEngine(SourceType.RE, new FactFireAction(event.getId(), event.getCommandToExecute()));
					}
					else {				
						if (DEBUG)
							System.out.println("Job already executed, wait for next occurence " +
									event.getCommandToExecute());
					}	
				}
			}
						
			synchronized (halt) {

				try {
					halt.wait(schedulerInterval*1000);
				} catch (InterruptedException e) {
				}	
			}
			
			if (terminate) {
				self = null;
				return;
			}

		}//while
		
	}
	
	public void stop() {
		isActive = false;
		synchronized(halt) {
			terminate = true;
			halt.notify();
		}
	}
	
}
