/**
 * Dasu Project
 *
 * File:         DasuCronScheduler.java
 * RCS:          $Id$
 * Description:  DasuCronScheduler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 4:55:53 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.timer
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.timer;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Properties;
import java.util.Map.Entry;

import org.gudy.azureus2.core3.logging.LogAlert;
import org.gudy.azureus2.core3.logging.Logger;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The DasuMessageManager class ...
 */

public class DasuMessageManager implements Runnable {

	public class Message {
		String text;
		int timeout;
		boolean repeatable;
		String id;
		
		public Message(String id, String text, int timeout, boolean repeatable) {
			this.id = id;
			this.text = text;
			this.timeout = timeout;
			this.repeatable = repeatable;
		}
		
		public boolean isRepeatable() {
			return repeatable;
		}
		
	}
	
	private static DasuMessageManager self = null;
	private static int schedulerInterval = 
		DasuConfiguration.getInstance().getMessageManagerInterval();
	
	private static LinkedHashMap<String, Long> messageLog =
		DasuManager.getPersistentDataManager().getMessageLog();
	private static HashSet<Message> messages = new HashSet<Message>();
	
	private Boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;
	private static boolean DEBUG = false;

	private DasuMessageManager() {
		super();
		isActive = true;
	}
	
    public static DasuMessageManager getInstance() {
    	if (self == null)
    		self = new DasuMessageManager();
	    return self;
    }
    
    
	public void addMessage (String id, String msg, int timeout, boolean repeatable) {
		
		/* if message already displayed, don't add it to the queue */
		synchronized(messageLog) {
			if (!messageLog.containsKey(id)) {
				messages.add(new Message(id, msg, timeout, repeatable));
				synchronized(halt) {
					halt.notify();
				}
			}
		}
	}
	
	public void run() {
				
		while (isActive) {
					
			if (DEBUG)
				System.out.println("DasuMessageManager(): I'm awake!!");
			
			checkMessages();
			
			Iterator<Message> iter = messages.iterator();
			while (iter.hasNext()) {
				
				Message m = iter.next();
  				iter.remove();
  				if (DasuConfiguration.getInstance().isEnablePopups()) {
  					Logger.log(new LogAlert(m.repeatable, LogAlert.AT_INFORMATION, "Dasu Plug-in:\n" + m.text, m.timeout));
  				}

				if (!m.isRepeatable()) { 
					synchronized(messageLog) {
						messageLog.put(m.id, System.currentTimeMillis());
					}
  					break;
  				}
			}
			
			synchronized (halt) {

				try {
				
					if (DEBUG)
						System.out.println("DasuMessageManager(): sleeping for " + schedulerInterval + " seconds...");

					
					halt.wait(schedulerInterval*1000);
				} catch (InterruptedException e) {
				}	
			}
			
			if (terminate) {
				self = null;
				return;
			}

		}//while
		
	}
	
	public void checkMessages() {

		String url = DasuManager.getInstance().getNewRandomDasuUrl();
		url = url + DasuConfiguration.DASU_MESSAGES;
		if (DEBUG)
			System.out.println("contacting: " + url);
		
		/*load file from server*/
		Properties props = Main.downloadFromURL(url,  2*1000);

		if (props != null) {

			try {
				Iterator<Entry<Object, Object>> iter = props.entrySet().iterator();
				while (iter.hasNext()) {
					Entry<Object, Object> ent = iter.next();
					String id = (String)ent.getKey();
					String msg = (String)ent.getValue();
	
					addMessage(id, msg, 0, false);
					
					if (DEBUG) {
						System.out.println("***New Message(): " + id + 
								" msg: " + "\"" + msg + "\"" + " timeout: " +
								0 + " repeat?: " +  false);
					}

				}
				
			} catch (Exception e) {
				if (DEBUG)
					System.out.println("DEBUG: " + e.getMessage());
				
			}
		}	
		
	}
			
	
	public void stop() {
		isActive = false;
		synchronized(halt) {
			terminate = true;
			halt.notify();
		}
	}
	
}
