/**
 * Dasu Project
 *
 * File:         UPNPDeviceHandler.java
 * RCS:          $Id$
 * Description:  UPNPDeviceHandler class (see below)
 * Author:       Zachary Bischof
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Feb 21, 2011 at 12:51:44 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.upnp
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.upnp;

import java.io.IOException;
import java.net.NetworkInterface;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sbbi.upnp.devices.UPNPDevice;
import net.sbbi.upnp.impls.InternetGatewayDevice;
import net.sbbi.upnp.messages.ActionMessage;
import net.sbbi.upnp.messages.UPNPMessageFactory;
import net.sbbi.upnp.messages.UPNPResponseException;
import net.sbbi.upnp.services.ServiceAction;
import net.sbbi.upnp.services.UPNPService;

/**
 * @author Zachary Bischof &lt;Zachary.Bischof@eecs.northwestern.edu&gt;
 *
 * The UPNPDeviceHandler class ...
 */

public class UPNPDeviceHandler {
	private static int timeoutUPNP = 3000;


	private static String[] xmlTags = {"manufacturer",
									   "manufacturerURL",
									   "modelName",
									   "modelNumber"};


	private static String matchKeyword(String xml, String keyword){
		Pattern patManufacturer = Pattern.compile("<" + keyword + ">(.*?)</"+keyword+">");
		Matcher m = patManufacturer.matcher(xml);
		if(m.find() && m.groupCount() >= 1){
			return m.group(1);
		}
		return null;
	}


	private static void parseXML(String xml, HashMap<String,String> tagsToVals){
		for(String tag : xmlTags){
			String result = matchKeyword(xml,tag);
			if(result != null){
				tagsToVals.put(tag, result);
			}
		}
	}


	public static InternetGatewayDevice[] getGatewayDevices(String interfaceName) {
		InternetGatewayDevice[] gatewayDevices = null;
		try {
			gatewayDevices = InternetGatewayDevice.getDevices(timeoutUPNP, 60,10,NetworkInterface.getByName(interfaceName));
			return gatewayDevices;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}
	

	public static boolean hasUPNPRouter(String interfaceName){
		if(getGatewayDevices(interfaceName) != null){
			return true;
		}
		return false;
	}


	public static String getDevDefLoc(String interfaceName){
		InternetGatewayDevice[] devices = getGatewayDevices(interfaceName);
		if(devices != null){
			return devices[0].getIGDRootDevice().getDeviceDefLocData();
		}
		return null;
	}

	@SuppressWarnings("unchecked")	
	public static HashMap<String,String> getUPNPStaticInfo(String interfaceName){
		HashSet<String> staticStates = new HashSet<String>();
		staticStates.add("UpstreamMaxBitRate");
		staticStates.add("DownstreamMaxBitRate");
		staticStates.add("PortMappingEnabled");
		staticStates.add("WANAccessType");

		HashSet<String> staticActions = new HashSet<String>();
		staticActions.add("GetNATRSIPStatus");
		staticActions.add("GetExternalIPAddress");
		staticActions.add("GetCommonLinkProperties");
		staticActions.add("GetConnectionTypeInfo");
		staticActions.add("GetDeviceInfo");		
		
		InternetGatewayDevice[] gatewayDevices = getGatewayDevices(interfaceName);
		if(gatewayDevices != null){
			//InternetGatewayDevice gatewayDevice = gatewayDevices[0];
			HashMap<String,String> staticTypesToVals = new HashMap<String,String>();
			for(InternetGatewayDevice gatewayDevice : gatewayDevices){
				System.out.flush();
				String deviceXml =  gatewayDevice.getIGDRootDevice().getDeviceDefLocData();
				parseXML(deviceXml,staticTypesToVals);
				List<UPNPDevice> upnpdevice = gatewayDevice.getIGDRootDevice().getChildDevices();
				for(Iterator<UPNPDevice> deviceIter = upnpdevice.iterator(); deviceIter.hasNext();){
					UPNPDevice mydev = deviceIter.next();
					List<UPNPService> devServices = mydev.getServices();
					for(Iterator<UPNPService> serviceIter = devServices.iterator(); serviceIter.hasNext();){
						UPNPService myserv = serviceIter.next();
						UPNPMessageFactory msgFactory = UPNPMessageFactory.getNewInstance(myserv);
						for(Iterator<String> actionIter = myserv.getAvailableActionsName(); actionIter.hasNext();){
							String action = actionIter.next();
							ServiceAction servact = myserv.getUPNPServiceAction(action);
							if (servact != null){
								if(staticActions.contains(action)){
									ActionMessage msg = msgFactory.getMessage(action);
									try {
										
										List<String> serviceActions = servact.getOutputActionArgumentsNames();
										if(serviceActions != null){
											for(int ii = 0; ii < serviceActions.size(); ii++){
												String servActArg = serviceActions.get(ii).toString();
												String response = msg.service().getOutActionArgumentValue(servActArg);
												staticTypesToVals.put(servActArg.replaceFirst("New", ""), response);
											}
										}
									} catch (IOException e) {
										// TODO Auto-generated catch block
										e.printStackTrace();
									} catch (UPNPResponseException e) {
										// TODO Auto-generated catch block
										e.printStackTrace();
									}
								}
							}
						}

						for(Iterator<String> stateIter = myserv.getAvailableStateVariableName(); stateIter.hasNext();){
							String stateName = stateIter.next();
							try {
								if(staticStates.contains(stateName)){
									staticTypesToVals.put(stateName, myserv.getUPNPServiceStateVariable(stateName).getValue());
								}
							} catch (UPNPResponseException e) {
								e.printStackTrace();
							} catch (IOException e) {
								e.printStackTrace();
							}
						}
					}
				}
			}
			return staticTypesToVals;
		}
		return null;
	}


	@SuppressWarnings("unchecked")
	public static HashMap<String,String> getUPNPDynamicInfo(String interfaceName) {
		HashSet<String> dynamicStates = new HashSet<String>();
		dynamicStates.add("NumberOfActiveConnections");

		HashSet<String> dynamicActions = new HashSet<String>();
		dynamicActions.add("GetTotalBytesReceived");
		dynamicActions.add("GetTotalBytesSent");
		dynamicActions.add("GetTotalPacketsReceived");
		dynamicActions.add("GetTotalPacketsSent");
		dynamicActions.add("GetStatusInfo");

		HashMap<String,String> dynamicTypesToVals = new HashMap<String,String>();
		InternetGatewayDevice[] gatewayDevices = getGatewayDevices(interfaceName);

		if (gatewayDevices != null){
			for(InternetGatewayDevice d : gatewayDevices){
				List<UPNPDevice> upnpdevice = d.getIGDRootDevice().getChildDevices();
				for(Iterator<UPNPDevice> deviceIter = upnpdevice.iterator(); deviceIter.hasNext();){
					UPNPDevice mydev = deviceIter.next();
					List<UPNPService> devServices = mydev.getServices();
					for(Iterator<UPNPService> serviceIter = devServices.iterator(); serviceIter.hasNext();){
						UPNPService myserv = serviceIter.next();
						UPNPMessageFactory msgFactory = UPNPMessageFactory.getNewInstance(myserv);
						for(Iterator<String> actionIter = myserv.getAvailableActionsName(); actionIter.hasNext();){
							String action = actionIter.next();
							ServiceAction servact = myserv.getUPNPServiceAction(action);
							if (servact != null){
								if(dynamicActions.contains(action)){
									ActionMessage msg = msgFactory.getMessage(action);
									try {
										List<String> serviceActions = servact.getOutputActionArgumentsNames();
										if(serviceActions != null){
											for(int ii = 0; ii < serviceActions.size(); ii++){
												String servActArg = serviceActions.get(ii).toString();
												String response = msg.service().getOutActionArgumentValue(servActArg);
												dynamicTypesToVals.put(servActArg.replaceFirst("New", ""), response);
											}
										}
									} catch (IOException e) {
										// TODO Auto-generated catch block
										e.printStackTrace();
									} catch (UPNPResponseException e) {
										// TODO Auto-generated catch block
										e.printStackTrace();
									}
								}
							}
						}
						for(Iterator<String> stateIter = myserv.getAvailableStateVariableName(); stateIter.hasNext();){
							String stateName = stateIter.next();
							try {
								if(dynamicStates.contains(stateName)){
									dynamicTypesToVals.put(stateName, myserv.getUPNPServiceStateVariable(stateName).getValue());
								}
							} catch (UPNPResponseException e) {
								e.printStackTrace();
							} catch (IOException e) {
								e.printStackTrace();
							}
						}
					}
				}
			}
		}
		if(dynamicTypesToVals.size() > 0){
			return dynamicTypesToVals;
		} else {
			return null;
		}
	}
}
